#include <stdio.h>
#include <string.h>

#include "config.h"
#include "lint.h"
#include "array.h"
#include "interpret.h"
#include "object.h"
#include "exec.h"
/*
 * Write statistics about objects on file.
 */

extern struct object *obj_list;

int svalue_size(struct svalue *v)
{
  int i, total;

  switch(v->type) {
  case T_FUNCTION:
  case T_FLOAT:
  case T_NUMBER:
  case T_OBJECT:
    return 0;

  case T_STRING:
    return my_strlen(v) + 4; /* Includes some malloc overhead. */

  case T_ALIST_PART:
  case T_LIST:
  case T_MAPPING:
  case T_POINTER:
    for (i=0, total = 0; i < v->u.vec->size; i++) {
      total += svalue_size(&v->u.vec->item[i]) + sizeof (struct svalue);
    }
    return total;
  default:

    fatal("Illegal type: %d\n", (void *)(int)v->type,0,0,0,0,0,0,0,0);
  }
  /*NOTREACHED*/
  return 0;
}

int data_size(ob)
    struct object *ob;
{
    int total = 0, i;
    if (ob->prog) {
        for (i = 0; i < ob->prog->num_variables; i++)
    	    total += svalue_size(&ob->variables[i]) + sizeof (struct svalue);
    }
    return total;
}

void dumpstat()
{
    FILE *f;
    struct object *ob;
    int tmp;

    f = fopen("OBJ_DUMP", "w");
    if (f == 0)
	return;
    for (ob = obj_list; ob; ob = ob->next_all)
    {
      tmp = 0;
      fprintf(f, "%s(%d) %d ref %d %s %s %s (%d)\n", ob->prog->name,
	      ob->clone_number,
	      (int)(tmp + data_size(ob) + sizeof (struct object)), ob->ref,
	      ob->flags & O_HEART_BEAT ? "HB" : "-",
	      ob->obj_index ? "HASHED" : "-",
	      ob->super ? ob->super->prog->name : "NO_ENV",ob->cpu);
    }
    fclose(f);
}
