#include "efuns.h"
#include <arpa/telnet.h>
#include "simulate.h"
#include "array.h"
#include "main.h"
#include "stralloc.h"
#include "call_out.h"

void f_get_function(int num_arg,struct svalue *argp)
{
  int function;
  struct object *ob;

  ob=argp[0].u.ob;

  if(argp[1].subtype==STRING_SHARED)
    function=find_shared_string_function(argp[1].u.string,ob->prog);
  else
    function=find_function(argp[1].u.string,ob->prog);

  pop_stack();
  pop_stack();
  if(function>=0)
  {
    sp++;
    sp->type=T_FUNCTION;
    sp->u.ob=ob;
    add_ref(ob, "F_GET_FUNCTION");
    sp->subtype=function;
  }else{
    push_number(0);
  }
}

void f_get_lfun(int num_arg,struct svalue *argp)
{
  int function;
  struct object *ob;

  ob=argp[0].u.ob;
  function=argp[1].u.number;
  
  if(function<0 || function>=LF_MAX)
    error("Too large lfun number.\n",0,0,0,0,0,0,0,0,0);
  if(ob->prog->funindex)
  {
    function=ob->prog->lfuns[function];
  }else{
    function=find_function(lfuns[function],ob->prog);
  }

  pop_stack();
  pop_stack();
  if(function>=0)
  {
    sp++;
    sp->type=T_FUNCTION;
    sp->u.ob=ob;
    add_ref(ob, "F_GET_FUNCTION");
    sp->subtype=function;
  }else{
    push_number(0);
  }
}

/* This is needed to write recursive lambda-functions */
void f_this_function(int num_arg,struct svalue *argp)
{
  if(current_object->flags & O_DESTRUCTED)
  {
    push_zero();
  }else{
    sp++;
    sp->type=T_FUNCTION;
    sp->u.ob=current_object;
    add_ref(current_object, "F_THIS_FUNCTION");
    sp->subtype=csp->func;
  }
}

void f_function_name(int num_arg,struct svalue *argp)
{
  char *name;
  if(sp->u.ob->flags & O_DESTRUCTED)
  {
    name=NULL;
  }else{
    name=copy_shared_string(FUNC(sp->u.ob->prog,sp->subtype)->name);
  }
  pop_stack();
  if(name!=NULL)
  {
    push_shared_string(name);
  }else{
    push_zero();
  }
}

void f_function_object(int num_arg,struct svalue *argp)
{
  if(sp->u.ob->flags & O_DESTRUCTED)
  {
    free_svalue(sp);    
  }else{
    sp->type=T_OBJECT;
  }
}

void f_call_function(int num_arg,struct svalue *argp)
{
  if(argp[0].type==T_NUMBER)
  {
    argp[0].subtype=NUMBER_UNDEFINED;
    pop_n_elems(num_arg-1);
    return;
  }
  if(argp[0].type==T_POINTER)
  {
    struct vector *v;
    v=map_array(argp[0].u.vec,&const0,argp+1,num_arg-1);
    pop_n_elems(num_arg);
    push_vector(v);
    v->ref--;
    return;
  }
  check_eval_cost();
  apply_lambda_low(argp->u.ob,argp->subtype,num_arg-1);
  if(sp==argp) push_zero();
  f_swap();
  pop_stack();
}

void f_expunge(int num_arg,struct svalue *argp)
{
  current_object->flags |= O_EXPUNGE;
}

void f_find_object(int num_arg,struct svalue *argp)
{
  struct object *o;
  char *name;
  name=sp->u.string;
  if(*name=='/') name++;
  o=lookup_my_object_hash(name);
  pop_stack();
  if(o) push_object(o);
  else push_zero();
}


void f_l_sizeof(int num_arg,struct svalue *argp)
{
  int a;
  switch(sp->type)
  {
  case T_STRING:  a=my_strlen(sp); break;
  case T_POINTER:
    a=sp->u.vec->size; break;
  case T_LIST:
  case T_MAPPING: a=sp->u.vec->item[0].u.vec->size; break;
  case T_OBJECT:  a=sp->u.ob->prog->total_size; break;
  default: bad_arg(1); return;
  }
  pop_stack();
  push_number(a);
}

void f_m_sizeof(int num_arg,struct svalue *argp) { f_l_sizeof(num_arg,argp); }
void f_size(int num_arg,struct svalue *argp) { f_l_sizeof(num_arg,argp); }

void f_make_object(int num_arg,struct svalue *argp)
{
  struct program *o;
  o=make_object(sp->u.string);
  pop_stack();
  if(!o) push_zero();
  else push_shared_string(copy_shared_string(o->name));
}

void f_strstr(int num_arg,struct svalue *argp)
{
  int e,start;
  char *a,*b;
  int alen,blen;

  if(num_arg==3)
  {
    check_argument(2,T_NUMBER);
    start=argp[2].u.number;
    pop_stack();
  }else{
    start=0;
  }

  a=argp[0].u.string;
  alen=my_strlen(argp+0);
  b=argp[1].u.string;
  blen=my_strlen(argp+1);
  
  if(blen)
  {
    for(e=start;e<alen-blen;e++)
    {
      if(memcmp(a+e,b,blen)) continue;
      pop_stack();
      pop_stack();
      push_number(e);
      return;
    }
  }else{
    pop_stack();
    pop_stack();
    push_zero();
    return;
  }
  
  pop_stack();
  pop_stack();
  push_number(-1);
}

void f_load(int num_arg,struct svalue *argp)
{
  if (!find_program(argp->u.string))
  {
    pop_stack();
    push_zero();
  }
}

void f_object_cpu(int num_arg,struct svalue *argp)
{
  register int e;
  e=sp->u.ob->cpu;
  pop_stack();
  push_number(e);
}

void f_clone_number(int num_arg,struct svalue *argp)
{
  register int e;
  e=sp->u.ob->clone_number;
  pop_stack();
  push_number(e);
}

void f_next_clone(int num_arg,struct svalue *argp)
{
  struct object *pp;
  char *f;

  if(sp->type==T_STRING)
  {
    pp=obj_list;
    f=sp->u.string;
  }else{
    pp=sp->u.ob->next_all;
    f=sp->u.ob->prog->name;
  }
  f=check_valid_path(f, current_object->eff_user, "next_clone", 0);
  if(!f) error("Invalid use of next_clone()\n",0,0,0,0,0,0,0,0,0);
  f=findstring(f);
  if(f)
    while(pp && (!pp->clone_number || pp->prog->name!=f)) pp=pp->next_all;
  else 
    pp=0;
  pop_stack(); 
  if(pp)
    push_object(pp);
  else
    push_zero();
}

void f_hash_name(int num_arg,struct svalue *argp)
{
  char *s;
  if((s=sp->u.ob->obj_index))
  {
    pop_stack();
    push_string(s,STRING_SHARED);
  }else{
    pop_stack();
    push_zero();
  }
}

void f_previous_objects(int num_arg,struct svalue *argp)
{
  struct vector *v;
  int e;
  struct control_stack *p;
  struct svalue sv;

  for (e=0,p = &control_stack[1]; p <= csp; p++) e++;
  v=allocate_array(e);
  for (e=0,p = &control_stack[1]; p <= csp; p++,e++)
  {
    if(!(p->ob->flags & O_DESTRUCTED))
    {
      sv.type=T_OBJECT;
      sv.u.ob=p->ob;
      assign_svalue_no_free(&(v->item[e]),&sv);
    }
  }
  push_vector(v);
  v->ref--;
}

void f_initiating_object(int num_arg,struct svalue *argp)
{
  if(control_stack[1].ob->flags & O_DESTRUCTED)
    push_zero();
  else
    push_object(control_stack[1].ob);
}

void f_created(int num_arg,struct svalue *argp)
{
  register int e;
  e=sp->u.ob->created;
  pop_stack();
  push_number(e);
}

void f_query_num_arg(int num_arg,struct svalue *argp)
{
  push_number(csp->num_of_arguments);
}

void f_get_varargs(int num_arg,struct svalue *argp)
{
  if(csp->va_args)
  {
    push_vector(csp->va_args);
  }else{
    push_zero();
  }
}

void f_update(int num_arg,struct svalue *argp)
{
  struct program *p;
  p=find_program2(sp->u.string);
  pop_stack();
  if(p)
  {
    remove_program_hash(p);
    free_prog(p,1);
    push_one();
  }else{
    push_zero();
  }
}

void f_array_of_index(int num_arg,struct svalue *argp)
{
  struct svalue *i;
  struct vector *v,*v1;
  int s,nr,e;

  v=argp[0].u.vec;
  s=v->size;
  nr=argp[1].u.number;
  v1=allocate_array(s);
  for(e=0;e<s;e++)
  {
    i=&(v->item[e]);
    if(i->type==T_POINTER && i->u.vec->size>nr)
      assign_svalue(&(v1->item[e]),&(i->u.vec->item[nr]));
  }
  pop_n_elems(2);
  push_vector(v1);
  v1->ref--;
}

int cp_cpu(struct object **o1,struct object **o2) 
{ return (*o1)->cpu - (*o2)->cpu; }

void f_top_cpu(int num_arg,struct svalue *argp)
{
#define NUM_TOP 20
  struct vector *vec;
  struct object *obs[NUM_TOP],*p;
  int e,d;
  for(p=obj_list,e=0;e<NUM_TOP && p;p=p->next_all)
    if(p->clone_number) obs[e++]=p;
  qsort((void *)obs,e,sizeof(struct object *),cp_cpu);
  for(;p;p=p->next_all)
  {
    if(!p->clone_number || p->cpu <= (*obs)->cpu ) continue;
    for(d=1;d<NUM_TOP && p->cpu > obs[d]->cpu;d++) obs[d-1]=obs[d];
    obs[d-1]=p;
  }
  vec=allocate_array(e);
  for(d=0;e--;d++)
  {
    vec->item[d].type=T_OBJECT;
    (vec->item[d].u.ob=obs[e])->ref++;
  }
  push_vector(vec);
  vec->ref--;
}

void f_set_this_player(int num_arg,struct svalue *argp)
{
  if((sp->u.ob->flags & O_ENABLE_COMMANDS) && !(sp->u.ob->flags & O_DESTRUCTED))
    command_giver=sp->u.ob;
  pop_stack();
  if(command_giver)
  {
    push_object(command_giver);
  }else{
    push_zero();
  }
}

#ifdef F_FILE_LENGTH

/*
 * file_length() efun, returns the number of lines in a file.
 */
int file_length(file)
  char *file;
{
  struct stat st;
  FILE *f;
  unsigned int i;
  int ret;

  file = check_valid_path(file, current_object->eff_user, "file_size", 0);
  if (!file) return -1;
  if (stat(file, &st) == -1)
    return -1;
  if (st.st_mode == S_IFDIR)
    return -2;
  if (!(f = fopen(file, "r")))
    return -1;
  ret = 0;
  for (i=0; i<st.st_size; i++) if (fgetc(f) == '\n') ret++;
  fclose(f);
  return ret;
} /* end of file_length() */

void f_file_length(int num_arg,struct svalue *argp)
{
  int l;

  l = file_length(sp->u.string);
  pop_stack();
  push_number(l);
}
#endif				/* F_FILE_LENGTH */

#ifdef F_REMOVE_ACTION

/*
 * removes an action from objects in the inventory and enviroment of
 * current_object.  the "cmd" must be defined by the object.
 */
int remove_action(cmd, from)
  char *cmd;
  struct object *from;
{
  struct sentence **s, *t;
  int n;

  n = 0;
  for (s=&(from->sent); *s; ) {
    if ((*s)->ob == current_object) {
      if (strcasecmp((*s)->verb, cmd)) {
        s = &((*s)->next);
        continue;
      }
      t = *s;
      s = &((*s)->next);
      free_sentence(t);
      n++;
    }
  }
  return n;
} /* end of remove_action() */

void f_remove_action(int num_arg,struct svalue *argp)
{
  int n;

  n = remove_action((sp-1)->u.string, sp->u.ob);
  pop_n_elems(2);
  push_number(n);

}
#endif				/* F_REMOVE_ACTION */

#ifdef F_QUERY_ACTIONS

void f_query_actions(int num_arg,struct svalue *argp)
{
  struct vector *v;
  if(num_arg==1)
  {
    v=query_actions(argp[0].u.ob,NULL);
  }else{
    v=query_actions(argp[0].u.ob,argp[1].u.ob);
    pop_stack();
  }
  pop_stack();
  push_vector(v);
  v->ref--;			/* fix ref count */
}
#endif				/* F_QUERY_ACTIONS */

void f_clone_object(int num_arg,struct svalue *argp)
{
  struct object *ob;
  ob = clone_object(sp->u.string,0);
  pop_stack();
  if (ob) {
    sp++;
    sp->type = T_OBJECT;
    sp->u.ob = ob;
    add_ref(ob, "F_CLONE_OBJECT");
  } else {
    push_zero();
  }
}


void f_tail(int num_arg,struct svalue *argp)
{
  char *c;
  if (c=tail(sp->u.string))
    push_shared_string(c);
  else
    assign_svalue(sp, &const0);
}

void f_save_object(int num_arg,struct svalue *argp)
{
  push_shared_string(save_object());
}

void f_code_value(int num_arg,struct svalue *argp)
{
  char *s;
  extern int one_line;
  one_line=0;
  switch(num_arg)
  {
    default: /* make gcc happy */
    case 2:
      one_line=1;
      pop_stack(); /* fall through */

    case 1:
      init_buf();
      save_svalue(argp);
      s=free_buf();
      pop_stack();
      break;

  }
  push_shared_string(s);
}

#ifdef F_WRITE_FILE

void f_write_file(int num_arg,struct svalue *argp)
{
  int i;
  i = write_file((sp-1)->u.string, sp);
  pop_n_elems(2);
  push_number(i);
}
#endif

#ifdef F_READ_FILE

void f_read_file(int num_arg,struct svalue *argp)
{
  char *str;
  int start = 0, len = 0;

  if (num_arg > 1)
    start = argp[1].u.number;
  if (num_arg == 3) {
    check_argument(2,T_NUMBER);
    len = argp[2].u.number;
  }

  str = read_file(argp[0].u.string, start, len);
  pop_n_elems(num_arg);
  if (str == 0)
    push_zero();
  else {
    push_string(str, STRING_MALLOC);
    free(str);
  }
}
#endif

#ifdef F_READ_BYTES

void f_read_bytes(int num_arg,struct svalue *argp)
{
  char *str;
  int start = 0, len = 0x7fffffff;

  if (num_arg > 1)
    start = argp[1].u.number;
  if (num_arg == 3) {
    check_argument(2,T_NUMBER);
    len = argp[2].u.number;
  }

  str = read_bytes(argp[0].u.string, start, len);
  pop_n_elems(num_arg);
  if (str == 0)
    push_zero();
  else
    push_shared_string(str);
}
#endif

#ifdef F_WRITE_BYTES

void f_write_bytes(int num_arg,struct svalue *argp)
{
  int i;
  i = write_bytes((sp-2)->u.string, (sp-1)->u.number, sp);
  pop_n_elems(3);
  push_number(i);
}
#endif

void f_file_size(int num_arg,struct svalue *argp)
{
  int i;
  i = file_size(sp->u.string);
  pop_stack();
  push_number(i);
}

void f_restore_object(int num_arg,struct svalue *argp)
{
  int i;
  i = restore_object(sp->u.string);
  pop_stack();
  push_number(i);
}

void f_decode_value(int num_arg,struct svalue *argp)
{
  char *i;
  struct svalue s;
  i = restore_svalue(sp->u.string,&s);
  pop_stack();
  push_svalue(&s);
  free_svalue(&s);
}

void f_this_player(int num_arg,struct svalue *argp)
{
  if (command_giver && !(command_giver->flags & O_DESTRUCTED))
    push_object(command_giver);
  else
    push_zero();
}

#ifdef F_FIRST_INVENTORY
void f_first_inventory(int num_arg,struct svalue *argp)
{
  struct object *ob;
  ob = first_inventory(sp);
  pop_stack();
  if (ob)
    push_object(ob);
  else
    push_zero();
}
#endif				/* F_FIRST_INVENTORY */

#ifdef F_GETUID
void f_getuid(int num_arg,struct svalue *argp)
{
  struct object *ob;
  char *tmp;
  ob = sp->u.ob;
#ifdef DEBUG
  if (ob->user == 0)
    fatal("User is null pointer\n",0,0,0,0,0,0,0,0,0);
#endif
  tmp = ob->user->name;
  pop_stack();
  push_new_shared_string(tmp);
}
#endif				/* F_GETUID */

#ifdef F_GETEUID
void f_geteuid(int num_arg,struct svalue *argp)
{
  struct object *ob;
  ob = sp->u.ob;

  if (ob->eff_user)
  {
    char *tmp;
    tmp = ob->eff_user->name;
    pop_stack();
    push_new_shared_string(tmp);
  } else {
    pop_stack();
    push_zero();
  }
}
#endif				/* F_GETEUID */

#ifdef F_EXPORT_UID
void f_export_uid(int num_arg,struct svalue *argp)
{
  struct object *ob;
  if (current_object->eff_user == 0)
    error("Illegal to export uid 0\n",0,0,0,0,0,0,0,0,0);
  ob = sp->u.ob;
  if (ob->eff_user)		/* Only allowed to export when null */
    return;
  ob->user = current_object->eff_user;
}
#endif				/* F_EXPORT_UID */

#ifdef F_SETEUID
void f_seteuid(int num_arg,struct svalue *argp)
{
  struct svalue *ret;

  if (sp->type == T_NUMBER)
  {
    if (sp->u.number != 0) bad_arg(1);
    current_object->eff_user = 0;
    pop_stack();
    push_one();
    return;
  }
  argp = sp;
  if (argp->type != T_STRING) bad_arg(1);
  push_object(current_object);
  push_svalue(argp);
  ret = apply_master_ob("valid_seteuid", 2);
  if (ret == 0 || ret->type != T_NUMBER || ret->u.number != 1)
  {
    pop_stack();
    push_zero();
    return;
  }
  current_object->eff_user = add_name(argp->u.string);
  pop_stack();
  push_one();
}
#endif				/* F_SETEUID */

void f_shutdown(int num_arg,struct svalue *argp)
{
  startshutdowngame();
}


void f_query_load_average(int num_arg,struct svalue *argp)
{
  push_new_shared_string(query_load_av());
}

void f_throw(int num_arg,struct svalue *argp)
{
  /* marion
   * the return from catch is now done by a 0 throw
   */
  assign_svalue(&catch_value, sp--);
  throw_error();	/* do the longjump, with extra checks... */
}

void f_query_host_name(int num_arg,struct svalue *argp)
{
  extern char hostname[100];
  push_string(hostname, STRING_CONSTANT);
}

#ifdef F_NEXT_INVENTORY
void f_next_inventory(int num_arg,struct svalue *argp)
{
  struct object *ob;
  ob = sp->u.ob;
  pop_stack();
  if (ob->next_inv)
    push_object(ob->next_inv);
  else
    push_zero();
}
#endif				/* F_NEXT_INVENTORY */

void f_all_inventory(int num_arg,struct svalue *argp)
{
  struct vector *vec;
  vec = all_inventory(sp->u.ob);
  pop_stack();
  if (vec == 0) {
    push_zero();
  } else {
    push_vector(vec);		/* This will make ref count == 2 */
    vec->ref--;
  }
}

void f_deep_inventory(int num_arg,struct svalue *argp)
{
  struct vector *vec;

  vec = deep_inventory(sp->u.ob, 0);
  free_svalue(sp);
  sp->type = T_POINTER;
  sp->u.vec = vec;
}

void f_environment(int num_arg,struct svalue *argp)
{
  struct object *ob;
  if (num_arg) {
    ob = environment(sp);
    pop_stack();
  } else if (!(current_object->flags & O_DESTRUCTED)) {
    ob = current_object->super;
  } else
    ob = 0;
  if (ob)
    push_object(ob);
  else
    push_zero();
}

void f_this_object(int num_arg,struct svalue *argp)
{
  if (current_object->flags & O_DESTRUCTED)
    push_zero();
  else
    push_object(current_object);
}

void f_previous_object(int num_arg,struct svalue *argp)
{
  if (previous_ob == 0 || (previous_ob->flags & O_DESTRUCTED))
    push_zero();
  else
    push_object(previous_ob);
}

void f_time(int num_arg,struct svalue *argp)
{
  if(batch_mode)
    current_time=get_current_time();
  push_number(current_time);
}

void f_wizlist(int num_arg,struct svalue *argp)
{
  struct vector *v;
  if (num_arg)
  {
    v=wizlist(sp->u.string);
    pop_stack();
  } else {
    v=wizlist(0);
  }
  if(v)
  {
    push_vector(v);
    v->ref--;
  }else{
    push_zero();
  }
}

#ifdef F_ADD_WORTH
void f_add_worth(int num_arg,struct svalue *argp)
{
  struct svalue *ret;
  ret=apply_master_ob("valid_add_worth",0);
  if(!ret || (ret->type==T_NUMBER && ret->u.number==0))
    error("Illegal call of add_worth.\n",0,0,0,0,0,0,0,0,0);
  
  if (num_arg == 2) {
    if (sp->u.ob->user)
      sp->u.ob->user->total_worth += (sp-1)->u.number;
    pop_stack();
  } else {
    if (previous_ob == 0)
      return;
    if (previous_ob->user)
      previous_ob->user->total_worth += sp->u.number;
  }
  pop_stack();
}
#endif				/* F_ADD_WORTH */

#ifdef F_LOG_FILE
void f_log_file(int num_arg,struct svalue *argp)
{
  log_file((sp-1)->u.string, sp->u.string);
  pop_stack();
  pop_stack();
}
#endif				/* F_LOG_FILE */


void f_intp(int num_arg,struct svalue *argp)
{
  pop_push_conditional(sp->type==T_NUMBER);
}

void f_floatp(int num_arg,struct svalue *argp)
{
  pop_push_conditional(sp->type==T_FLOAT);
}

void f_stringp(int num_arg,struct svalue *argp)
{
  pop_push_conditional(sp->type==T_STRING);
}

void f_objectp(int num_arg,struct svalue *argp)
{
  pop_push_conditional(sp->type==T_OBJECT);
}

void f_pointerp(int num_arg,struct svalue *argp)
{
  pop_push_conditional(sp->type==T_POINTER);
}

void f_zero_type(int num_arg,struct svalue *argp)
{
  sp->u.number=sp->subtype;
}

void f_mappingp(int num_arg,struct svalue *argp)
{
  pop_push_conditional(sp->type==T_MAPPING);
}

void f_listp(int num_arg,struct svalue *argp)
{
  pop_push_conditional(sp->type==T_LIST);
}

void f_functionp(int num_arg,struct svalue *argp)
{
  pop_push_conditional(sp->type==T_FUNCTION);
}

void f_extract(int num_arg,struct svalue *argp)
{
  int len, from, to;
  char *res;

  len = strlen(argp[0].u.string);
  if (num_arg == 1)
    return;			/* Simply return argument */
  from = argp[1].u.number;
  if (from < 0)
    from = len + from;
  if (from >= len)
  {
    pop_n_elems(num_arg);
    push_string("", STRING_CONSTANT);
    return;
  }
  if (num_arg == 2) {
    res = string_copy(argp->u.string + from);
    pop_n_elems(2);
    push_malloced_string(res);
    return;
  }
  if (argp[2].type != T_NUMBER)
    error("Bad third argument to extract()\n",0,0,0,0,0,0,0,0,0);
  to = argp[2].u.number;
  if (to < 0)
    to = len + to;
  if (to < from) {
    pop_n_elems(3);
    push_new_shared_string("");
    return;
  }
  if (to >= len)
    to = len-1;
  if (to == len-1) {
    res = string_copy(argp->u.string + from);
    pop_n_elems(3);
    push_malloced_string(res);
    return;
  }
  res = xalloc(to - from + 2);
  strncpy(res, argp[0].u.string + from, to - from + 1);
  res[to - from + 1] = '\0';
  pop_n_elems(3);
  push_malloced_string(res);
}

void f_query_verb(int num_arg,struct svalue *argp)
{
  if (last_verb == 0) {
    push_zero();
    return;
    }
  push_new_shared_string(last_verb);
}

void f_file_name(int num_arg,struct svalue *argp)
{
  char *name;

  name = copy_shared_string(sp->u.ob->prog->name);
  pop_stack();
  push_shared_string(name);
}

void f_write(int num_arg,struct svalue *argp)
{
  do_write(sp);
  pop_stack();
}

void f_move_object(int num_arg,struct svalue *argp)
{
  struct object *o;
  o = sp->u.ob;
  move_object(current_object,o);
  pop_stack();
}

void f_function_exists(int num_arg,struct svalue *argp)
{
  char *str;

  str = function_exists((sp-1)->u.string, sp->u.ob);
  pop_n_elems(2);
  if (str){
    push_shared_string(copy_shared_string(str));
  } else {
    push_zero();
  }
}

void f_add_action(int num_arg,struct svalue *argp)
{
  char *a;
  int c;

  a=argp[0].u.string;
  if(argp[1].u.ob!=current_object)
  {
    error("Add action with function pointer to other object not allowed.\n",0,0,0,0,0,0,0,0,0);
  }
  if(num_arg>2 && argp[2].type==T_NUMBER)
    c=argp[2].u.number; else c=0;

  add_action(a,argp+1,c);
  pop_n_elems(num_arg);
}

void f_crypt(int num_arg,struct svalue *argp)
{
  char salt[2];
  char *res;
  char *choise =
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";

  if (sp->type == T_STRING && strlen(sp->u.string) >= 2) {
    salt[0] = sp->u.string[0];
    salt[1] = sp->u.string[1];
  } else {
    salt[0] = choise[random_number(strlen(choise))];
    salt[1] = choise[random_number(strlen(choise))];
  }
#ifdef sun
  res = make_shared_string(_crypt((sp-1)->u.string, salt));
#else
  res = make_shared_string((char *)crypt((sp-1)->u.string, salt));
#endif
  pop_n_elems(2);
  push_shared_string(res);
}

void f_destruct(int num_arg,struct svalue *argp)
{
  if (num_arg == 0) push_object(current_object);
  else if(sp->u.ob != current_object)
  {
    struct object *a;
    struct svalue *r; 
    a=sp->u.ob;
    r=apply_master_ob("query_valid_destruct",1);
    if(!r || (r->type==T_NUMBER && r->u.number==0))
      error("Illegal call to destruct().\n",0,0,0,0,0,0,0,0,0);
    push_object(a);
  }
  destruct_object(sp->u.ob);
  pop_stack();
}

void f_random(int num_arg,struct svalue *argp)
{

  if (sp->u.number <= 0)
  {
    sp->u.number = 0;
    return;
  }
  sp->u.number = random_number(sp->u.number);
}

#ifdef F_TELL_ROOM
void f_tell_room(int num_arg,struct svalue *argp)
{
  struct object *ob;
  extern struct vector null_vector;
  struct vector *avoid;

  ob = argp[0].u.ob;
  if (num_arg == 2) {
    avoid = &null_vector;
    avoid->ref++;
  } else {
    static struct vector *vtmp;
    check_argument(2,T_POINTER);
    vtmp=allocate_array(1);

    if(argp[2].u.vec->ref > 1)
    {
      vtmp->item[0].u.vec = slice_array(argp[2].u.vec, 0,
					argp[2].u.vec->size-1);
      vtmp->item[0].type=T_POINTER;
    }else{
      vtmp->item[0].u.vec = argp[2].u.vec;
      vtmp->item[0].u.vec->ref++;
      vtmp->item[0].type=T_POINTER;
    }
    pop_stack();
    order_alist(vtmp);
    avoid = vtmp->item[0].u.vec;
  }
  tell_room(ob, sp, avoid);
  free_vector(avoid);
  pop_stack();
  pop_stack();
}
#endif				/* F_TELL_ROOM */

void f_strlen(int num_arg,struct svalue *argp)
{
  int i;
  i = my_strlen(sp);
  pop_stack();
  push_number(i);
}

void f_lower_case(int num_arg,struct svalue *argp)
{
  int i;
  char *str = string_copy(sp->u.string);
  for (i = strlen(str)-1; i>=0; i--)
    if (str[i]>='A' && str[i]<='Z')
      str[i] -= 'A' - 'a';
  pop_stack();
  push_malloced_string(str);
}

void f_upper_case(int num_arg,struct svalue *argp)
{
  int i;
  char *str = string_copy(sp->u.string);
  for (i = strlen(str)-1; i>=0; i--)
    if (str[i]>='a' && str[i]<='z')
      str[i] += 'A' - 'a';
  pop_stack();
  push_malloced_string(str);
}

void f_set_heart_beat(int num_arg,struct svalue *argp)
{
  int i;
  i = set_heart_beat(current_object, sp->u.number);
  sp->u.number = i;
}

void f_capitalize(int num_arg,struct svalue *argp)
{
  if (islower(sp->u.string[0]))
  {
    char *str;

    str = string_copy(sp->u.string);
    str[0] += 'A' - 'a';
    pop_stack();
    push_malloced_string(str);
  }
}

void f_command(int num_arg,struct svalue *argp)
{
  int i;
  if (num_arg == 1)
  {
    i = command_for_object(argp[0].u.string, 0);
  }else{
    i = command_for_object(argp[0].u.string, argp[1].u.ob);
  }
  pop_n_elems(num_arg);
  push_number(i);
}

void f_rm(int num_arg,struct svalue *argp)
{
  int i;
  i = remove_file(sp->u.string);
  pop_stack();
  push_number(i);
}

void f_mkdir(int num_arg,struct svalue *argp)
{
  char *path;
  int i;
  path = check_valid_path(sp->u.string, current_object->eff_user, "mkdir", 1);

  if(path && mkdir(path, 0770) != -1)
    i=1;
  else
    i=0;
  pop_stack();
  push_number(i);
}

void f_rmdir(int num_arg,struct svalue *argp)
{
  char *path;
  path = check_valid_path(sp->u.string, current_object->eff_user, "rmdir", 1);
  push_conditional(path == 0 || rmdir(path) == -1);
}

void f_enable_commands(int num_arg,struct svalue *argp)
{
  enable_commands(1);
}
void f_disable_commands(int num_arg,struct svalue *argp)
{
  enable_commands(0);
}

void f_present(int num_arg,struct svalue *argp)
{
  struct object *ob;
  ob = object_present(argp, argp[1].u.ob);
  pop_n_elems(2);
  if (ob)
    push_object(ob);
  else
    push_zero();
}

void f_ctime(int num_arg,struct svalue *argp)
{
  char *cp,*str;
  str = time_string(sp->u.number);
  cp = strchr(str, '\n');
  if (cp) *cp = '\0';
  pop_stack();
  push_new_shared_string(str); /* Per, som vill ha ctime :) */
}


void f_cindent(int num_arg,struct svalue *argp)
{
  char *path;

  path = check_valid_path(sp->u.string, current_object->eff_user, "cindent", 1);
  if (path) {
    if (indent_program(path))
    {
      pop_stack();
      push_one();
      return;
    }
  } else {
    error("Illegal attempt to indent\n",0,0,0,0,0,0,0,0,0);
  }
  pop_stack();
  push_zero();
}

/*
 * do_rename is used by the efun rename. It is basically a combination
 * of the unix system call rename and the unix command mv. Please shoot
 * the people at ATT who made Sys V.
 */

#ifdef F_RENAME
int
do_rename(fr, t)
    char *fr, *t;
{
    char *from, *to;
    
    from = check_valid_path(fr, current_object->eff_user, "do_rename", 1);
    if(!from)
	return 1;
    to = check_valid_path(t, current_object->eff_user, "do_rename", 1);
    if(!to)
	return 1;
    if(!strlen(to) && !strcmp(t, "/")) {
	to = (char *)allocaX(3);
	sprintf(to, "./");
    }
    strip_trailing_slashes (from);
    if (isdir (to))
	{
	    /* Target is a directory; build full target filename. */
	    char *cp;
	    char *newto;
	    
	    cp = strrchr (from, '/');
	    if (cp)
		cp++;
	    else
		cp = from;
	    
	    newto = (char *) allocaX (strlen (to) + 1 + strlen (cp) + 1);
	    sprintf (newto, "%s/%s", to, cp);
	    return do_move (from, newto);
	}
    else
	return do_move (from, to);
}

void f_rename(int num_arg,struct svalue *argp)
{
  int i;
  i = do_rename((sp-1)->u.string, sp->u.string);
  pop_n_elems(2);
  push_number(i);
}
#endif /* F_RENAME */

void f_debug_info(int num_arg,struct svalue *argp)
{
  struct svalue res;
  struct object *ob;
  int i;

  switch ( argp[0].u.number ) {
  case 0:
  {
    int flags;
    struct object *obj2;

    if (num_arg != 2)
      error("bad number of arguments to debug_info",0,0,0,0,0,0,0,0,0);
    check_argument(1,T_OBJECT);

    ob = argp[1].u.ob;
    flags = ob->flags;
    vtell_object("O_HEART_BEAT      : %s\n",
		 flags&O_HEART_BEAT      ?"TRUE":"FALSE");
    vtell_object("O_IS_WIZARD       : %s\n",
		 flags&O_IS_WIZARD       ?"TRUE":"FALSE");
    vtell_object("O_ENABLE_COMMANDS : %s\n",
		 flags&O_ENABLE_COMMANDS ?"TRUE":"FALSE");
    vtell_object("O_DESTRUCTED      : %s\n",
		 flags&O_DESTRUCTED      ?"TRUE":"FALSE");
    vtell_object("O_APPROVED        : %s\n",
		 flags&O_APPROVED        ?"TRUE":"FALSE");
    vtell_object("O_RESET_STATE     : %s\n",
		 flags&O_RESET_STATE     ?"TRUE":"FALSE");
    vtell_object("O_WILL_CLEAN_UP   : %s\n",
		 flags&O_WILL_CLEAN_UP   ?"TRUE":"FALSE");
    vtell_object("O_IS_ARCH         : %s\n",
		 flags&O_IS_ARCH         ?"TRUE":"FALSE");

    /*	        vtell_object("total light : %d\n", ob->total_light);  */
    vtell_object("next_reset  : %d\n", ob->next_reset);
    vtell_object("time_of_ref : %d\n", ob->time_of_ref);
    vtell_object("ref         : %d\n", ob->ref);
#ifdef DEBUG
    vtell_object("extra_ref   : %d\n", ob->extra_ref);
#endif
    vtell_object("clone_number: %d\n",ob->clone_number);
    vtell_object("name        : '%s'\n", ob->prog->name);
    vtell_object("next_all    : OBJ(%s)\n",
		 ob->next_all?ob->next_all->prog->name:"NULL");
    if (obj_list == ob) vtell_object(
				     "This object is the head of the object list.\n");
    for (obj2=obj_list,i=1; obj2; obj2=obj2->next_all,i++)
      if (obj2->next_all == ob)
      {
	vtell_object(
		     "Previous object in object list: OBJ(%s)\n",
		     obj2->prog->name);
	vtell_object("position in object list:%d\n",i);
      }
    assign_svalue_no_free(&res,&const0);
    break;
  }
  case 1: {
    if (num_arg != 2)
      error("bad number of arguments to debug_info",0,0,0,0,0,0,0,0,0);
    check_argument(1,T_OBJECT);

    ob = argp[1].u.ob;
		
    vtell_object("program ref's %d\n", ob->prog->ref);
    vtell_object("Name %s\n", ob->prog->name);
    vtell_object("program size %d\n",
		 ob->prog->program_size);
    vtell_object("num func's %d (%d) \n", ob->prog->num_functions
		 ,ob->prog->num_functions * sizeof(struct function));
    vtell_object("num strings %d\n", ob->prog->num_strings);
    vtell_object("num vars %d (%d)\n", ob->prog->num_variables
		 ,ob->prog->num_variables * sizeof(struct variable));
    vtell_object("num inherits %d (%d)\n", ob->prog->num_inherited
		 ,ob->prog->num_inherited * sizeof(struct inherit));
    vtell_object("total size %d\n", ob->prog->total_size);
    assign_svalue_no_free(&res,&const0);
    break;
  }
  default: bad_arg(1);
  }
  pop_n_elems(num_arg);
  sp++;
  *sp=res;
}

void f_master(int num_arg, struct svalue *argp) { push_object(master_ob); }

extern struct object *obj_list;
extern struct program **obj_table;
void gc_svalue(struct svalue *foo);

void gc_object(struct object *ob)
{
  int e;
  for(e=0;e<ob->prog->num_variables;e++)
    gc_svalue(ob->variables+e);
}

void gc_svalue(struct svalue *foo)
{
  int e;
  switch(foo->type)
  {
    case T_OBJECT:
    case T_FUNCTION:
     if(foo->u.ob->flags & O_DESTRUCTED)
     {
       free_svalue(foo);
     }else{
       gc_object(foo->u.ob);
     }
     break;

    case T_POINTER:
    case T_MAPPING:
    case T_ALIST_PART:
    case T_LIST:
     check_alist_for_destruct(foo->u.vec);
     for(e=0;e<foo->u.vec->size;e++)
     {
       gc_svalue(foo->u.vec->item+e);
     }
  }
}

void do_gc(int foo)
{
  struct object *o;
  int e,d;
  struct program *p,*next;

  for(o=obj_list;o;o=o->next_all) gc_object(o);

  for (e=0;e<num_pending_calls;e++)
    for(d=0;d<pending_calls[e]->args->size;d++)
      gc_svalue(pending_calls[e]->args->item+d);
  
  if(!foo) return;
  if(obj_table)
  {
    int om;

    do{
      om=0;
      for(e=0;e<OTABLE_SIZE;e++)
      {
	for(p=obj_table[e];p;p=next)
	{
	  next=p->next_hash;
	  if(p->ref==1)
	  {
	    remove_program_hash(p);
	    free_prog(p,1);
	    p=obj_table[e];
	    om=1;
	  }
	}
      }
    }while(om);
  }
}

void f_gc(int num_arg, struct svalue *argp)
{
  do_gc(argp->u.number);
  pop_stack();
}

#ifdef F_DUMP_MALLOC_DATA
void f_dump_malloc_data(int num_arg, struct svalue *argp)
{
  push_shared_string(dump_malloc_data());
}
#endif

void f_dump_string_data(int num_arg, struct svalue *argp)
{
  int i;
  i=sp->u.number;
  pop_stack();
  push_shared_string(add_string_status(i));
}

void f_dump_heart_beat_data(int num_arg, struct svalue *argp)
{
  int i;
  i=sp->u.number;
  pop_stack();
  push_shared_string(heart_beat_status(i));
}

void f_dump_call_out_data(int num_arg, struct svalue *argp)
{
  int i;
  i=sp->u.number;
  pop_stack();
  push_shared_string(print_call_out_usage(i));
}

void f_dump_prog_table_data(int num_arg, struct svalue *argp)
{
  int i;
  i=sp->u.number;
  pop_stack();
  push_shared_string(show_otable_status(i));
}

void f_dump_index_table_data(int num_arg, struct svalue *argp)
{
  int i;
  i=sp->u.number;
  pop_stack();
  push_shared_string((char *)show_ohash_status(i));
}

void f_dumpallobj(int num_arg, struct svalue *argp) { dumpstat(); }

void f_set_max_eval_cost(int num_arg, struct svalue *argp)
{
  int i;
  extern int max_eval_cost;
  i=argp[0].u.number;
  assert_master_ob_loaded();
  if(!current_object->eff_user || current_object->eff_user!=master_ob->user)
    return;
  max_eval_cost=i;
  pop_stack();
}

void f_reset_eval_cost(int num_arg, struct svalue *argp)
{
  int i;
  extern int eval_cost;
  i=argp[0].u.number;
  assert_master_ob_loaded();
  if(!current_object->eff_user || current_object->eff_user!=master_ob->user)
    return;
  eval_cost=0;
}

void f_range(int num_arg, struct svalue *argp)
{
  check_argument(1,T_NUMBER);

  if (argp[0].type == T_POINTER)
  {
    struct vector *v;

    v = slice_array(sp[-2].u.vec, sp[-1].u.number, sp[0].u.number);
    pop_n_elems(3);
    if (v)
    {
      push_vector(v);
      v->ref--;			/* Will make ref count == 1 */
    }else{
      push_zero();
    }
  } else if (argp[0].type == T_STRING) {
    int len, from, to;
    char *res;
	    
    len = my_strlen(argp);
    from = argp[1].u.number;
    to = argp[2].u.number;

    if(from<0) from+=len;
    if (from >= len || to<from)
    {
      pop_n_elems(3);
      push_string("", STRING_CONSTANT);
      return;
    }
    if (to >= len) to = len-1;
    if (to == len-1 && from==0)
    {
      pop_stack();
      pop_stack();
      return;
    }
    res = make_shared_binary_string(sp[-2].u.string + from, to-from+1);
    pop_n_elems(3);
    push_shared_string(res);
  } else {
    error("Bad argument to [ .. ] range operand.\n",0,0,0,0,0,0,0,0,0);
  }
}

void f_combine_path(int num_arg, struct svalue *argp)
{
  char *path;
  if(num_arg==2)
  {
    path=combine_path(argp[1].u.string,argp[0].u.string);
  }else{
    path=combine_path(0,argp[0].u.string);
  }
  pop_n_elems(2);
  push_malloced_string(path);
}

void f_copy_value(int num_arg,struct svalue *argp)
{
  struct svalue s;
  s=*argp;
  copy_svalue(argp,argp);
  free_svalue(&s);
}
