#ifndef INTERPRET_H
#define INTERPRET_H
union u {
  float fnum;
  char *string;
  int number;
  struct object *ob;
  struct vector *vec;
  struct lvalue *lvalue;
};

/*
 * The value stack element.
 * If it is a string, then the way that the string has been allocated differ,
 * wich will affect how it should be freed.
 */
struct svalue {
    short type;
    unsigned short subtype;
    union u u;
};

#define T_INVALID	0x0
#define T_LVALUE	0x1
#define T_NUMBER	0x2
#define T_INT T_NUMBER
#define T_STRING	0x4
#define T_POINTER	0x8
#define T_OBJECT	0x10
#define T_MAPPING	0x20
#define T_LIST		0x40
#define T_ALIST_PART    0x100
#define T_FUNCTION      0x200
#define T_FLOAT         0x400
#define T_ANY (T_NUMBER|T_STRING|T_POINTER|T_OBJECT|T_MAPPING|T_LIST|T_ALIST_PART|T_FUNCTION|T_FLOAT)
/* these types uses the .u.vec */
#define T_VECTOR (T_ALIST_PART | T_POINTER | T_MAPPING | T_LIST)

#define STRING_MALLOC	0	/* Allocated by malloc() */
#define STRING_CONSTANT	1	/* Pointer to static buffer */
#define STRING_SHARED	2	/* Allocated by the shared string library */
#define STRING_TEMPORARY 3	/* Pointer to some buffer */

/* different types of zeros */
#define NUMBER_NUMBER 0 /* normal number */
#define NUMBER_UNDEFINED 1 /* value not present */
#define NUMBER_DESTRUCTED_OBJECT 2 /* a destructed object */
#define NUMBER_DESTRUCTED_FUNCTION 3 /* a function in a destructed object */

#define LVALUE_VECTOR_INDEX 0
#define LVALUE_STRING_INDEX 1
#define LVALUE_MAPPING_INDEX 2
#define LVALUE_LIST_INDEX 3
#define LVALUE_RANGE 4

#define LVALUES 200

struct lvalue
{
  int type;
  struct svalue *sval;
  struct svalue *parent;
  int start,end;
};

struct lnode_def;
void free_vector PROT((struct vector *));
void free_all_values();

/*
 * Control stack element.
 * 'prog' is usually same as 'ob->prog' (current_object), except when
 * when the current function is defined by inheritance.
 * The pointer, csp, will point to the values that will be used at return.
 */
struct control_stack
{
    struct object *comgiver;	/* Always save the command giver... */
    struct object *ob;		/* Current object */
    struct object *prev_ob;	/* Save previous object */
    struct program *prog;	/* Current program */
    int num_local_variables;	/* Local + arguments */
    int num_of_arguments;	/* Local + arguments */
    char *pc;
    struct svalue *fp;
    struct function *funp;	/* Only used for tracebacks */
    int func;			/* funp==prog->funcions[func] */
    int function_index_offset;	/* Used when executing functions in inherited
				   programs */
    int variable_index_offset;	/* Same */
    struct vector *va_args;     /* remaining arguments */
};

#define IS_ZERO(x) (!(x) || (((x)->type == T_NUMBER) && ((x)->u.number == 0)))

struct svalue *apply_lfun(int fun,struct object *ob,int num_arg);
struct svalue *apply PROT((char *, struct object *, int));
void push_string PROT((char *, int));
void push_number PROT((int));
void push_object PROT((struct object *));
int inter_sscanf PROT((int));
struct svalue *apply_master_ob PROT((char *fun, int num_arg));
void assert_master_ob_loaded();
int apply_lambda_low(struct object *,int,int);
struct svalue *apply_lambda(struct svalue *,int);
struct svalue *safe_apply (char *fun,struct object * ob,int num_arg);
void check_eval_cost();
char *function_exists PROT((char *, struct object *));
int get_line_number_if_any PROT((void));
void check_a_lot_ref_counts PROT((struct program *));
void clear_state PROT((void));
void reset_machine PROT((int));
void push_svalue PROT((struct svalue *));
void push_constant_string PROT((char *));
void push_new_shared_string(char *p);
void call_function PROT((struct program *, struct function *));
int find_function(char *fun,struct program *prog);
int find_shared_string_function(char *fun,struct program *prog);
void pop_stack();
void push_zero();
void push_shared_string(char *);
void push_malloced_string(char *);
void pop_n_elems(int);
void bad_arg(int arg);
void push_vector(struct vector *v);
void push_one();
struct svalue *find_value(int num);
void eval_instruction(void);

void assign_svalue PROT((struct svalue *, struct svalue *));
void move_svalue PROT((struct svalue *, struct svalue *));
void assign_svalue_no_free PROT((struct svalue *to, struct svalue *from));
void free_svalue PROT((struct svalue *));

#endif
