#include <stdio.h>
#include <setjmp.h>
#include "config.h"
#include "lint.h"
#include "lang.h"
#include "interpret.h"
#include "las.h"
#include "instrs.h"
#include "array.h"
#include "exec.h"
#include "object.h"
#include "simul_efun.h"
#include "incralloc.h"
#include "stralloc.h"
#include "main.h"
#include "operators.h"
#include "dynamic_buffer.h"

#define LASDEBUG
int lasdebug=0;

node *eval(node *);
int eval_low(node *n);
node *optimize(node *n,node *parent);

struct vector *current_switch_mapping=0;
int *break_stack=0;
int current_break,break_stack_size;
int *continue_stack=0;
int current_continue,continue_stack_size;
extern int exact_types;

struct program fake_prog;
struct object fake_object;

struct mem_block mem_block[NUMAREAS];

extern int num_parse_error;
extern char **local_names;
extern unsigned short *type_of_locals;
extern int current_number_of_locals;
extern int current_line;
extern char *get_type_name(int);
void nice_print_tree(node *);

#define PC mem_block[A_PROGRAM].current_size

INLINE void add_to_mem_block(int n,char *data,int size)
{
  struct mem_block *mbp = &mem_block[n];
  while (mbp->current_size + size > mbp->max_size) {
    mbp->max_size <<= 1;
    mbp->block = realloc((char *)mbp->block, mbp->max_size);
  }
  memcpy(mbp->block + mbp->current_size, data, size);
  mbp->current_size += size;
}

static int store_linenumbers=1;
void store_line_number_info(unsigned short line)
{
  unsigned short offset,entrys,*ent;
  if(!store_linenumbers) return;
  offset = PC;
  entrys = mem_block[A_LINENUMBERS].current_size/sizeof(unsigned short);
  ent =(unsigned short *)mem_block[A_LINENUMBERS].block;
  if(entrys && ent[entrys-1]==line) return;
  add_to_mem_block(A_LINENUMBERS, (char *)&offset, sizeof offset);
  add_to_mem_block(A_LINENUMBERS, (char *)&line, sizeof line);
}

void ins_byte(char b)
{
  add_to_mem_block(A_PROGRAM, &b, 1);
}

/*
 * Store a 2 byte number. It is stored in such a way as to be sure
 * that correct byte order is used, regardless of machine architecture.
 * Also beware that some machines can't write a word to odd addresses.
 */
void ins_short(short l)
{
  add_to_mem_block(A_PROGRAM, (char *)&l, sizeof(short));
}

#if YYDEBUG
extern int yydebug;
#endif
void upd_short(int offset, short l)
{
#if YYDEBUG
  if(yydebug>1)
    fprintf(stderr,"Upd short %d %d\n",offset,l);
#endif
  mem_block[A_PROGRAM].block[offset + 0] = ((char *)&l)[0];
  mem_block[A_PROGRAM].block[offset + 1] = ((char *)&l)[1];
}

void upd_branch(int offset, int l)
{
  short tmp;
#if YYDEBUG
  if(yydebug>1)
    fprintf(stderr,"Upd branch %d %d\n",offset,l);
#endif
  tmp=l-offset;
  mem_block[A_PROGRAM].block[offset + 0] = ((char *)&tmp)[0];
  mem_block[A_PROGRAM].block[offset + 1] = ((char *)&tmp)[1];
}

short read_short(int offset)
{
  short l;

  ((char *)&l)[0] = mem_block[A_PROGRAM].block[offset + 0];
  ((char *)&l)[1] = mem_block[A_PROGRAM].block[offset + 1];
  return l;
}

/*
 * Store a 4 byte number. It is stored in such a way as to be sure
 * that correct byte order is used, regardless of machine architecture.
 */
void ins_long(long l)
{
  add_to_mem_block(A_PROGRAM, (char *)&l+0, sizeof(long));
}
#if YYDEBUG
extern int yydebug;
#endif

void ins_f_byte(unsigned int b)
{
#if YYDEBUG
  if(yydebug)
    fprintf(stderr,"Inserting f_byte %s (%d) at %d\n",get_f_name(b),b,
	    PC);
#endif
#if defined(LASDEBUG)
  if(lasdebug>1)
    if(store_linenumbers)
      fprintf(stderr,"Inserting f_byte %s (%d) at %d\n",get_f_name(b),b,
	      PC);
#endif
  b-=F_OFFSET;
#ifdef OPCPROF
  add_compiled(b);
#endif
  if(b>255)
  {
    ins_f_byte(F_ADD_256);
    b-=256;
  }
  ins_byte((char)b);
}

/*
 * A mechanism to remember addresses on a stack. The size of the stack is
 * defined in config.h.
 */
int comp_stackp;
int comp_stack[COMPILER_STACK_SIZE];

void push_address()
{
 if (comp_stackp >= COMPILER_STACK_SIZE)
 {
   yyerror("Compiler stack overflow");
   comp_stackp++;
   return;
 }
 comp_stack[comp_stackp++] = PC;
}

void push_explicit(int address)
{
  if (comp_stackp >= COMPILER_STACK_SIZE)
  {
    yyerror("Compiler stack overflow");
    comp_stackp++;
    return;
  }
  comp_stack[comp_stackp++] = address;
}

int pop_address()
{
  if (comp_stackp == 0)
     fatal("Compiler stack underflow.\n",0,0,0,0,0,0,0,0,0);
  if (comp_stackp > COMPILER_STACK_SIZE)
  {
    --comp_stackp;
    return 0;
  }
  return comp_stack[--comp_stackp];
}

void ins_int(int i)
{
  if ( i == 0 ) {
    ins_f_byte(F_CONST0);
  }else if ( i == 1 ) {
    ins_f_byte(F_CONST1);
  } else if ( i == -1) {
    ins_f_byte(F_CONST_1);
  } else if ( i>1 && i<258) {
    ins_f_byte(F_BYTE); ins_byte(i-2);
  } else if ( i<-1 && i>-258) {
    ins_f_byte(F_NEG_BYTE); ins_byte(-i-2);
  } else if( i>=-37768 && i<32768){
    ins_f_byte(F_SHORT); ins_short(i);
  }else{
    ins_f_byte(F_NUMBER); ins_long(i);
  }
}

void ins_float(float f)
{
  ins_f_byte(F_FLOAT);
  add_to_mem_block(A_PROGRAM,(char *)&f,sizeof(float));
}

short store_prog_string(char *str)
{
    short i;
    char **p;

    p = (char **) mem_block[A_STRINGS].block;

    for (i=mem_block[A_STRINGS].current_size / sizeof str -1; i>=0; --i)
      if (p[i] == str)
	return i;

    str=copy_shared_string(str);
    add_to_mem_block(A_STRINGS, (char *)&str, sizeof str);
    return mem_block[A_STRINGS].current_size / sizeof str - 1;
}

int store_constant(struct svalue *foo)
{
  struct svalue *s,tmp;
  int size,e;
  s=(struct svalue *)mem_block[A_CONSTANTS].block;
  size=mem_block[A_CONSTANTS].current_size / sizeof(struct svalue);
  for(e=0;e<size;e++)
    if(is_equal(s+e,foo))
      return e;

  assign_svalue_no_free(&tmp,foo);
  add_to_mem_block(A_CONSTANTS,(char *)&tmp,sizeof(struct svalue));
  return size;
}

#ifdef LINMALLOC_DEBUG
extern void assert_malloced_memory(void *);

void assert_node(node *n)
{
  if(!n) return;
  assert_malloced_memory((char *)n);
  switch(n->token)
  {
  case F_CONSTANT:
    /* assert_svalue_malloced(&(n->u.sval)); */
    break;

  case F_STRING:
    /* assert_malloced_memory(n->u.s.a.str); */
    break;

  default:
    if(n->opt_info & OPT_A_IS_NODE)
      assert_node(n->u.s.a.node);
    if(n->opt_info & OPT_B_IS_NODE)
      assert_node(n->u.s.b.node);
  }
}
#endif

int compute_opt_info(node *n,int needlval,int recompute)
{
  int tmp1,tmp2,try_optimize;

  if(!n) return OPT_CONST;

  if((n->opt_info & OPT_OPT_COMPUTED) && !recompute && !needlval)
    return n->opt_info;

  if(recompute==1) recompute=0;

#ifdef LINMALLOC_DEBUG
  assert_node(n);
#endif

  try_optimize=OPT_TRY_OPTIMIZE | OPT_COULD_BE_CONST;
  tmp1=tmp2=OPT_CONST;
  switch(n->token)
  {
  case F_NUMBER:
  case F_STRING:
  case F_FLOAT:
  case F_CONSTANT:
    try_optimize=OPT_COULD_BE_CONST;
    break;

  case F_CAST:
    tmp1=compute_opt_info(n->u.s.a.node,needlval,recompute);
    if(n->type==TYPE_OBJECT)
    {
      if((n->u.s.a.node->type & TYPE_MOD_MASK) == TYPE_STRING ||
	 (n->u.s.a.node->type & TYPE_MOD_MASK) == TYPE_ANY)
      {
	tmp2=OPT_SIDE_EFFECT;
      }else if((n->u.s.a.node->type & TYPE_MOD_MASK) == TYPE_NUMBER){
	tmp2=0;
      }
    }
    break;

  case F_CONSTANT_FUNCTION:
    try_optimize=OPT_COULD_BE_CONST;
    if(needlval!=-1)
    {
      tmp1=OPT_CONST;
    }else{
      if((FUNCTIONP(n->u.s.a.number)->flags & NAME_PROTOTYPE) ||
	!(FUNCTIONP(n->u.s.a.number)->type & TYPE_MOD_INLINE))
      {
	tmp1=OPT_SIDE_EFFECT | OPT_RESTORE_OBJECT; /* assume the worst */
      }else{
	tmp1=FUNCTIONP(n->u.s.a.number)->type;
	tmp1=(tmp1&TYPE_MOD_CONSTANT?OPT_CONST:0) |
	  (tmp1&TYPE_MOD_SIDE_EFFECT?OPT_SIDE_EFFECT:0) ;
	if(!(tmp1&OPT_CONST)) try_optimize=OPT_TRY_OPTIMIZE;
      }
    }
    break;

  case F_PUSH_SIMUL_EFUN:
    try_optimize=OPT_COULD_BE_CONST;
    if(needlval!=-1)
    {
      tmp1=OPT_CONST;
    }else{
      tmp1=simul_efun_prog->function_ptrs[n->u.s.a.number].type;
      tmp1=(tmp1&TYPE_MOD_CONSTANT?OPT_CONST:0) |
	(tmp1&TYPE_MOD_SIDE_EFFECT?OPT_SIDE_EFFECT:0) ;
      if(!(tmp1&OPT_CONST)) try_optimize=OPT_TRY_OPTIMIZE;
    }
    break;

  case F_EFUN_CALL:
    switch(n->u.s.a.number)
    {
    case F_RESTORE_OBJECT:
      tmp1=instrs[n->u.s.a.number-F_OFFSET].ret_type;
      tmp1=(tmp1&TYPE_MOD_CONSTANT?OPT_CONST:0) |
	(tmp1&TYPE_MOD_SIDE_EFFECT?OPT_SIDE_EFFECT:0)|
	  OPT_RESTORE_OBJECT;
      tmp2=compute_opt_info(n->u.s.b.node,0,recompute);
      try_optimize=0;
      break;
      
    case F_CALL_FUNCTION:
      tmp1=compute_opt_info(n->u.s.b.node,-1,recompute);
      if(!(tmp1&OPT_CONST)) tmp1|=OPT_SIDE_EFFECT;
      break;

    default:
      tmp1=instrs[n->u.s.a.number-F_OFFSET].ret_type;
      tmp1=(tmp1&TYPE_MOD_CONSTANT?OPT_CONST:0) |
	(tmp1&TYPE_MOD_SIDE_EFFECT?OPT_SIDE_EFFECT:0) ;
      tmp2=compute_opt_info(n->u.s.b.node,0,recompute);
      if(!(tmp1&OPT_CONST)) try_optimize=OPT_TRY_OPTIMIZE;
    }
    break;

  case F_CONTINUE:
    tmp1=OPT_CONT;
    break;

  case F_BREAK:
    tmp1=OPT_BREAK;
    break;

  case F_BRANCH:
  case F_DO:
    tmp1=compute_opt_info(n->u.s.a.node,0,recompute) & ~OPT_CONT & ~OPT_BREAK;
    tmp2=compute_opt_info(n->u.s.b.node,0,recompute);
    break;

  case F_INC:
  case F_INC_AND_POP:
  case F_POST_INC:
  case F_DEC:
  case F_DEC_AND_POP:
  case F_POST_DEC:
    tmp1=compute_opt_info(n->u.s.a.node,OPT_LVALUE_CHANGE,recompute);
    break;

  case F_FOREACH:
  case F_INC_NEQ_LOOP:
  case F_DEC_NEQ_LOOP:
  case F_INC_LOOP:
  case F_DEC_LOOP:
  case F_FOR:
    tmp2=compute_opt_info(n->u.s.a.node,0,recompute);
    tmp1=compute_opt_info(n->u.s.b.node,0,recompute) & ~OPT_CONT & ~OPT_BREAK;

  case F_ASSIGN_AND_POP:
  case F_ASSIGN:
    tmp1=compute_opt_info(n->u.s.a.node,OPT_LVALUE_OVERWRITE,recompute);
    tmp2=compute_opt_info(n->u.s.b.node,0,recompute);
    break;

  case F_AND_EQ:
  case F_OR_EQ:
  case F_XOR_EQ:
  case F_LSH_EQ:
  case F_RSH_EQ:
  case F_ADD_EQ:
  case F_SUB_EQ:
  case F_MULT_EQ:
  case F_MOD_EQ:
  case F_DIV_EQ:
    tmp1=compute_opt_info(n->u.s.a.node,OPT_LVALUE_CHANGE,recompute);
    tmp2=compute_opt_info(n->u.s.b.node,0,recompute);
    break;

  case ' ':
    tmp1=compute_opt_info(n->u.s.a.node,0,recompute);
    tmp2=compute_opt_info(n->u.s.b.node,OPT_LVALUE_CHANGE,recompute);
    break;

  case F_LOCAL:
    if(needlval)
    {
      tmp1=OPT_LOCAL_ASSIGNMENT;
    }else{
      tmp1=0;
    }
    break;

  case F_GLOBAL:
    if(needlval)
    {
      tmp1=OPT_GLOBAL_ASSIGNMENT;
    }else{
      tmp1=OPT_GLOBAL_DEPEND;
    }
    break;

  case F_INDEX:
    /* hm */
    tmp1=compute_opt_info(n->u.s.a.node,needlval?OPT_LVALUE_CHANGE:0,recompute);
    tmp2=compute_opt_info(n->u.s.b.node,0,recompute);
    needlval=0;
    break;

  case '{':
  case F_ARG_LIST:
    try_optimize=OPT_COULD_BE_CONST;
    if(needlval==-1)
    {
      tmp1=compute_opt_info(n->u.s.a.node,needlval,recompute);
      tmp2=compute_opt_info(n->u.s.b.node,0,recompute);
    }else{
      tmp1=compute_opt_info(n->u.s.a.node,0,recompute);
      tmp2=compute_opt_info(n->u.s.b.node,needlval,recompute);
    }
    break;

  case F_LVALUE_LIST:
    try_optimize=OPT_COULD_BE_CONST;
    tmp1=compute_opt_info(n->u.s.a.node,OPT_LVALUE_CHANGE,recompute);
    tmp2=compute_opt_info(n->u.s.b.node,OPT_LVALUE_CHANGE,recompute);
    break;

  case F_CASE:
  case F_DEFAULT:
    tmp1=OPT_CASE | compute_opt_info(n->u.s.a.node,0,recompute);
    tmp2=OPT_CASE | compute_opt_info(n->u.s.b.node,0,recompute);
    break;

  case F_RETURN:
    tmp1=OPT_RETURN;
    tmp2=compute_opt_info(n->u.s.a.node,0,recompute);
    break;

  case F_SWITCH:
    tmp1=compute_opt_info(n->u.s.a.node,0,recompute);
    tmp2=compute_opt_info(n->u.s.b.node,0,recompute) & ~OPT_CASE & ~OPT_BREAK;
    break;
    
  default:
    if(n->opt_info & OPT_A_IS_NODE)
    {
      tmp1=compute_opt_info(n->u.s.a.node,needlval,recompute);
    }else{
      tmp1=OPT_CONST;
    }

    if(n->opt_info & OPT_B_IS_NODE)
    {
      tmp2=compute_opt_info(n->u.s.b.node,needlval,recompute);
    }else{
      tmp2=OPT_CONST;
    }
  }

  n->opt_info=(tmp1 & tmp2 & OPT_CONST) |
	 ((tmp1 | tmp2) & (OPT_SIDE_EFFECT |
			 OPT_GLOBAL_ASSIGNMENT |
			 OPT_LOCAL_ASSIGNMENT |
			 OPT_RESTORE_OBJECT |
			 OPT_TRY_OPTIMIZE |
			 OPT_CONT |
			 OPT_BREAK |
			 OPT_RETURN |
			 OPT_GLOBAL_DEPEND |
			 OPT_CASE)) |
	 (n->opt_info & (OPT_A_IS_NODE | OPT_B_IS_NODE) |
	  OPT_OPT_COMPUTED |
	  try_optimize |
	  (needlval & (OPT_LVALUE_CHANGE | OPT_LVALUE_OVERWRITE))
	  );

#ifdef LASDEBUG
  if(lasdebug>3)
  {
    fprintf(stderr,"Computed [%c%c%c%c%c%c%c%c%c%c%c%c%c%c] opt info for:\n",
	   (n->opt_info & OPT_A_IS_NODE)?'A':'-',
	   (n->opt_info & OPT_B_IS_NODE)?'B':'-',
	   (n->opt_info & OPT_CONST)?'C':'-',
	   (n->opt_info & OPT_SIDE_EFFECT)?'X':'-',
	   (n->opt_info & OPT_LOCAL_ASSIGNMENT)?'l':'-',
	   (n->opt_info & OPT_GLOBAL_ASSIGNMENT)?'g':'-',
	   (n->opt_info & OPT_RESTORE_OBJECT)?'r':'-',
	   (n->opt_info & OPT_OPT_COMPUTED)?'O':'-',
	   (n->opt_info & OPT_TRY_OPTIMIZE)?'T':'-',
	   (n->opt_info & OPT_CASE)?'c':'-',
	   (n->opt_info & OPT_CONT)?'c':'-',
	   (n->opt_info & OPT_CONT)?'b':'-',
	   (n->opt_info & OPT_RETURN)?'r':'-',
	   (n->opt_info & OPT_GLOBAL_DEPEND)?'G':'-'
	   );
    nice_print_tree(n);
  }
#endif

  return n->opt_info;
}


void free_node(node *n)
{
  if(!n) return;
  switch(n->token)
  {
  case F_CONSTANT:
    free_svalue(&(n->u.sval));
    break;

  case F_STRING:
    free_string(n->u.s.a.str);
    break;

  default:
    if(n->opt_info & OPT_A_IS_NODE)
      free_node(n->u.s.a.node);
    if(n->opt_info & OPT_B_IS_NODE)
      free_node(n->u.s.b.node);
  }
  free((char *)n);
}

node *mknode(short token,node *a,node *b)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->u.s.a.node=a;
  res->u.s.b.node=b;
  res->token=token;
  res->type=0;
  res->opt_info=OPT_A_IS_NODE | OPT_B_IS_NODE;
  res->lnum=current_line;
  return res;
}

node *mkstrnode(char *str)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->token=F_STRING;
  res->type=TYPE_STRING;
  res->opt_info=OPT_CONST | OPT_OPT_COMPUTED | OPT_OPTIMIZED;
#ifdef DEBUG
  free_string(copy_shared_string(str));
#endif
  res->u.s.a.str=str;
  res->lnum=current_line;
  return res;
}

node *mkintnode(int nr)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->token=F_NUMBER;
  res->type=TYPE_NUMBER;
  res->opt_info=OPT_CONST | OPT_OPT_COMPUTED | OPT_OPTIMIZED;
  res->u.s.a.number=nr;
  res->lnum=current_line;
  return res;
}

node *mkfloatnode(float foo)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->token=F_FLOAT;
  res->type=TYPE_FLOAT;
  res->opt_info=OPT_CONST | OPT_OPT_COMPUTED | OPT_OPTIMIZED;
  res->u.s.a.fnum=foo;
  res->lnum=current_line;
  return res;
}

node *mkconstfunnode(int fun)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->token=F_CONSTANT_FUNCTION;
  res->type=TYPE_FUNCTION;
  res->opt_info=OPT_CONST;
  res->u.s.a.number=fun;
  res->lnum=current_line;
  return res;
}

node *mkglobalnode(int var)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->token=F_GLOBAL;
  res->type=VARIABLE(var)->type & TYPE_MOD_MASK;
  res->opt_info=OPT_COULD_BE_CONST;
  res->u.s.a.number=var;
  res->lnum=current_line;
  return res;
}

node *mklocalnode(int var)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->token=F_LOCAL;
  res->type=type_of_locals[var] & TYPE_MOD_MASK;
  res->opt_info=OPT_COULD_BE_CONST;
  res->u.s.a.number=var;
  res->lnum=current_line;
  return res;
}

node *mksimulefunnode(int fun)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->token=F_PUSH_SIMUL_EFUN;
  res->type=TYPE_FUNCTION;
  res->opt_info=OPT_CONST;
  res->u.s.a.number=fun;
  res->lnum=current_line;
  return res;
}

node *mkcastnode(int type,node *n)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->token=F_CAST;
  res->type=type;
  res->opt_info=OPT_A_IS_NODE;
  res->u.s.a.node=n;
  res->u.s.b.node=NULL;
  res->lnum=current_line;
  return res;
}

node *mkefunnode(int efun,node *args)
{
  node *res;
  res=(node *)malloc(sizeof(node));
  res->token=F_EFUN_CALL;
  res->type=instrs[efun-F_OFFSET].ret_type;
  res->opt_info=OPT_B_IS_NODE;
  if(instrs[efun-F_OFFSET].ret_type & TYPE_MOD_CONSTANT)
  {
    res->opt_info|=OPT_CONST;
  }else{
    res->opt_info|=OPT_SIDE_EFFECT;
  }
  res->u.s.a.number=efun;
  res->u.s.b.node=args;
  res->lnum=current_line;
  return res;
}

int can_use_constants(struct svalue *s,int svalues)
{
  int e;
  for(e=0;e<svalues;e++)
  {
    switch(s[e].type)
    {
    case T_FUNCTION:
    case T_OBJECT:
      return 0;
      break;

    case T_LIST:
    case T_MAPPING:
    case T_POINTER:
    case T_ALIST_PART:
      if(!can_use_constants(s[e].u.vec->item,s[e].u.vec->size))
	return 0;
    }
  }
  return 1;
}

node *mksvaluenode(struct svalue *s)
{
  node *res;
  int e;
  switch(s->type)
  {
  case T_STRING:
    if(s->subtype==STRING_SHARED)
    {
      return mkstrnode(copy_shared_string(s->u.string));
    }else{
      return mkstrnode(make_shared_string(s->u.string));
    }
  case T_INT: return mkintnode(s->u.number);
  case T_FLOAT: return mkfloatnode(s->u.fnum);

  case T_POINTER:
    for(e=0;e<s->u.vec->size;e++)
      if(s->u.vec->item[e].type!=T_NUMBER || s->u.vec->item[e].u.number)
	break;
    
    if(e==s->u.vec->size)
    {
      res=mkefunnode(F_ALLOCATE,mkintnode(s->u.vec->size));
      compute_opt_info(res,0,0);
      res->type=TYPE_NUMBER | TYPE_MOD_POINTER;
      return res;
    }
    
  case T_ALIST_PART:
  case T_LIST:
  case T_MAPPING:
    if(can_use_constants(s,1))
    {
      res=(node *)malloc(sizeof(node));
      res->token=F_CONSTANT;
      res->lnum=current_line;
      assign_svalue_no_free(&(res->u.sval),s);
      if(res->u.sval.type==T_ALIST_PART)
	res->u.sval.type=T_POINTER;
      res->opt_info=OPT_CONST;
      if(s->type==T_POINTER || s->type==T_ALIST_PART)
	res->type=TYPE_ANY | TYPE_MOD_POINTER;
      else if(s->type==T_LIST)
	res->type=TYPE_LIST;
      else
	res->type=TYPE_MAPPING;
      return res;
    }else{
      switch(s->type)
      {
      case T_POINTER:
      case T_ALIST_PART:
	res=mksvaluenode(s->u.vec->item);
	for(e=1;e<s->u.vec->size;e++)
	  res=mknode(F_ARG_LIST,res,mksvaluenode(s->u.vec->item+e));
	res=mkefunnode(F_AGGREGATE,res);
	compute_opt_info(res,0,0);
	res->type=TYPE_ANY | TYPE_MOD_POINTER;
	return res;

      case T_MAPPING:
	res=mkefunnode(F_MKMAPPING,
	  mknode(F_ARG_LIST,mksvaluenode(s->u.vec->item),
		 mksvaluenode(s->u.vec->item+1)));
	compute_opt_info(res,0,0);
	res->type=TYPE_MAPPING;
	return res;

      case T_LIST:
	res=mksvaluenode(s->u.vec->item[0].u.vec->item);
	for(e=1;e<s->u.vec->size;e++)
	  res=mknode(F_ARG_LIST,res,
		     mksvaluenode(s->u.vec->item[0].u.vec->item+e));
	res=mkefunnode(F_L_AGGREGATE,res);
	compute_opt_info(res,0,0);
	res->type=TYPE_LIST;
	return res;
      }
    }
    return NULL;

  case T_OBJECT:
    if(s->u.ob==&fake_object)
    {
      return mkefunnode(F_THIS_OBJECT,0);
    }else{
      res=(node *)malloc(sizeof(node));      
      res->lnum=current_line;
      res->token=F_CONSTANT;
      assign_svalue_no_free(&(res->u.sval),s);
      res->opt_info=OPT_CONST;
      res->type=TYPE_OBJECT;
      return res;
    }

  case T_FUNCTION:
    if(s->u.ob==&fake_object)
    {
      return mkconstfunnode(s->subtype);
    }else if(s->u.ob==simul_efun_object){
      return mksimulefunnode(s->subtype);
    }else{
      res=(node *)malloc(sizeof(node));      
      res->lnum=current_line;
      res->token=F_CONSTANT;
      assign_svalue_no_free(&(res->u.sval),s);
      res->opt_info=OPT_CONST;
      res->type=TYPE_FUNCTION;
      return res;
    }
  }
  return 0;
}


/* here comes the typechecking */
/* note that the typechecking does certain casts that won't be made */
/* if typechecking is turned off. */

int cant_check_types;
int typep;
int types[256];

void get_types(node *n)
{
  if(typep>255 || !n)
    return;

  switch(n->token)
  {
    case F_ARG_LIST:
      get_types(n->u.s.a.node);
      get_types(n->u.s.b.node);
      break;
      
    case F_PUSH_ARRAY:
      cant_check_types=1;
      break;
      
    default:
      if(n->type!=TYPE_VOID)
	types[typep++]=n->type;
  }
}

int match_arg(int *argp,int type)
{
  for(;*argp;argp++)
    if(TYPE(type,*argp)) return 1;
  return 0;
}

int check_efun_types(int **argp,int offset,int left,node **n,char *name,int *changed)
{
  if(!*n) return offset;
  if(offset==left) return offset;
  if((*n)->token==F_ARG_LIST)
  {
    offset=check_efun_types(argp,offset,left,&((*n)->u.s.a.node),name,changed);
    offset=check_efun_types(argp,offset,left,&((*n)->u.s.b.node),name,changed);
    return offset;
  }
  if((*n)->type==TYPE_STRING && !match_arg(*argp,TYPE_STRING))
  {
    if(match_arg(*argp,TYPE_OBJECT))
    {
      *n=mkcastnode(TYPE_OBJECT,*n);
      *changed=1;
    }else if(match_arg(*argp,TYPE_FUNCTION)){
      *changed=1;
      *n=mkcastnode(TYPE_FUNCTION,*n);
      compute_opt_info(*n,0,1);
      *n=optimize(*n,0);
    }
  }
  if(!match_arg(*argp,(*n)->type))
  {
    char buf[200];
    sprintf(buf, "Bad argument %d type to efun %s()",offset+1,name);
    yyerror(buf);
  }
  while(**argp) (*argp)++;
  (*argp)++;
  return offset+1;
}

void check_args_for_summation(node *n)
{
  int e,type;
  char buf[200];
  type=types[0];

  if(typep<1)
  {
    sprintf(buf,"Too few arguments to sumation.");
    yyerror(buf);
  }else{
    for(e=0;e<typep;e++)
    {
      types[e]&=TYPE_MOD_MASK;
      if(types[e]==TYPE_FUNCTION || types[0]==TYPE_OBJECT)
      {
	sprintf(buf,"Bad argument %d to + (%s).",e+1,get_type_name(types[e]));
	yyerror(buf);
      }
      if(type==TYPE_ANY) continue;
      if(types[e]==TYPE_ANY) { type=TYPE_ANY; continue; }
      if(types[e]==type) continue;
      if(types[e] & TYPE_MOD_POINTER) type=TYPE_MOD_POINTER;
      if(types[e]&type&TYPE_MOD_POINTER) continue;
      switch(type)
      {
      case TYPE_STRING:
	if(types[e]==TYPE_NUMBER || types[e]==TYPE_FLOAT) continue;
	break;

      case TYPE_NUMBER:
      case TYPE_FLOAT:
	if(types[e]==TYPE_STRING)
	{
	  type=TYPE_STRING;
	  continue;
	}
	break;

      }
    }
    if(type!=TYPE_ANY)
    {
      for(e=0;e<typep;e++)
      {
	switch(type)
	{
	case TYPE_STRING:
	  if(types[e]!=TYPE_STRING &&
	     types[e]!=TYPE_NUMBER &&
	     types[e]!=TYPE_FLOAT)
	  {
	    sprintf(buf,"Bad argument %d to + (%s).",e+1,
		    get_type_name(types[e]));
	    yyerror(buf);
	  }
	  break;

	case TYPE_MOD_POINTER:
	  if(!(types[e]&TYPE_MOD_POINTER))
	  {
	    sprintf(buf,"Bad argument %d to + (%s).",e+1,
		    get_type_name(types[e]));
	    yyerror(buf);
	  }
	  break;

	case TYPE_LIST:
	case TYPE_MAPPING:
	case TYPE_NUMBER:
	case TYPE_FLOAT:
	  if(types[e]!=type)
	  {
	    sprintf(buf,"Bad argument %d to + (%s).",e+1,
		    get_type_name(types[e]));
	    yyerror(buf);
	  }
	  break;
	}
      }
    }
  }
  n->type=type;
}

void check_types(node *n)
{
  char buf[200];
  if(!n) return;
  typep=0;
  cant_check_types=0;

#ifdef LINMALLOC_DEBUG
  assert_node(n);
#endif

  switch(n->token)
  {
  case F_INDEX:
    get_types(n->u.s.a.node);
    get_types(n->u.s.b.node);
    if(typep!=2)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(types[0]==TYPE_STRING || (types[0]&TYPE_MOD_POINTER))
    {
      if(!TYPE(types[1],TYPE_NUMBER))
      {
	sprintf(buf,"Index not integer.");
	yyerror(buf);
      }
      if(types[0]==TYPE_STRING)
      {
	n->type=TYPE_NUMBER;
      }else{
	n->type=types[0] & ~ TYPE_MOD_POINTER;
      }
    }else if(types[0]==TYPE_OBJECT || types[0]==TYPE_FUNCTION ||
	   types[0]==TYPE_NUMBER || types[0]==TYPE_FLOAT)
    {
      sprintf(buf,"Indexing on illegal type.");
      yyerror(buf);
    }else{
      n->type=TYPE_ANY;
    }
    break;

  case F_ADD:
    get_types(n->u.s.a.node);
    get_types(n->u.s.b.node);
    check_args_for_summation(n);
    break;

  case F_SUBTRACT:
  case F_AND:
  case F_XOR:
  case F_OR:
    get_types(n->u.s.a.node);
    get_types(n->u.s.b.node);
    if(typep!=2)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(types[0]==TYPE_FUNCTION || types[0]==TYPE_FUNCTION ||
       (n->token!=F_SUBTRACT && types[0]==TYPE_FLOAT))
    {
      sprintf(buf,"Bad argument 1 to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(!TYPE(types[0],types[1]))
    {
      sprintf(buf,"Bad arg 2 to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    n->type=types[0];
    break;

  case F_MULTIPLY:
    get_types(n->u.s.a.node);
    get_types(n->u.s.b.node);
    if(typep!=2)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(types[0]==TYPE_FUNCTION || types[0]==TYPE_OBJECT ||
       types[0]==TYPE_LIST || types[0]==TYPE_MAPPING)
    {
      sprintf(buf,"Bad arg 1 to *.");
      yyerror(buf);
    }
    if(types[1]==TYPE_FUNCTION || types[1]==TYPE_OBJECT ||
       types[1]==TYPE_LIST || types[1]==TYPE_MAPPING)
    {
      sprintf(buf,"Bad arg 2 to *.");
      yyerror(buf);
    }
    if((types[0]&TYPE_MOD_POINTER) && types[1]==TYPE_STRING)
    {
      n->type=TYPE_STRING;
    }else if(types[0]==TYPE_NUMBER && types[1]==TYPE_NUMBER){
      n->type=TYPE_NUMBER;
    }else if(types[0]==TYPE_FLOAT && types[1]==TYPE_FLOAT){
      n->type=TYPE_FLOAT;
    }else{
      n->type=TYPE_ANY;
    }
    break;

  case F_DIVIDE:
    get_types(n->u.s.a.node);
    get_types(n->u.s.b.node);
    if(typep!=2)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(types[0]==TYPE_FUNCTION || types[0]==TYPE_OBJECT ||
       types[0]==TYPE_LIST || types[0]==TYPE_MAPPING)
    {
      sprintf(buf,"Bad arg 1 to /.");
      yyerror(buf);
    }
    if(types[1]==TYPE_FUNCTION || types[1]==TYPE_OBJECT ||
       types[1]==TYPE_LIST || types[1]==TYPE_MAPPING)
    {
      sprintf(buf,"Bad arg 2 to /.");
      yyerror(buf);
    }
    if(types[0]==TYPE_STRING && types[1]==TYPE_STRING)
    {
      n->type=TYPE_STRING | TYPE_MOD_POINTER;
    }else if(types[0]==TYPE_NUMBER && types[1]==TYPE_NUMBER){
      n->type=TYPE_NUMBER;
    }else if(types[0]==TYPE_FLOAT && types[1]==TYPE_FLOAT){
      n->type=TYPE_FLOAT;
    }else{
      n->type=TYPE_ANY;
    }
    break;

  case F_NEGATE:
    get_types(n->u.s.a.node);
    if(typep!=1)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(!BASIC_TYPE(types[0],TYPE_NUMBER) && !BASIC_TYPE(types[0],TYPE_FLOAT))
    {
      sprintf(buf,"Bad 1 arg to subtract.");
      yyerror(buf);
    }
    n->type=types[0];
    break;

  case F_COMPL:
    get_types(n->u.s.a.node);
    if(typep!=1)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(!BASIC_TYPE(types[0],TYPE_NUMBER))
    {
      sprintf(buf,"Bad 1 arg to ~.");
      yyerror(buf);
    }
    n->type=types[0];
    break;

  case F_EQ:
  case F_NE:
    get_types(n->u.s.a.node);
    get_types(n->u.s.b.node);
    if(typep!=2)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    n->type=TYPE_NUMBER;
    break;
      
  case F_GT:
  case F_GE:
  case F_LT:
  case F_LE:
    get_types(n->u.s.a.node);
    get_types(n->u.s.b.node);
    if(typep!=2)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(!BASIC_TYPE(types[0],TYPE_NUMBER) && !BASIC_TYPE(types[0],TYPE_FLOAT) &&
       !BASIC_TYPE(types[1],TYPE_STRING))
    {
      sprintf(buf,"Bad arg 1 to %s.",get_f_name(n->token));
      yyerror(buf);
	
    }
    if(!BASIC_TYPE(types[0],types[1]))
    {
      sprintf(buf,"Bad arg 2 to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    n->type=TYPE_NUMBER;
    break;

  case F_NOT:
    n->type=TYPE_NUMBER;
    break;

  case F_RSH:
  case F_LSH:
  case F_MOD:
    get_types(n->u.s.a.node);
    get_types(n->u.s.b.node);
    if(typep!=2)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(!BASIC_TYPE(types[0],TYPE_NUMBER))
    {
      sprintf(buf,"Bad arg 1 to %s.",get_f_name(n->token));
      yyerror(buf);
	
    }
    if(!BASIC_TYPE(types[0],TYPE_NUMBER))
    {
      sprintf(buf,"Bad arg 2 to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    n->type=TYPE_NUMBER;
    break;

  case F_SSCANF:
    get_types(n->u.s.a.node);
    if(typep!=2)
    {
      sprintf(buf,"Wrong number of arguments to sscanf.");
      yyerror(buf);
    }
    if(!BASIC_TYPE(types[0],TYPE_STRING))
    {
      sprintf(buf,"Bad arg 1 to sscanf.");
      yyerror(buf);
    }
    if(!BASIC_TYPE(types[1],TYPE_STRING))
    {
      sprintf(buf,"Bad arg 2 to sscanf.");
      yyerror(buf);
    }
    n->type=TYPE_NUMBER;
    break;

  case F_SWAP:
    n->type=TYPE_VOID;
    break;
      
  case F_LAND:
  case F_LOR:
    get_types(n->u.s.a.node);
    get_types(n->u.s.b.node);
    if(typep!=2)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    if(types[0]==types[1])
      n->type=types[0];
    else
      n->type=TYPE_ANY;
    break;

  case F_INC:
  case F_DEC:
  case F_POST_INC:
  case F_POST_DEC:
    get_types(n->u.s.a.node);
    if(typep!=1)
    {
      sprintf(buf,"Wrong number of arguments to %s.",get_f_name(n->token));
      yyerror(buf);
    }
    n->type=TYPE_NUMBER;
    break;

  case F_EFUN_CALL:
  {
    int min,max,def,ret,*argp;
    extern int efun_arg_types[];

    get_types(n->u.s.b.node);

    if(n->u.s.a.number==F_SUM)
    {
      if(cant_check_types) break;
      check_args_for_summation(n);
      break;
    }

    min=instrs[n->u.s.a.number-F_OFFSET].min_arg;
    max=instrs[n->u.s.a.number-F_OFFSET].max_arg;
    def=instrs[n->u.s.a.number-F_OFFSET].Default;
    ret=instrs[n->u.s.a.number-F_OFFSET].ret_type;

    if(cant_check_types)
    {
      if(min==max)
	yyerror("Cannot use '@' on efun with fixed number of arguments.");
      break;
    }
      
    argp = &efun_arg_types[instrs[n->u.s.a.number-F_OFFSET].arg_index];
    ret&=TYPE_MOD_MASK;
    n->type=ret;
    if(def && typep==min-1)
    {
      node *tmp;
      if(instrs[def-F_OFFSET].efunc)
	tmp=mkefunnode(def,0);
      else
	tmp=mknode(def,0,0);
      n->u.s.b.node=mknode(F_ARG_LIST,n->u.s.b.node,tmp);
      n->opt_info&=~(OPT_OPT_COMPUTED | OPT_OPTIMIZED);
      compute_opt_info(n,0,0);
      n->u.s.b.node=optimize(n->u.s.b.node,0);
      typep=0;
      get_types(n->u.s.b.node);
    }
    if(typep<min)
    {
      sprintf(buf, "Too few arguments to %s", get_f_name(n->u.s.a.number));
      yyerror(buf);
    }
    if(typep>max && max!=-1)
    {
      sprintf(buf, "Too many arguments to %s", get_f_name(n->u.s.a.number));
      yyerror(buf);
    }
    if(exact_types)
    {
      int foo=0;
      check_efun_types(&argp,0,max==-1?min:typep,&(n->u.s.b.node),get_f_name(n->u.s.a.number),&foo);
      if(foo) compute_opt_info(n,0,2);
    }
    break;
  }

  case '?':
    get_types(n->u.s.b.node->u.s.a.node);
    get_types(n->u.s.b.node->u.s.b.node);
    if(typep!=2)
      n->type=TYPE_VOID;
    else if(types[0]==types[1])
      n->type=types[0];
    else
      n->type=TYPE_ANY;
    break;

  default:
    if(!n->type) n->type=TYPE_ANY;
  }
}

/* these routines operates on parsetrees and are mostly used by the
 * optimizer
 */

node *copy_node(node *n)
{
  node *b;
  if(!n) return n;
  switch(n->token)
  {
    case F_NUMBER:
    case F_FLOAT:
    case F_LOCAL:
    case F_GLOBAL:
    case F_CONSTANT_FUNCTION:
    case F_PUSH_SIMUL_EFUN:
      b=mkintnode(0);
      *b=*n;
      return b;

    case F_STRING:
      b=mkstrnode(copy_shared_string(n->u.s.a.str));
      break;
    case F_EFUN_CALL:
      b=mkefunnode(n->u.s.a.number,copy_node(n->u.s.b.node));
      break;
    case F_CAST:
      b=mkcastnode(n->type,copy_node(n->u.s.a.node));
      break;
    case F_CONSTANT:
      b=mksvaluenode(&(n->u.sval));
      break;

    default:
      if((n->opt_info & OPT_A_IS_NODE) && (n->opt_info & OPT_B_IS_NODE))
	b=mknode(n->token,copy_node(n->u.s.a.node),copy_node(n->u.s.b.node));
      else if(n->opt_info & OPT_A_IS_NODE)
	b=mknode(n->token,copy_node(n->u.s.a.node),n->u.s.b.node);
      else if(n->opt_info & OPT_B_IS_NODE)
	b=mknode(n->token,n->u.s.a.node,copy_node(n->u.s.b.node));
      else
	b=mknode(n->token,n->u.s.a.node,n->u.s.b.node);
  }
  b->type=n->type;
  b->lnum=n->lnum;
  b->opt_info=n->opt_info;
  return b;
}


int is_const(node *n)
{
  return (n->opt_info & OPT_CONST) &&
	 !(n->opt_info & (OPT_SIDE_EFFECT |
			OPT_CASE |
			OPT_CONT |
			OPT_BREAK |
			OPT_LOCAL_ASSIGNMENT |
			OPT_GLOBAL_ASSIGNMENT |
			OPT_RESTORE_OBJECT |
			OPT_RETURN
	   ));
         
}

/* this one supposes that the value is optimized */
int node_is_true(node *n)
{
  if(!n) return 0;
  switch(n->token)
  {
    case F_NUMBER: return n->u.s.a.number!=0;
    case F_STRING: return 1;
    case F_FLOAT: return n->u.s.a.fnum!=0.0;
    case F_CONSTANT:
    {
      switch(n->u.sval.type)
      {
      case F_NUMBER: return n->u.sval.u.number!=0;
      case F_FLOAT: return n->u.sval.u.fnum!=0.0;
      default: return 1;
      }
    }
    default: return 0;
  }
}

/* this one supposes that the value is optimized */
int node_is_false(node *n)
{
  if(!n) return 0;
  switch(n->token)
  {
    case F_NUMBER: return !n->u.s.a.number;
    case F_FLOAT: return n->u.s.a.fnum==0.0;
    case F_CONSTANT:
    {
      switch(n->u.sval.type)
      {
      case F_NUMBER: return n->u.sval.u.number==0;
      case F_FLOAT: return n->u.sval.u.fnum==0.0;
      default: return 1;
      }
    }
    default: return 0;
  }
}

int node_is_eq(node *a,node *b)
{
  if(a==b) return 1;
  if(!a || !b) return 0;
  if(a->token!=b->token) return 0;
  switch(a->token)
  {
  case F_STRING:
    return a->u.s.a.str==b->u.s.a.str;

  case F_CONSTANT_FUNCTION:
  case F_PUSH_SIMUL_EFUN:
  case F_GLOBAL:
  case F_LOCAL:
  case F_NUMBER:
    return a->u.s.a.number==b->u.s.a.number;

  case F_FLOAT:
    return a->u.s.a.fnum==b->u.s.a.fnum;

  case F_CAST:
    return a->type==b->type &&
      a->opt_info==b->opt_info &&
	node_is_eq(a->u.s.a.node,b->u.s.a.node);

  case F_EFUN_CALL:
    return a->u.s.a.number==b->u.s.a.number &&
      a->type==b->type &&
      a->opt_info==b->opt_info &&
	node_is_eq(a->u.s.b.node,b->u.s.b.node);

  case F_CONSTANT:
    return is_equal(&(a->u.sval),&(b->u.sval));

  default:
    return a->type==b->type &&
      a->opt_info==b->opt_info &&
	((a->opt_info & OPT_A_IS_NODE) && node_is_eq(a->u.s.a.node,b->u.s.a.node)) &&
	  ((a->opt_info & OPT_B_IS_NODE) && node_is_eq(a->u.s.b.node,b->u.s.b.node));
  }

}

node **last_cmd(node **a)
{
  node **n;
  if(!a || !*a) return (node **)NULL;
  if((*a)->token==F_CAST) return last_cmd(&(*a)->u.s.a.node);
  if((*a)->token!='{' && (*a)->token!=F_ARG_LIST) return a;
  if((*a)->u.s.b.node)
  {
    if((*a)->u.s.b.node->token!='{' && (*a)->u.s.b.node->token!=F_CAST &&
       (*a)->u.s.a.node->token!=F_ARG_LIST)
      return &((*a)->u.s.b.node);
    if(n=last_cmd(&(*a)->u.s.b.node))
      return n;
  }
  if((*a)->u.s.a.node)
  {
    if((*a)->u.s.a.node->token!='{' && (*a)->u.s.a.node->token!=F_CAST &&
       (*a)->u.s.a.node->token!=F_ARG_LIST)
      return &((*a)->u.s.a.node);
    if(n=last_cmd(&(*a)->u.s.a.node))
      return n;
  }
  return 0;
}

node **first_arg(node *a)
{
  node **n;
  if(a->token!=F_ARG_LIST) return 0;
  if(a->u.s.a.node)
  {
    if(a->u.s.a.node->token!=F_ARG_LIST)
      return &(a->u.s.a.node);
    if(n=first_arg(a->u.s.a.node))
      return n;
  }
  if(a->u.s.b.node)
  {
    if(a->u.s.b.node->token!=F_ARG_LIST)
      return &(a->u.s.b.node);
    if(n=first_arg(a->u.s.b.node))
      return n;
  }
  return 0;
}

void print_tree(node *foo,int needlval)
{
  if(!foo) return;
#ifdef LINMALLOC_DEBUG
  assert_node(foo);
#endif
  switch(foo->token)
  {
  case F_CONSTANT_FUNCTION:
    printf("%s",FUNC(&fake_prog,foo->u.s.a.number)->name);
    break;

  case F_LOCAL:
    if(needlval) putchar('&');
    printf("$%d",foo->u.s.a.number);
    break;

  case '?':
    printf("(");
    print_tree(foo->u.s.a.node,0);
    printf(")?(");
    print_tree(foo->u.s.b.node->u.s.a.node,0);
    printf("):(");
    print_tree(foo->u.s.b.node->u.s.b.node,0);
    printf(")");
    break;

  case F_GLOBAL:
    if(needlval) putchar('&');
    printf("%s",VARIABLE(foo->u.s.a.number)->name);
    break;

  case F_NUMBER:
    printf("%d",foo->u.s.a.number);
    break;

  case F_FLOAT:
    printf("%f",foo->u.s.a.fnum);
    break;

  case F_STRING:
    printf("\"%s\"",foo->u.s.a.str);
    break;

  case F_ASSIGN:
  case F_ASSIGN_AND_POP:
    print_tree(foo->u.s.a.node,1);
    printf("=");
    print_tree(foo->u.s.b.node,0);
    break;

  case F_CAST:
    printf("(%s){",get_type_name(foo->type));
    print_tree(foo->u.s.a.node,0);
    printf("}");
    break;

  case F_ARG_LIST:
    print_tree(foo->u.s.a.node,0);
    if(foo->u.s.a.node && foo->u.s.b.node) putchar(',');
    print_tree(foo->u.s.b.node,needlval);
    return;

  case F_LVALUE_LIST:
    print_tree(foo->u.s.a.node,1);
    if(foo->u.s.a.node && foo->u.s.b.node) putchar(',');
    print_tree(foo->u.s.b.node,1);
    return;

  case '{':
    print_tree(foo->u.s.a.node,0);
    if(foo->u.s.a.node) printf(";\n ");
    print_tree(foo->u.s.b.node,needlval);
    return;

  case F_EFUN_CALL:
    printf("efun::%s(",get_f_name(foo->u.s.a.number));
    print_tree(foo->u.s.b.node,0);
    printf(")");
    return;

  default:
    if(!(foo->opt_info & OPT_A_IS_NODE) && !(foo->opt_info & OPT_B_IS_NODE))
    {
      printf("%s",get_f_name(foo->token));
      return;
    }
    if(foo->token<256)
    {
      printf("%c(",foo->token);
    }else{
      printf("%s(",get_f_name(foo->token));
    }
    if(foo->opt_info & OPT_A_IS_NODE) print_tree(foo->u.s.a.node,0);
    if((foo->opt_info & OPT_A_IS_NODE) && (foo->opt_info & OPT_B_IS_NODE)
       && foo->u.s.a.node && foo->u.s.b.node)
      putchar(',');
    if(foo->opt_info & OPT_B_IS_NODE) print_tree(foo->u.s.b.node,0);
    printf(")");
    return;
  }
}

void nice_print_tree(node *n)
{
  print_tree(n,0);
  printf("\n");
  fflush(stdout);
}

int depend_p3(node *a,node *b)
{
  if(!a || is_const(a)) return 0;
  if(!(a->opt_info & OPT_COULD_BE_CONST)) return 2;
  if(a->token==F_GLOBAL)
    if(b->opt_info & OPT_RESTORE_OBJECT) return 2;

  if((a->opt_info & OPT_A_IS_NODE) && depend_p3(a->u.s.a.node,b))
    return 1;
  if((a->opt_info & OPT_B_IS_NODE) && depend_p3(a->u.s.b.node,b))
    return 1;
  return 0;
}

/* return 1 if a depends on b */
int depend_p2(node *a,node *b)
{
  int foo;
  if(!a || !b || is_const(a)) return 0;

  switch(b->token)
  {
  case F_INDEX:
    if(b->opt_info & ( OPT_LVALUE_CHANGE | OPT_LVALUE_OVERWRITE ))
    {
      if(a->token==F_INDEX) return 1;
      switch(a->token)
      {
      case F_ASSIGN:
      case F_ASSIGN_AND_POP:
	if((a->opt_info & OPT_B_IS_NODE) && (foo=depend_p2(a->u.s.b.node,b)))
	  return foo;
	if((a->opt_info & OPT_A_IS_NODE) && (foo=depend_p2(a->u.s.a.node,b)))
	  return foo;
	break;

      default:
	if((a->opt_info & OPT_A_IS_NODE) && (foo=depend_p2(a->u.s.a.node,b)))
	  return foo;
	if((a->opt_info & OPT_B_IS_NODE) && (foo=depend_p2(a->u.s.b.node,b)))
	  return foo;
      }
    }
    break;
    
  case F_GLOBAL:
  case F_LOCAL:
    if(b->opt_info & ( OPT_LVALUE_CHANGE | OPT_LVALUE_OVERWRITE ))
    {
      if(a->token==b->token && a->u.s.a.number==b->u.s.a.number)
      {
/*	if(a->opt_info & OPT_LVALUE_OVERWRITE)
	  return -1;
	else
*/	  return 1;
      }
      switch(a->token)
      {
      case F_ASSIGN:
      case F_ASSIGN_AND_POP:
	if((a->opt_info & OPT_B_IS_NODE) && (foo=depend_p2(a->u.s.b.node,b)))
	  return foo;
	if((a->opt_info & OPT_A_IS_NODE) && (foo=depend_p2(a->u.s.a.node,b)))
	  return foo;
	break;

      default:
	if((a->opt_info & OPT_A_IS_NODE) && (foo=depend_p2(a->u.s.a.node,b)))
	  return foo;
	if((a->opt_info & OPT_B_IS_NODE) && (foo=depend_p2(a->u.s.b.node,b)))
	  return foo;
      }
    }
    break;
  }
  if((b->opt_info & OPT_A_IS_NODE) && (0<depend_p2(a,b->u.s.a.node)))
    return 1;
  if((b->opt_info & OPT_B_IS_NODE) && (0<depend_p2(a,b->u.s.b.node)))
    return 1;
  return 0;
}

int depend_p(node *a,node *b)
{
  if(!b) return 0;
  if(!(b->opt_info & OPT_SIDE_EFFECT))
  {
    return depend_p2(a,b);
  }else{
    return depend_p3(a,b) || depend_p2(a,b);
  }
}

int find_and_clear_optimize_flag(node *n,node *a)
{
  if(!a || !n) return 0;
  if(a==n) return 1;
  if(((n->opt_info & OPT_A_IS_NODE) &&
      find_and_clear_optimize_flag(n->u.s.a.node,a)) ||
     ((n->opt_info & OPT_B_IS_NODE) &&
      find_and_clear_optimize_flag(n->u.s.b.node,a)))
  {
    n->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
    return 1;
  }
  return 0;
}

int cntargs(node *n)
{
  if(!n) return 0;
  switch(n->token)
  {
    case F_ASSIGN_AND_POP:
    case F_INC_AND_POP:
    case F_DEC_AND_POP:
    case '{':         return 0;

    case F_CAST:
    case F_EFUN_CALL: return n->type!=TYPE_VOID;

    case F_FOREACH:   return 3;
    case F_INC_NEQ_LOOP:
    case F_DEC_NEQ_LOOP:
    case F_INC_LOOP:
    case F_DEC_LOOP:  return 2;

    case ' ':
    case F_LVALUE_LIST:
    case F_ARG_LIST:
      return cntargs(n->u.s.a.node)+cntargs(n->u.s.b.node);

    /* this might not be true, but it doesn't matter very much */
    default: return 1;
  }
}

node *do_optimize(node *n,node *parent);
node *optimize(node *n,node *parent)
{
  int save_current_line=current_line;
  if(!n) return n;

#ifdef LINMALLOC_DEBUG
  assert_node(n);
#endif

  if(n->opt_info & OPT_OPTIMIZED) return n;
  current_line=n->lnum;
  n=do_optimize(n,parent);
  if(n) n->opt_info|=OPT_OPTIMIZED;
  current_line=save_current_line;
  return n;
}

node *do_optimize(node *n,node *parent)
{
  node *tmp1,*tmp2,*tmp3;

  if(!n) return NULL;
  
  if(n->opt_info & OPT_A_IS_NODE) n->u.s.a.node=optimize(n->u.s.a.node,n);
  if(n->opt_info & OPT_B_IS_NODE) n->u.s.b.node=optimize(n->u.s.b.node,n);
  check_types(n);
  compute_opt_info(n,0,1);	/* hmmm */

  if(lasdebug>2)
  {
    fprintf(stderr,"do_optimizing:");
    nice_print_tree(n);
  }

  switch(n->token)
  {
  case ' ':
  case ':': return n;  /* hands of optimizer ! */

  case '?':
    if(node_is_true(n->u.s.a.node))
    {
      tmp1=n->u.s.b.node->u.s.a.node;
      n->u.s.b.node->u.s.a.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(node_is_false(n->u.s.a.node))
    {
      tmp1=n->u.s.b.node->u.s.b.node;
      n->u.s.b.node->u.s.b.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(node_is_eq(n->u.s.b.node->u.s.a.node,n->u.s.b.node->u.s.b.node))
    {
      n->token=F_ARG_LIST;
      n->u.s.a.node=mkcastnode(TYPE_VOID,n->u.s.a.node);
      tmp1=n->u.s.b.node;
      n->u.s.b.node=tmp1->u.s.a.node;
      tmp1->u.s.a.node=NULL;
      free_node(tmp1);
      n->opt_info&=~OPT_OPT_COMPUTED;
      return optimize(n,parent);
    }
    break;

  case F_ARG_LIST:
  case '{':
      
    if(!n->u.s.a.node && !n->u.s.b.node)
    {
      free_node(n);
      return NULL;
    }

    if(!n->u.s.a.node)
    {
      tmp1=n->u.s.b.node;
      n->u.s.b.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(!n->u.s.b.node)
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if((n->u.s.a.node->token==F_RETURN ||
	n->u.s.a.node->token==F_BREAK ||
	n->u.s.a.node->token==F_CONTINUE ||
	((n->u.s.a.node->token=='{' || n->u.s.a.node->token==F_ARG_LIST) &&
	 n->u.s.a.node->u.s.b.node &&
	 (n->u.s.a.node->u.s.b.node->token==F_RETURN ||
	  n->u.s.a.node->u.s.b.node->token==F_BREAK ||
	  n->u.s.a.node->u.s.b.node->token==F_CONTINUE))) &&
       !(n->u.s.b.node->opt_info & OPT_CASE))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(n->token=='{')
    {
      node **prev,**curr;
      while(n->u.s.a.node->token==F_CAST && n->u.s.a.node->type==TYPE_VOID)
      {
	tmp1=n->u.s.a.node;
	n->u.s.a.node=tmp1->u.s.a.node;
	tmp1->u.s.a.node=0;
	free_node(tmp1);
      }
      while(n->u.s.b.node->token==F_CAST && n->u.s.b.node->type==TYPE_VOID)
      {
	tmp1=n->u.s.b.node;
	n->u.s.b.node=tmp1->u.s.a.node;
	tmp1->u.s.a.node=0;
	free_node(tmp1);
      }
      curr=&(n->u.s.b.node);
      prev=last_cmd(&(n->u.s.a.node));
      if(prev && (*prev) && (*curr))
      {
	if(((*curr)->token==F_INC || (*curr)->token==F_POST_INC ||
	    (*curr)->token==F_INC_AND_POP) &&
	   ((*prev)->token==F_ASSIGN ||
	    (*prev)->token==F_ASSIGN_AND_POP) &&
	   node_is_eq((*prev)->u.s.a.node,(*curr)->u.s.a.node))
	{
	  free_node(*curr);
	  *curr=0;
	  (*prev)->u.s.b.node=mknode(F_ADD,(*prev)->u.s.b.node,mkintnode(1));
	  find_and_clear_optimize_flag(n,*prev);
	  return optimize(n,parent);
	}

	if(((*curr)->token==F_DEC || (*curr)->token==F_POST_DEC ||
	    (*curr)->token==F_DEC_AND_POP) &&
	   ((*prev)->token==F_ASSIGN ||
	    (*prev)->token==F_ASSIGN_AND_POP) &&
	   node_is_eq((*prev)->u.s.a.node,(*curr)->u.s.a.node))
	{
	  free_node(*curr);
	  *curr=0;
	  (*prev)->u.s.b.node=mknode(F_ADD,(*prev)->u.s.b.node,mkintnode(-1));
	  find_and_clear_optimize_flag(n,*prev);
	  return optimize(n,parent);
	}
      }
      if(!(n->opt_info & (OPT_CONT | OPT_BREAK | OPT_CASE)) &&
	 cntargs(n->u.s.a.node)+cntargs(n->u.s.b.node)<10)
      {
	n->token=F_ARG_LIST;
	return optimize(mkcastnode(TYPE_VOID,n),parent);
      }
    }

    break;

  case F_MULTIPLY:
    if((n->u.s.a.node->type & TYPE_MOD_MASK)==(n->u.s.b.node->type & TYPE_MOD_MASK))
    {
      if((n->u.s.a.node->token==F_NUMBER && n->u.s.a.node->u.s.a.number==1) ||
	 (n->u.s.a.node->token==F_FLOAT && n->u.s.a.node->u.s.a.fnum==1.0) ||
	 (n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==0) ||
	 (n->u.s.b.node->token==F_FLOAT && n->u.s.b.node->u.s.a.fnum==0.0))
      {
	tmp1=n->u.s.b.node;
	n->u.s.b.node=0;
	free_node(n);
	return optimize(tmp1,parent);
      }
      if((n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==1) ||
	 (n->u.s.b.node->token==F_FLOAT && n->u.s.b.node->u.s.a.fnum==1.0) ||
	 (n->u.s.a.node->token==F_NUMBER && n->u.s.a.node->u.s.a.number==0) ||
	 (n->u.s.a.node->token==F_FLOAT && n->u.s.a.node->u.s.a.fnum==0.0))
      {
	tmp1=n->u.s.a.node;
	n->u.s.a.node=0;
	free_node(n);
	return optimize(tmp1,parent);
      }
    }
    break;

  case F_DIVIDE:
    if((n->u.s.a.node->type & TYPE_MOD_MASK)==(n->u.s.b.node->type & TYPE_MOD_MASK))
    {
      if((n->u.s.a.node->token==F_NUMBER && n->u.s.a.node->u.s.a.number==0) ||
	 (n->u.s.a.node->token==F_FLOAT && n->u.s.a.node->u.s.a.fnum==0.0) ||
	 (n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==1) ||
	 (n->u.s.b.node->token==F_FLOAT && n->u.s.b.node->u.s.a.fnum==1.0))
      {
	tmp1=n->u.s.a.node;
	n->u.s.a.node=0;
	free_node(n);
	return optimize(tmp1,parent);
      }
    }
    break;

  case F_ADD:
    if(((n->u.s.a.node->type & TYPE_MOD_MASK) == TYPE_NUMBER &&
	n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==0) ||
       ((n->u.s.a.node->type & TYPE_MOD_MASK) == TYPE_FLOAT &&
	n->u.s.b.node->token==F_FLOAT && n->u.s.b.node->u.s.a.fnum==0.0) ||
       ((n->u.s.a.node->type & TYPE_MOD_MASK) == TYPE_STRING &&
	n->u.s.b.node->token==F_STRING && !strlen(n->u.s.b.node->u.s.a.str)))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    if(((n->u.s.b.node->type & TYPE_MOD_MASK) == TYPE_NUMBER &&
	n->u.s.a.node->token==F_NUMBER && n->u.s.a.node->u.s.a.number==0) ||
       ((n->u.s.b.node->type & TYPE_MOD_MASK) == TYPE_FLOAT &&
	n->u.s.a.node->token==F_FLOAT && n->u.s.a.node->u.s.a.fnum==0.0) ||
       ((n->u.s.b.node->type & TYPE_MOD_MASK) == TYPE_STRING &&
	n->u.s.a.node->token==F_STRING && !strlen(n->u.s.a.node->u.s.a.str)))
    {
      tmp1=n->u.s.b.node;
      n->u.s.b.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    if(n->u.s.a.node->token==F_ADD)
    {
      n->u.s.a.node->token=F_ARG_LIST;
      n->u.s.a.node->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      n->token=F_ARG_LIST;
      n->opt_info&=~OPT_OPT_COMPUTED;
      return optimize(mkefunnode(F_SUM,n),parent);
    }
    if(n->u.s.a.node->token==F_EFUN_CALL &&
       n->u.s.a.node->u.s.a.number==F_SUM)
    {
      n->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      n->u.s.a.node->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      tmp1=n->u.s.a.node;
      n->token=F_ARG_LIST;
      n->u.s.a.node=tmp1->u.s.b.node;
      tmp1->u.s.b.node=n;
      return optimize(tmp1,parent);
    }
    break;

  case F_ASSIGN:
    if(node_is_eq(n->u.s.a.node,n->u.s.b.node))
    {
      free_node(n);
      return NULL;
    }

    if(!n->u.s.a.node) break;
    if(n->u.s.b.node->token==F_ADD &&
       n->u.s.a.node->type==TYPE_NUMBER &&
       node_is_eq(n->u.s.a.node,n->u.s.b.node->u.s.a.node) &&
       n->u.s.b.node->u.s.b.node->token==F_NUMBER &&
       n->u.s.b.node->u.s.b.node->u.s.a.number==1)
    {
      tmp1=mknode(F_INC,n->u.s.a.node,0);
      n->u.s.a.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(n->u.s.b.node->token==F_SUBTRACT &&
       n->u.s.a.node->type==TYPE_NUMBER &&
       node_is_eq(n->u.s.a.node,n->u.s.b.node->u.s.a.node) &&
       n->u.s.b.node->u.s.b.node->token==F_NUMBER &&
       n->u.s.b.node->u.s.b.node->u.s.a.number==1)
    {
      tmp1=mknode(F_DEC,n->u.s.a.node,0);
      n->u.s.a.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }
    break;

  case F_EFUN_CALL:
    switch(n->u.s.a.number)
    {
    case F_SIZEOF:
      if(n->u.s.b.node &&
	 n->u.s.b.node->token==F_EFUN_CALL)
      {
	switch(n->u.s.b.node->u.s.a.number)
	{
	case F_M_INDICES:
	case F_M_VALUES:
	case F_INDICES:
	  n->u.s.a.number=F_SIZE;
	  tmp1=n->u.s.b.node;
	  n->u.s.b.node=tmp1->u.s.b.node;
	  tmp1->u.s.b.node=NULL;
	  free_node(tmp1);
	  n->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
	  return optimize(n,parent);
	}
      }

    case F_GET_FUNCTION:
      if(n->u.s.b.node &&
	 n->u.s.b.node->token==F_ARG_LIST &&
	 n->u.s.b.node->u.s.a.node &&
	 n->u.s.b.node->u.s.b.node &&
	 n->u.s.b.node->u.s.a.node->token==F_EFUN_CALL &&
	 n->u.s.b.node->u.s.a.node->u.s.a.number==F_THIS_OBJECT &&
	 n->u.s.b.node->u.s.b.node->token==F_STRING)
      {
	int i;
	if((i=defined_function(n->u.s.b.node->u.s.b.node->u.s.a.str))!=-1)
	{
	  free_node(n);
	  return mkconstfunnode(i);
	}
      }

      if(n->u.s.b.node &&
	 n->u.s.b.node->token==F_ARG_LIST &&
	 n->u.s.b.node->u.s.b.node &&
	 n->u.s.b.node->u.s.b.node->token==F_STRING)

      {
	int e;
	for(e=0;e<LF_MAX;e++)
	{
	  if(!strcmp(lfuns[e],n->u.s.b.node->u.s.b.node->u.s.a.str))
	  {
	    n->u.s.a.number=F_GET_LFUN;
	    free_node(n->u.s.b.node->u.s.b.node);
	    n->u.s.b.node->u.s.b.node=mkintnode(e);
	    return optimize(n,parent);
	  }
	}
      }
      break;
	
    case F_SUM:
    {
      node **t;
      if((t=first_arg(n->u.s.b.node)) && (*t)->token==F_ADD)
      {
	(*t)->token=F_ARG_LIST;
	(*t)->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
	return optimize(n,parent);
      }
    }
      break;
    }
    break;

  case F_XOR:
    if(((n->u.s.a.node->type & TYPE_MOD_MASK) == TYPE_NUMBER &&
	n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==0))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    if(((n->u.s.a.node->type & TYPE_MOD_MASK) == TYPE_NUMBER &&
	n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==-1))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=0;
      free_node(n);
      return optimize(mknode(F_COMPL,tmp1,0),parent);
    }
    break;

  case F_AND:
    if((n->u.s.a.node->type== TYPE_NUMBER &&
	n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==-1) ||
       (n->u.s.b.node->type== TYPE_NUMBER &&
	n->u.s.a.node->token==F_NUMBER && n->u.s.a.node->u.s.a.number==0))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    if((n->u.s.b.node->type == TYPE_NUMBER &&
	n->u.s.a.node->token==F_NUMBER && n->u.s.a.node->u.s.a.number==-1) ||
       (n->u.s.a.node->type == TYPE_NUMBER &&
	n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==0))
    {
      tmp1=n->u.s.b.node;
      n->u.s.b.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    break;

  case F_LAND:
    /* !a and !b -> !(a or b) */
    if(n->u.s.a.node->token==F_NOT && n->u.s.b.node->token==F_NOT)
    {
      tmp1=mknode(F_NOT,mknode(F_LOR,n->u.s.a.node->u.s.a.node,
			       n->u.s.b.node->u.s.a.node),0);
      n->u.s.a.node->u.s.a.node=NULL;
      n->u.s.b.node->u.s.a.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }
    /* (a && b) && c -> a && (b && c) */
    while(n->u.s.a.node->token==F_LAND)
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=tmp1->u.s.a.node;
      tmp1->u.s.a.node=tmp1->u.s.b.node;
      tmp1->u.s.b.node=n->u.s.b.node;
      n->u.s.b.node=tmp1;
    }
    if(node_is_true(n->u.s.a.node))
    {
      tmp1=n->u.s.b.node;
      n->u.s.b.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(node_is_true(n->u.s.b.node))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(node_is_false(n->u.s.a.node))
    {
      free_node(n);
      return mkintnode(0);
    }
    break;

  case F_LOR:
    /* !a || !b -> !(a || b) */
    if(n->u.s.a.node->token==F_NOT && n->u.s.b.node->token==F_NOT)
    {
      tmp1=mknode(F_NOT,mknode(F_LAND,n->u.s.a.node->u.s.a.node,
			       n->u.s.b.node->u.s.a.node),0);
      n->u.s.a.node->u.s.a.node=NULL;
      n->u.s.b.node->u.s.a.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }
    /* (a && b) && c -> a && (b && c) */
    while(n->u.s.a.node->token==F_LOR)
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=tmp1->u.s.a.node;
      tmp1->u.s.a.node=tmp1->u.s.b.node;
      tmp1->u.s.b.node=n->u.s.b.node;
      n->u.s.b.node=tmp1;
    }
    if(node_is_false(n->u.s.a.node))
    {
      tmp1=n->u.s.b.node;
      n->u.s.b.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(node_is_false(n->u.s.b.node))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(node_is_true(n->u.s.a.node))
    {
      free_node(n);
      return mkintnode(1);
    }
    break;

  case F_OR:
    if((n->u.s.a.node->type == TYPE_NUMBER &&
	n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==-1) ||
       (n->u.s.b.node->type == TYPE_NUMBER &&
	n->u.s.a.node->token==F_NUMBER && n->u.s.a.node->u.s.a.number==0))
    {
      tmp1=n->u.s.b.node;
      n->u.s.b.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    if((n->u.s.b.node->type == TYPE_NUMBER &&
	n->u.s.a.node->token==F_NUMBER && n->u.s.a.node->u.s.a.number==-1) ||
       (n->u.s.a.node->type== TYPE_NUMBER &&
	n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==0))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    break;

  case F_CAST:
    if(n->type==TYPE_VOID)
    {
      if(!(n->u.s.a.node) || !(n->opt_info & (OPT_SIDE_EFFECT |
					      OPT_LOCAL_ASSIGNMENT |
					      OPT_GLOBAL_ASSIGNMENT |
					      OPT_RESTORE_OBJECT |
					      OPT_CASE |
					      OPT_RETURN |
					      OPT_CONT |
					      OPT_BREAK
					      )))
      {
	free_node(n);
	return NULL;
      }
      tmp1=n->u.s.a.node;
      switch(tmp1->token)
      {
      case F_ASSIGN:
	tmp1->token=F_ASSIGN_AND_POP;
	tmp1->type=TYPE_VOID;
	n->u.s.a.node=0;
	free_node(n);
	return optimize(tmp1,parent);

      case F_INC:
      case F_POST_INC:
	tmp1->token=F_INC_AND_POP;
	tmp1->type=TYPE_VOID;
	n->u.s.a.node=0;
	free_node(n);
	return optimize(tmp1,parent);

      case F_DEC:
      case F_POST_DEC:
	tmp1->token=F_DEC_AND_POP;
	tmp1->type=TYPE_VOID;
	n->u.s.a.node=0;
	free_node(n);
	return optimize(tmp1,parent);
      }
      /*
	 if(opt_to_assignments)
	 {
	 remember_all_variables();
	 eval(n->u.s.b.node);
	 compare_all_variables_and_make_assignments();
	 
	 }
	 */
     
    }

    if(!n->u.s.a.node)
    {
      switch(n->type)
      {
      case TYPE_NUMBER:
	free_node(n);
	return mkintnode(0);

      case TYPE_STRING:
	free_node(n);
	return mkstrnode(make_shared_string(""));

      }
    }
    if(n->type==(n->u.s.a.node->type & TYPE_MOD_MASK))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    if(n->type==TYPE_FUNCTION && n->u.s.a.node->token==F_STRING)
    {
      int i;
      if(-1!=(i=defined_function(n->u.s.a.node->u.s.a.str)))
      {
	free_node(n);
	return mkconstfunnode(i);
      }
    }
    break;

  case F_FOR:
  {
    node **last;
    int inc,less;
    if(node_is_false(n->u.s.a.node))
    {
      free_node(n);
      return NULL;
    }

    if(n->u.s.a.node &&
       (n->u.s.a.node->token==F_INC ||
	n->u.s.a.node->token==F_DEC ||
	n->u.s.a.node->token==F_POST_INC ||
	n->u.s.a.node->token==F_POST_DEC) &&
       (!n->u.s.b.node->u.s.a.node &&
	!(n->u.s.b.node->u.s.b.node->opt_info & OPT_CONT)))
    {
      if(n->u.s.a.node->token==F_DEC || n->u.s.a.node->token==F_POST_DEC)
	n->token=F_DEC_LOOP;
      else
	n->token=F_INC_LOOP;
      n->u.s.a.node->token=' ';
      n->u.s.a.node->u.s.b.node=n->u.s.a.node->u.s.a.node;
      n->u.s.a.node->u.s.a.node=mkintnode(0);
      n->u.s.a.node->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      n->u.s.b.node->token='{';
      n->u.s.b.node->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      n->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      return optimize(n,parent);
    }

    last=&(n->u.s.b.node->u.s.b.node);
    tmp1=last?*last:(node *)NULL;
    if(tmp1 && (
		tmp1->token==F_INC ||
		tmp1->token==F_POST_INC ||
		tmp1->token==F_INC_AND_POP ||
		tmp1->token==F_DEC_AND_POP ||
		tmp1->token==F_DEC ||
		tmp1->token==F_POST_DEC))
    {
      if(tmp1->token==F_INC || tmp1->token==F_POST_INC ||
	 tmp1->token==F_INC_AND_POP)
	inc=1;
      else
	inc=0;

      less=0;
      if(n->u.s.a.node->token!=F_GT &&
	 n->u.s.a.node->token!=F_GE &&
	 n->u.s.a.node->token!=F_LE &&
	 n->u.s.a.node->token!=F_NE &&
	 n->u.s.a.node->token!=F_LT)
	break;

      if(!node_is_eq(n->u.s.a.node->u.s.a.node,tmp1->u.s.a.node) ||
	 depend_p(n->u.s.a.node->u.s.b.node,n->u.s.b.node->u.s.a.node) ||
	 depend_p(n->u.s.a.node->u.s.b.node,n->u.s.a.node->u.s.a.node))
      {
	if(node_is_eq(n->u.s.a.node->u.s.b.node,tmp1->u.s.a.node) &&
	   !depend_p(n->u.s.a.node->u.s.a.node,n->u.s.b.node->u.s.a.node) &&
	   !depend_p(n->u.s.a.node->u.s.a.node,n->u.s.a.node->u.s.b.node))
	{
	  switch(n->u.s.a.node->token)
	  {
	  case F_LT: n->u.s.a.node->token=F_GT; break;
	  case F_LE: n->u.s.a.node->token=F_GT; break;
	  case F_GT: n->u.s.a.node->token=F_LE; break;
	  case F_GE: n->u.s.a.node->token=F_LE; break;
	  }
	  tmp2=n->u.s.a.node->u.s.a.node;
	  n->u.s.a.node->u.s.a.node=n->u.s.a.node->u.s.b.node;
	  n->u.s.a.node->u.s.b.node=tmp2;
	}else{
	  break;
	}
      }
      if(inc && n->u.s.a.node->token==F_LE)
	tmp3=mknode(F_ADD,n->u.s.a.node->u.s.b.node,mkintnode(1));
      else if(inc && n->u.s.a.node->token==F_LT)
	tmp3=n->u.s.a.node->u.s.b.node;
      else if(!inc && n->u.s.a.node->token==F_GE)
	tmp3=mknode(F_SUBTRACT,n->u.s.a.node->u.s.b.node,mkintnode(1));
      else if(!inc && n->u.s.a.node->token==F_GT)
	tmp3=n->u.s.a.node->u.s.b.node;
      else
	break;

      *last=0;
      if(n->u.s.a.node->token==F_NE)
      {
	if(inc)
	{
	  tmp2=mknode(F_INC_NEQ_LOOP,mknode(' ',tmp3,n->u.s.a.node->u.s.a.node),n->u.s.b.node->u.s.a.node);
	}else{
	  tmp2=mknode(F_DEC_NEQ_LOOP,mknode(' ',tmp3,n->u.s.a.node->u.s.a.node),n->u.s.b.node->u.s.a.node);
	}
      }else{
	if(inc)
	{
	  tmp2=mknode(F_INC_LOOP,mknode(' ',tmp3,n->u.s.a.node->u.s.a.node),n->u.s.b.node->u.s.a.node);
	}else{
	  tmp2=mknode(F_DEC_LOOP,mknode(' ',tmp3,n->u.s.a.node->u.s.a.node),n->u.s.b.node->u.s.a.node);
	}
      }
      n->u.s.a.node->u.s.a.node=0;
      n->u.s.a.node->u.s.b.node=0;
      n->u.s.b.node->u.s.a.node=0;
      n->u.s.b.node->u.s.b.node=0;
      free_node(n);
      if(inc)
      {
	tmp1->token=F_DEC;
      }else{
	tmp1->token=F_INC;
      }
      tmp1=mknode('{',mkcastnode(TYPE_VOID,tmp1),tmp2);
      return optimize(tmp1,parent);
    }
  }
    break;

  case F_ADD_EQ:
    if(n->u.s.a.node->type==TYPE_NUMBER &&
       n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==1)
    {
      tmp1=mknode(F_INC,n->u.s.a.node,0);
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(n->u.s.a.node->type==TYPE_NUMBER &&
       n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==-1)
    {
      tmp1=mknode(F_DEC,n->u.s.a.node,0);
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    break;

  case F_SUB_EQ:
    if(n->u.s.a.node->type==TYPE_NUMBER &&
       n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==1)
    {
      tmp1=mknode(F_DEC,n->u.s.a.node,0);
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }

    if(n->u.s.a.node->type==TYPE_NUMBER &&
       n->u.s.b.node->token==F_NUMBER && n->u.s.b.node->u.s.a.number==-1)
    {
      tmp1=mknode(F_ADD,n->u.s.a.node,0);
      n->u.s.a.node=0;
      free_node(n);
      return optimize(tmp1,parent);
    }
    break;

  case F_INC_NEQ_LOOP:
  case F_DEC_NEQ_LOOP:
  case F_INC_LOOP:
  case F_DEC_LOOP:
    if(!n->u.s.b.node)
    {
      n->u.s.a.node->token=F_ASSIGN;
      n->token=F_CAST;
      n->type=TYPE_VOID;
      n->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      n->u.s.a.node->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      n->u.s.a.node->type=n->u.s.a.node->u.s.b.node->type;
      tmp1=n->u.s.a.node->u.s.a.node;
      n->u.s.a.node->u.s.a.node=n->u.s.a.node->u.s.b.node;
      n->u.s.a.node->u.s.b.node=tmp1;
      return optimize(n,parent);
    }
    if(!((n->u.s.a.node->opt_info | n->u.s.b.node->opt_info)&
	 (OPT_SIDE_EFFECT | OPT_RETURN | OPT_CONT | OPT_BREAK)) &&
       !depend_p(n->u.s.a.node,n->u.s.b.node) &&
       !depend_p(n->u.s.b.node,n->u.s.b.node) &&
       !depend_p(n->u.s.b.node,n->u.s.a.node))
    {
      n->u.s.a.node->token=F_ASSIGN;
      n->token='{';
      n->type=TYPE_VOID;
      n->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      n->u.s.a.node->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      n->u.s.a.node->type=n->u.s.a.node->u.s.b.node->type;
      tmp1=n->u.s.a.node->u.s.a.node;
      n->u.s.a.node->u.s.a.node=n->u.s.a.node->u.s.b.node;
      n->u.s.a.node->u.s.b.node=tmp1;
      return optimize(n,parent);
    }
  break;

  case F_SWITCH:
  {
    node **tmp;
    while((tmp=last_cmd(&(n->u.s.b.node))) && (*tmp)->token==F_BREAK)
    {
      free_node(*tmp);
      *tmp=0;
    }
    if(!n->u.s.b.node)
    {
      n->token=F_CAST;
      n->type=TYPE_VOID;
      n->opt_info&=~(OPT_OPT_COMPUTED|OPT_OPTIMIZED);
      return optimize(n,parent);
    }
  }
  case F_DO:
  {
    if(node_is_true(n->u.s.b.node))
    {
      n->token=F_BRANCH;
      free_node(n->u.s.b.node);
      n->u.s.b.node=NULL;
    }
    if(node_is_false(n->u.s.b.node))
    {
      tmp1=n->u.s.a.node;
      n->u.s.a.node=NULL;
      free_node(n);
      return optimize(tmp1,parent);
    }
  }
}

  if(is_const(n) && (n->opt_info & OPT_TRY_OPTIMIZE) &&
     (!parent || !is_const(parent)) &&
     !num_parse_error)
  {
    tmp1=eval(n);
    free_node(n);
    return tmp1;
  }
#ifdef LINMALLOC_DEBUG
  assert_node(n);
#endif
  return n;
}


/* these will be used for inlining */
node **locals[MAX_LOCAL],**globals[256];

/* depend_optim: nullify a node at a time and see anything depends on it
 * beeing there
 */
int depend_optim(node **a,node *b,int chronologie,int voided,node *tree2)
{
  node *n,*n2;
  if(!(n=*a)) return 0;
#ifdef LINMALLOC_DEBUG
  assert_node(*a);
#endif
  if(!(n->opt_info & (OPT_SIDE_EFFECT | OPT_CASE | OPT_CONT | OPT_BREAK |
	 OPT_GLOBAL_ASSIGNMENT | OPT_RETURN)))
  {
    *a=0;
    if(!depend_p(b,n) && !depend_p(tree2,n))
    {
      free_node(n);
      return 1;
    }
    *a=n;
  }
  switch(n->token)
  {
  case F_DO:
  {
    int i;
    tree2=mknode('{',n->u.s.b.node,tree2);
    i=depend_optim(&(n->u.s.a.node),b,0,0,tree2);
    n2=tree2->u.s.a.node;
    tree2->u.s.b.node=tree2->u.s.a.node=NULL;
    free_node(tree2);
    tree2=n2;
    if(i)
    {
      n->opt_info&=~OPT_OPTIMIZED;
      return 1;
    }
    return 0;
  }

  case F_INC_NEQ_LOOP:
  case F_DEC_NEQ_LOOP:
  case F_INC_LOOP:
  case F_DEC_LOOP:
  case F_FOREACH:
  case F_FOR:
  {
    int i;
    tree2=mknode('{',n->u.s.a.node,tree2);
    i=depend_optim(&(n->u.s.b.node),b,0,0,tree2);
    n2=tree2->u.s.a.node;
    tree2->u.s.b.node=tree2->u.s.a.node=NULL;
    free_node(tree2);
    tree2=n2;
    if(i)
    {
      n->opt_info&=~OPT_OPTIMIZED;
      return 1;
    }
    return 0;
  }

  case F_CAST:
    if(n->type==TYPE_VOID)
    {
      if(depend_optim(&(n->u.s.a.node),b,chronologie,1,tree2))
      {
	n->opt_info&=~OPT_OPTIMIZED;
	return 1;
      }else{
	return 0;
      }
    }
    return 0;

  case F_ARG_LIST:
  case '{':
    if(n->token == F_ARG_LIST && !voided) return 0;
    if(chronologie)
    {
      int i;
      /* if chronological dependency is wanted, nullify and ignore
       * things that happened before this node while testing
       */
      while(depend_optim(&(n->u.s.a.node),b,chronologie,1,tree2))
	n->opt_info&=~OPT_OPTIMIZED;

      n2=n->u.s.a.node;
      n->u.s.a.node=0;
      i=depend_optim(&(n->u.s.b.node),b,chronologie,1,tree2);
      if(i) n->opt_info&=~OPT_OPTIMIZED;
      n->u.s.a.node=n2;
      return i;
    }else{
      int i;

      tree2=mknode('{',n->u.s.b.node,tree2);
      n->u.s.b.node=NULL;
      i=depend_optim(&(n->u.s.a.node),b,chronologie,1,tree2);
      n->u.s.b.node=tree2->u.s.a.node;
      n2=tree2;
      tree2=tree2->u.s.b.node;
      n2->u.s.a.node=n2->u.s.b.node=NULL;
      free_node(n2);

      tree2=mknode('{',tree2,n->u.s.a.node);
      n->u.s.a.node=NULL;
      i|=depend_optim(&(n->u.s.b.node),b,chronologie,1,tree2);
      n->u.s.a.node=tree2->u.s.b.node;
      n2=tree2;
      tree2=tree2->u.s.a.node;
      n2->u.s.b.node=n2->u.s.a.node=NULL;
      free_node(n2);
      if(i) n->opt_info&=~OPT_OPTIMIZED;
      return i;
    }
  }
  return 0;
}

/* we need to dome computations to see if the arguments are pointer arguments
 * if so, let the function depend on those, so we won't remove code that
 * changes shared arrays.
 */
void do_depend_optim(node **n,int args)
{
  node *tree2;
  int e;
  tree2=NULL;
  for(e=0;e<args;e++)
  {
    if(type_of_locals[e]==TYPE_MAPPING ||
       type_of_locals[e]&TYPE_MOD_POINTER)
    {
      tree2=mknode(F_ARG_LIST,mklocalnode(e),tree2);
    }
  }
  tree2=mknode(F_RETURN,mkefunnode(F_AGGREGATE,tree2),0);
  compute_opt_info(tree2,0,0);
  while(depend_optim(n,*n,1,0,tree2));
  free_node(tree2);
}


void do_pop(int nr)
{
  if(!nr) return;
  if(nr==1)
  {
    ins_f_byte(F_POP_VALUE);
  }else if(nr<256){
    ins_f_byte(F_POP_N_ELEMS);
    ins_byte(nr);
  }else{
    ins_f_byte(F_POP_N_ELEMS);
    ins_byte(255);
    do_pop(nr-255);
  }
}



#define JUMP_CONDITIONAL 2
#define JUMP_SHORT 4
#define JUMP_UNSET 8

struct jump
{
  unsigned short flags;
  short token;
  int whereto;
  int wherefrom;
  int address;
};

int jump_ptr, max_jumps;
struct jump jumps[256];

void set_branch(int address,int whereto)
{
  int j,e;

  j=jump_ptr;
  if(j<0 || (jumps[j].address!=address || !(jumps[j].flags & JUMP_UNSET)))
  {
    for(j=max_jumps;j>=0;j--)
      if(jumps[j].address==address && (jumps[j].flags & JUMP_UNSET))
	break;
  }

  for(e=0;e<=max_jumps;e++)
  {
    if(jumps[e].flags & ( JUMP_UNSET)) continue;
    if(jumps[e].flags & JUMP_CONDITIONAL) continue;
    if(jumps[e].wherefrom!=whereto) continue;
#if defined(LASDEBUG)
    if(lasdebug>1) printf("Optimized Jump to a jump\n");
#endif
    whereto=jumps[e].whereto;
    break;
  }
  
  if(j>=0)
  {
    if(!(jumps[j].flags & JUMP_CONDITIONAL))
    {
      for(e=0;e<=max_jumps;e++)
      {
	if(jumps[e].flags & (JUMP_UNSET|JUMP_SHORT)) continue;
	if(jumps[e].whereto==jumps[j].wherefrom)
	{
	  upd_branch(jumps[e].address,whereto);
#if defined(LASDEBUG)
	  if(lasdebug>1) printf("Optimized Jump to a jump\n");
#endif
	}
      }
    }
    jumps[j].whereto=whereto;
    jumps[j].flags&=~JUMP_UNSET;
  }
  upd_branch(address,whereto);
}

void push_break_stack()
{
  push_explicit((int)break_stack);
  push_explicit(current_break);
  push_explicit(break_stack_size);
  break_stack_size=10;
  break_stack=(int *)xalloc(sizeof(int)*break_stack_size);
  current_break=0;
}

int do_jump(int token,int whereto)
{
  jump_ptr=(jump_ptr+1)%NELEM(jumps);
  if(jump_ptr>max_jumps) max_jumps=jump_ptr;

  jumps[jump_ptr].flags=JUMP_UNSET;
  jumps[jump_ptr].whereto=-1;
  jumps[jump_ptr].token=token;

  if(token!=F_BRANCH && token!=F_SHORT_BRANCH)
    jumps[jump_ptr].flags|=JUMP_CONDITIONAL;

  if(token==F_SHORT_BRANCH ||
     token==F_SHORT_BRANCH_WHEN_ZERO ||
     token==F_SHORT_BRANCH_WHEN_NON_ZERO) /* not used (yet) */
    jumps[jump_ptr].flags|=JUMP_SHORT;

  jumps[jump_ptr].wherefrom=PC;
  ins_f_byte(token);
  jumps[jump_ptr].address=PC;
  ins_short(0);
  if(whereto!=-1) set_branch(jumps[jump_ptr].address,whereto);
  return jumps[jump_ptr].address;
}

void clean_jumptable() { max_jumps=jump_ptr=-1; }

void pop_break_stack(int jump)
{
  for(current_break--;current_break>=0;current_break--)
    set_branch(break_stack[current_break],jump);

  free((char *)break_stack);
  break_stack_size=pop_address();
  current_break=pop_address();
  break_stack=(int *)pop_address();
}

void push_continue_stack()
{
  push_explicit((int)continue_stack);
  push_explicit(current_continue);
  push_explicit(continue_stack_size);
  continue_stack_size=10;
  continue_stack=(int *)xalloc(sizeof(int)*continue_stack_size);
  current_continue=0;
}

void pop_continue_stack(int jump)
{
  for(current_continue--;current_continue>=0;current_continue--)
    set_branch(continue_stack[current_continue],jump);

  free((char *)continue_stack);
  continue_stack_size=pop_address();
  current_continue=pop_address();
  continue_stack=(int *)pop_address();
}

int do_docode2(node *n,int needlval);

int do_docode(node *n,int needlval)
{
  int i;
  int save_current_line=current_line;
  if(!n) return 0;
  current_line=n->lnum;
  i=do_docode2(n,needlval);
  current_line=save_current_line;
  return i;
}

int docode(node *n)
{
  clean_jumptable();
  return do_docode(n,0);
}

int do_jump_when_non_zero(node *n,int j)
{
  if(!(n->opt_info & OPT_SIDE_EFFECT))
  {
    if(node_is_true(n))
      return do_jump(F_BRANCH,j);

    if(node_is_false(n))
      return -1;
  }
  if(do_docode(n,0)!=1)
    fatal("Infernal compiler skiterror.\n");
  return do_jump(F_BRANCH_WHEN_NON_ZERO,j);
}

int do_jump_when_zero(node *n,int j)
{
  if(!(n->opt_info & OPT_SIDE_EFFECT))
  {
    if(node_is_true(n))
      return -1;

    if(node_is_false(n))
      return do_jump(F_BRANCH,j);
  }
  if(do_docode(n,0)!=1)
    fatal("Infernal compiler skiterror.\n");
  return do_jump(F_BRANCH_WHEN_ZERO,j);
}

int do_docode2(node *n,int needlval)
{
  int tmp1,tmp2,tmp3;

  if(!n) return 0;

  store_line_number_info(n->lnum);
  if(needlval & 1)
  {
    switch(n->token)
    {
    default:
      yyerror("Illegal lvalue.");
      ins_int(0);
      return 1;

    case F_LVALUE_LIST:
    case F_LOCAL:
    case F_GLOBAL:
    case F_INDEX:
    case F_ARG_LIST:
    }
  }
  switch(n->token)
  {
  case F_PUSH_ARRAY:
    tmp1=do_docode(n->u.s.a.node,0);
    if(tmp1!=1)
    {
      fatal("Internal compiler error, Yikes!\n");
    }
    ins_f_byte(F_PUSH_ARRAY);
    return -0x7ffffff;

  case F_STRING:
    tmp1=store_prog_string(n->u.s.a.str);
    if(tmp1<256)
    {
      ins_f_byte(F_SHORT_STRING);
      ins_byte(tmp1);
    }else{
      ins_f_byte(F_STRING);
      ins_short(tmp1);
    }
    return 1;

  case F_NUMBER:
    ins_int(n->u.s.a.number);
    return 1;

  case F_FLOAT:
    ins_float(n->u.s.a.fnum);
    return 1;

  case '?':
  {
    int tmp5=0;			/* make gcc happy */
    if(do_docode(n->u.s.a.node,0)!=1)
      fatal("Infernal compiler error, please consult a shrink.\n");
    tmp1=do_jump(F_BRANCH_WHEN_ZERO,-1);
    tmp3=do_docode(n->u.s.b.node->u.s.a.node,0);
    if(tmp3)
    {
      ins_f_byte(F_POP_N_ELEMS);
      tmp5=PC;
      ins_byte(0);
    }
    if(n->u.s.b.node->u.s.b.node)
    {
      int tmp4;
      tmp2=do_jump(F_BRANCH,-1);
      set_branch(tmp1,PC);
      tmp4=do_docode(n->u.s.b.node->u.s.b.node,0);
      if(tmp4>tmp3)
      {
	do_pop(tmp4-tmp3);
      }else if(tmp4<tmp3){
	mem_block[A_PROGRAM].block[tmp5]=tmp3-tmp4;
	tmp3=tmp4;
      }
      tmp1=tmp2;
    }else{
      if(tmp3)
      {
	mem_block[A_PROGRAM].block[tmp5]=tmp3;
	tmp3=0;
      }
    }
    set_branch(tmp1,PC);
    return tmp3;
  }
      
  case F_AND_EQ:
  case F_OR_EQ:
  case F_XOR_EQ:
  case F_LSH_EQ:
  case F_RSH_EQ:
  case F_ADD_EQ:
  case F_SUB_EQ:
  case F_MULT_EQ:
  case F_MOD_EQ:
  case F_DIV_EQ:
    tmp1=do_docode(n->u.s.a.node,1);
    if(do_docode(n->u.s.b.node,0)!=1)
      fatal("Internal compiler error, I am afraid I have to dump core now.\n");
    ins_f_byte(F_PUSH_LTOSVAL);
    switch(n->token)
    {
    case F_AND_EQ: ins_f_byte(F_AND); break;
    case F_OR_EQ:  ins_f_byte(F_OR);  break;
    case F_XOR_EQ: ins_f_byte(F_XOR); break;
    case F_LSH_EQ: ins_f_byte(F_LSH); break;
    case F_RSH_EQ: ins_f_byte(F_RSH); break;
    case F_ADD_EQ: ins_f_byte(F_ADD); break;
    case F_SUB_EQ: ins_f_byte(F_SUBTRACT); break;
    case F_MULT_EQ:ins_f_byte(F_MULTIPLY);break;
    case F_MOD_EQ: ins_f_byte(F_MOD); break;
    case F_DIV_EQ: ins_f_byte(F_DIVIDE); break;
    }
    ins_f_byte(F_ASSIGN);
    return tmp1;

  case F_ASSIGN:
  case F_ASSIGN_AND_POP:
    tmp1=do_docode(n->u.s.a.node,1);
    switch(n->u.s.b.node->token)
    {
    case F_AND:
    case F_OR:
    case F_XOR:
    case F_LSH:
    case F_RSH:
    case F_ADD:
    case F_MOD:
    case F_SUBTRACT:
    case F_DIVIDE:
    case F_MULTIPLY:
      if(node_is_eq(n->u.s.a.node,n->u.s.b.node->u.s.a.node))
      {
	if(do_docode(n->u.s.b.node->u.s.b.node,0)!=1)
	  fatal("Infernal compiler error (dumpar core |ver hela mattan).\n");
	ins_f_byte(F_PUSH_LTOSVAL);
	ins_f_byte(n->u.s.b.node->token);
	ins_f_byte(n->token);
	return n->token==F_ASSIGN;
      }
    }
    if(do_docode(n->u.s.b.node,0)!=1)
      fatal("Infernal compiler error *sigh*.\n");
    ins_f_byte(n->token);
    return n->token==F_ASSIGN;

  case F_LAND:
  case F_LOR:
    if(do_docode(n->u.s.a.node,0)!=1)
      fatal("Compiler internal error.\n");
    tmp1=do_jump(n->token,-1);
    if(do_docode(n->u.s.b.node,0)!=1)
      fatal("Compiler internal error.\n");
    upd_branch(tmp1,PC);
    return 1;

  case F_EQ:
  case F_NE:
  case F_LT:
  case F_LE:
  case F_GT:
  case F_GE:
  case F_ADD:
  case F_SUBTRACT:
  case F_MULTIPLY:
  case F_DIVIDE:
  case F_MOD:
  case F_LSH:
  case F_RSH:
  case F_XOR:
  case F_OR:
  case F_AND:
    tmp1=do_docode(n->u.s.a.node,0);
    if(do_docode(n->u.s.b.node,0)!=1)
      fatal("Compiler internal error.\n");
    ins_f_byte(n->token);
    return tmp1;

  case F_RANGE:
    tmp1=do_docode(n->u.s.a.node,0);
    if(do_docode(n->u.s.b.node,0)!=2)
      fatal("Compiler internal error.\n");
    ins_f_byte(n->token);
    return tmp1;

  case F_INC:
  case F_DEC:
  case F_POST_INC:
  case F_POST_DEC:
    tmp1=do_docode(n->u.s.a.node,1);
    ins_f_byte(n->token);
    return tmp1;

  case F_INC_AND_POP:
  case F_DEC_AND_POP:
    tmp1=do_docode(n->u.s.a.node,1);
    ins_f_byte(n->token);
    return tmp1-1;

  case F_NOT:
  case F_COMPL:
  case F_NEGATE:
    tmp1=do_docode(n->u.s.a.node,0);
    ins_f_byte(n->token);
    return tmp1;

  case F_FOR:
  {
    struct vector *tmp_store;
    tmp_store=current_switch_mapping;
    current_switch_mapping=(struct vector *)NULL;
    push_break_stack();
    push_continue_stack();
    tmp1=do_jump(F_BRANCH,-1);
    tmp2=PC;
    if(n->u.s.b.node)
      do_pop(do_docode(n->u.s.b.node->u.s.a.node,0));
    pop_continue_stack(PC);
    if(n->u.s.b.node)
      do_pop(do_docode(n->u.s.b.node->u.s.b.node,0));
    set_branch(tmp1,PC);
    do_jump_when_non_zero(n->u.s.a.node,tmp2);
    pop_break_stack(PC);
    current_switch_mapping=tmp_store;
    return 0;
  }

  case ' ':
    return do_docode(n->u.s.a.node,0)+do_docode(n->u.s.b.node,1);

  case F_FOREACH:
  {
    struct vector *tmp_store;
    tmp_store=current_switch_mapping;
    current_switch_mapping=(struct vector *)NULL;
    tmp2=do_docode(n->u.s.a.node,0);
    ins_f_byte(F_CONST0);
    tmp3=do_jump(F_BRANCH,-1);
    tmp1=PC;
    push_break_stack();
    push_continue_stack();
    do_pop(do_docode(n->u.s.b.node,0));
    pop_continue_stack(PC);
    set_branch(tmp3,PC);
    do_jump(n->token,tmp1);
    pop_break_stack(PC);
    current_switch_mapping=tmp_store;
    return tmp2+1;
  }

  case F_INC_NEQ_LOOP:
  case F_DEC_NEQ_LOOP:
  case F_INC_LOOP:
  case F_DEC_LOOP:
  {
    struct vector *tmp_store;
    tmp_store=current_switch_mapping;
    current_switch_mapping=(struct vector *)NULL;
    tmp2=do_docode(n->u.s.a.node,0);
    tmp3=do_jump(F_BRANCH,-1);
    tmp1=PC;
    push_break_stack();
    push_continue_stack();
    do_pop(do_docode(n->u.s.b.node,0));
    pop_continue_stack(PC);
    set_branch(tmp3,PC);
    do_jump(n->token,tmp1);
    pop_break_stack(PC);
    current_switch_mapping=tmp_store;
    return tmp2;
  }

  case F_DO:
  {
    struct vector *tmp_store;
    tmp_store=current_switch_mapping;
    current_switch_mapping=(struct vector *)NULL;
    tmp2=PC;
    push_break_stack();
    push_continue_stack();
    do_pop(do_docode(n->u.s.a.node,0));
    pop_continue_stack(PC);
    do_jump_when_non_zero(n->u.s.b.node,tmp2);
    pop_break_stack(PC);
    current_switch_mapping=tmp_store;
    return 0;
  }

  case F_BRANCH:
  {
    struct vector *tmp_store;
    tmp_store=current_switch_mapping;
    current_switch_mapping=(struct vector *)NULL;
    tmp2=PC;
    push_break_stack();
    push_continue_stack();
    do_pop(do_docode(n->u.s.a.node,0));
    pop_continue_stack(PC);
    do_jump(F_BRANCH,tmp2);
    pop_break_stack(PC);
    current_switch_mapping=tmp_store;
    return 0;
  }

  case F_CAST:
    tmp1=do_docode(n->u.s.a.node,0);
    if(n->type==TYPE_VOID)
    {
      do_pop(tmp1);
      return 0;
    }
    if(!tmp1) { ins_f_byte(F_CONST0); tmp1=1; }
    switch(n->type)
    {
    case TYPE_STRING:
      ins_f_byte(F_CAST_TO_STRING);
      break;

    case TYPE_NUMBER:
      ins_f_byte(F_CAST_TO_INT);
      break;

    case TYPE_FLOAT:
      ins_f_byte(F_CAST_TO_FLOAT);
      break;

    case TYPE_OBJECT:
      ins_f_byte(F_CAST_TO_OBJECT);
      break;

    case TYPE_FUNCTION:
      ins_f_byte(F_CAST_TO_FUNCTION);
      break;
    }
    return tmp1;

  case F_EFUN_CALL:
    if(instrs[n->u.s.a.number-F_OFFSET].min_arg!=instrs[n->u.s.a.number-F_OFFSET].max_arg)
    {
      ins_f_byte(F_MARK);
      do_docode(n->u.s.b.node,0);
      ins_f_byte(n->u.s.a.number);
    }else{
      tmp1=do_docode(n->u.s.b.node,0);
      if(tmp1>256)
        yyerror("To many arguments to function call.");

      if(tmp1<0)
        yyerror("Can't use '@' with fixed argument efun.\n");

      if(tmp1!=instrs[n->u.s.a.number-F_OFFSET].min_arg)
        fatal("GAAH, INTERNAL COMPILER ERROR.\n",0,0,0,0,0,0,0,0);
      ins_f_byte(n->u.s.a.number);
    }
    if((n->type & TYPE_MOD_MASK)==TYPE_VOID)
      return 0;
    else
      return 1;

  case '{':
    do_pop(do_docode(n->u.s.a.node,0));
    do_pop(do_docode(n->u.s.b.node,0));
    return 0;

  case F_ARG_LIST:
    return do_docode(n->u.s.a.node,0)+
      do_docode(n->u.s.b.node,needlval);

  case F_SWITCH:
  {
    int e,addr=0;
    struct vector *v;
    struct vector *tmp_store=current_switch_mapping;
    if(do_docode(n->u.s.a.node,0)!=1)
      fatal("Internal compiler error, time to panic\n");        
    current_switch_mapping=allocate_mapping(
					    allocate_array(0),allocate_array(0));
    ins_f_byte(F_SWITCH);
    tmp1=PC;
    ins_short(0);
    push_break_stack();
    do_pop(do_docode(n->u.s.b.node,0));

    v=current_switch_mapping->item[1].u.vec;
    for(e=0;e<v->size;e++)
    {
      if(v->item[e].subtype==2 && !addr)
      {
	char *tmp;
	init_buf();
	my_strcat("Case range error at ");
	save_svalue(current_switch_mapping->item[0].u.vec->item+e);
/*	save_svalue(current_switch_mapping->item); */
	tmp=simple_free_buf();
	yyerror(tmp);
	free(tmp);
      }
      if(v->item[e].subtype==1)
      {
	if(addr)
	{
	  char *tmp;
	  init_buf();
	  my_strcat("Case range overlap at ");
	  save_svalue(current_switch_mapping->item[0].u.vec->item+e);
	  tmp=simple_free_buf();
	  yyerror(tmp);
	  free(tmp);
	}
	addr=v->item[e].u.number;
      }
      if(v->item[e].subtype==2)
      {
	v->item[e].subtype=addr;
	addr=0;
      }else{
	v->item[e].subtype=addr;
      }
    }
        
    if(!current_switch_mapping->item[2].u.number)
    {
      current_switch_mapping->item[2].u.number=
	PC;
    }
    pop_break_stack(PC);
    upd_short(tmp1,mem_block[A_SWITCH_MAPPINGS].current_size /
	      sizeof(struct vector *));
    add_to_mem_block(A_SWITCH_MAPPINGS, (char *)&current_switch_mapping,
		     sizeof(struct vector *));
    current_switch_mapping=tmp_store;
  }
    return 0;

  case F_CASE:
  {
    extern struct svalue *sp;
    struct svalue bar;
    if(!current_switch_mapping)
    {
      yyerror("Case outside switch.");
    }else{
      if(!is_const(n->u.s.a.node))
	yyerror("Case label isn't constant.");

      tmp1=eval_low(n->u.s.a.node);
      if(tmp1<1)
      {
	yyerror("Error in case label.\n");
	return 0;
      }

      if(!n->u.s.b.node)
      {
	if(assoc(sp,current_switch_mapping)>=0)
	{
	  yyerror("Duplicate case");
	}else{
	  bar.type=T_NUMBER;
	  bar.subtype=0;
	  bar.u.number=PC;
	  insert_alist(sp,&bar,current_switch_mapping);
	}
      }else{
	if(assoc(sp,current_switch_mapping)>=0)
	{
	  yyerror("Duplicate case");
	}else{
	  bar.type=T_NUMBER;
	  bar.subtype=1;
	  bar.u.number=PC;
	  insert_alist(sp,&bar,current_switch_mapping);
	}
	if(!is_const(n->u.s.b.node))
	  yyerror("Case label isn't constant.");

	tmp1+=tmp2=eval_low(n->u.s.b.node);
	if(tmp2<1)
	{
	  pop_n_elems(tmp1);
	  yyerror("Error in case label.\n");
	  return 0;
	}

	if(sp->type!=sp[-1].type)
	  yyerror("Mixed types in case range");

	if(assoc(sp,current_switch_mapping)>=0)
	{
	  yyerror("Duplicate case");
	}else{
	  bar.subtype=2;
	  insert_alist(sp,&bar,current_switch_mapping);
	}
      }
      pop_n_elems(tmp1);
    }
  }
    return 0;

  case F_DEFAULT:
  {
    if(!current_switch_mapping)
    {
      yyerror("Default outside switch.");
    }else{
      if(current_switch_mapping->item[2].u.number)
      {
	yyerror("Dual Default in switch.");
      }
      current_switch_mapping->item[2].u.number=
	PC;
    }
  }
    return 0;

  case F_BREAK:
    if(!break_stack)
    {
      yyerror("Break outside loop or switch.");
    }else{
      if(current_break>=break_stack_size)
      {
	break_stack_size*=2;
	break_stack=(int *)realloc((char *)break_stack,
				   sizeof(int)*break_stack_size);
      }
      break_stack[current_break++]=do_jump(F_BRANCH,-1);
    }
    return 0;

  case F_CONTINUE:
  {
    if(!continue_stack)
    {
      yyerror("continue outside loop or switch.");
    }else{
      if(current_continue>=continue_stack_size)
      {
	continue_stack_size*=2;
	continue_stack=(int *)realloc((char *)continue_stack,
				      sizeof(int)*continue_stack_size);
      }
      continue_stack[current_continue++]=do_jump(F_BRANCH,-1);
    }
  }
    return 0;

  case F_RETURN:
    if(n->u.s.a.node->token==F_EFUN_CALL &&
       n->u.s.a.node->u.s.a.number==F_CALL_FUNCTION)
    {
      ins_f_byte(F_MARK);
      do_docode(n->u.s.a.node->u.s.b.node,0);
      ins_f_byte(F_TAILRECURSE);
    }else{
      if(do_docode(n->u.s.a.node,0)!=1)
	ins_f_byte(F_CONST0);
      ins_f_byte(F_RETURN);
    }
    return 0;

  case F_SSCANF:
    tmp1=do_docode(n->u.s.a.node,0);
    tmp2=do_docode(n->u.s.b.node,0);
    ins_f_byte(F_SSCANF);
    ins_byte(tmp2+2);
    return tmp1-1;

  case F_SWAP_VARIABLES:
    tmp1=do_docode(n->u.s.a.node,0);
    if(tmp1!=2)
      fatal("Infernal compiler error (skit oxo)\n");        
    ins_f_byte(F_SWAP_VARIABLES);
    return 0;

  case F_CATCH:
  {
    struct vector *tmp_store;
    tmp_store=current_switch_mapping;
    current_switch_mapping=(struct vector *)NULL;
    tmp1=do_jump(F_CATCH,-1);
    push_break_stack();
    push_continue_stack();
    do_pop(do_docode(n->u.s.a.node,0));
    pop_continue_stack(PC);
    pop_break_stack(PC);
    ins_f_byte(F_DUMB_RETURN);
    set_branch(tmp1,PC);
    current_switch_mapping=tmp_store;
  }
    return 1;

  case F_GAUGE:
  {
    struct vector *tmp_store;
    tmp_store=current_switch_mapping;
    current_switch_mapping=(struct vector *)NULL;
    ins_f_byte(F_PUSH_COST);
    do_pop(do_docode(n->u.s.a.node,0));
    ins_f_byte(F_PUSH_COST);
    ins_f_byte(F_SUBTRACT);
    ins_f_byte(F_NEGATE);
    current_switch_mapping=tmp_store;
    return 1;
  }

  case F_LVALUE_LIST:
    return do_docode(n->u.s.a.node,1)+do_docode(n->u.s.b.node,1);

  case F_INDEX:
    tmp1=do_docode(n->u.s.a.node,needlval | 2);
    if(do_docode(n->u.s.b.node,0)!=1)
      fatal("Internal compiler error, please report this.");
    while(n && n->token==F_INDEX) n=n->u.s.a.node;

    if(needlval & 1)
    {
      ins_f_byte(F_PUSH_INDEXED_LVALUE);
    }else{
      ins_f_byte(F_INDEX);
      if(n->token==F_CONSTANT) ins_f_byte(F_COPY_VALUE);
    }
    return tmp1;

  case F_CONSTANT:
    tmp1=store_constant(&(n->u.sval));
    if(tmp1>255)
    {
      ins_f_byte(F_CONSTANT);
      ins_short(tmp1);
    }else{
      ins_f_byte(F_SHORT_CONSTANT);
      ins_byte(tmp1);
    }
    if(needlval!=2) /* copy later */
      if(n->u.sval.type & (T_POINTER | T_LIST | T_MAPPING | T_ALIST_PART))
	ins_f_byte(F_COPY_VALUE);
    return 1;

  case F_LOCAL:
    if(needlval & 1)
    {
      ins_f_byte(F_PUSH_LOCAL_VARIABLE_LVALUE);
      ins_byte(n->u.s.a.number);
    }else{
      ins_f_byte(F_LOCAL_NAME);
      ins_byte(n->u.s.a.number);
    }
    return 1;

  case F_GLOBAL:
    if(needlval & 1)
    {
      ins_f_byte(F_PUSH_IDENTIFIER_LVALUE);
      ins_byte(n->u.s.a.number);
    }else{
      ins_f_byte(F_IDENTIFIER);
      ins_byte(n->u.s.a.number);
    }
    return 1;

  case F_CONSTANT_FUNCTION:
    if(needlval & 1)
    {
      yyerror("Function name is not a lvalue.");
    }
    if(n->u.s.a.number<256)
    {
      ins_f_byte(F_SHORT_CONSTANT_FUNCTION);
      ins_byte(n->u.s.a.number);
    }else{
      ins_f_byte(F_CONSTANT_FUNCTION);
      ins_short(n->u.s.a.number);
    }
    return 1;

  case F_PUSH_SIMUL_EFUN:
    ins_f_byte(F_PUSH_SIMUL_EFUN);
    ins_short(n->u.s.a.number);
    return 1;
    
  default:
    fatal("Infernal compiler error (unknown parse-tree-token).\n");
    return 0;			/* make gcc happy */
  }
}

void setup_fake_program()
{
  fake_prog.next_hash=0;
  fake_prog.program=mem_block[A_PROGRAM].block;
  fake_prog.program_size=mem_block[A_PROGRAM].current_size;
  fake_prog.line_numbers=(unsigned short *)mem_block[A_LINENUMBERS].block;
  fake_prog.num_line_numbers=
    mem_block[A_LINENUMBERS].current_size/sizeof(unsigned short);
  fake_prog.strings=(char **)mem_block[A_STRINGS].block;
  fake_prog.num_strings=
    mem_block[A_STRINGS].current_size/sizeof(char **);
  fake_prog.constants=(struct svalue *)mem_block[A_CONSTANTS].block;
  fake_prog.num_constants=
    mem_block[A_CONSTANTS].current_size/sizeof(struct svalue);
  fake_prog.functions=(struct function *)mem_block[A_FUNCTIONS].block;
  fake_prog.num_functions=
    mem_block[A_FUNCTIONS].current_size/sizeof(struct function);
  fake_prog.function_ptrs=
    (struct function_p *)mem_block[A_FUNCTION_PTRS].block;
  fake_prog.num_function_ptrs=
    mem_block[A_FUNCTION_PTRS].current_size/sizeof(struct function_p);
  fake_prog.switch_mappings=(struct vector **)
    mem_block[A_SWITCH_MAPPINGS].block;
  fake_prog.num_switch_mappings=
    mem_block[A_SWITCH_MAPPINGS].current_size/sizeof(struct vector **);
  fake_prog.inherit=(struct inherit *)mem_block[A_INHERITS].block;
  fake_prog.num_inherited = mem_block[A_INHERITS].current_size /
    sizeof (struct inherit);
  fake_prog.inherit[0].prog=&fake_prog;
  fake_object.prog=&fake_prog;
  fake_object.ref=1;
}

int eval_low(node *n)
{
  extern struct svalue *sp;
  extern int error_recovery_context_exists;

  extern jmp_buf error_recovery_context;

  extern struct object *command_giver,*current_object;
  struct object *save_command_giver = command_giver;
  struct object *save_current_object = current_object;
  extern struct svalue catch_value;
  extern struct control_stack *csp;
  extern struct program *current_prog;
  extern struct svalue *fp;

  int returnval;
  int foo,fo2,e;
  char **bar;
  struct svalue *save_sp,*gazonk;
  int jump=PC;
  save_sp=sp;

  if(num_parse_error)
    return -1;

  store_linenumbers=0;
  do_docode(n,0);
  ins_f_byte(F_DUMB_RETURN);
  store_linenumbers=1;

  if(num_parse_error)
    return -1;

  foo=mem_block[A_STRINGS].current_size/sizeof(char *);
  fo2=mem_block[A_CONSTANTS].current_size/sizeof(struct svalue);
  
  push_control_stack(0);
  push_pop_error_context (1);

  csp->num_local_variables = 0; /* No extra variables */
  fp = sp - csp->num_local_variables + 1;

  setup_fake_program();
  current_prog=&fake_prog;
  current_object=&fake_object;
  /* signal catch OK - print no err msg */
  error_recovery_context_exists = 2;

  if (setjmp(error_recovery_context))
  {
    push_pop_error_context(-1);
    pop_control_stack();
    if(catch_value.type==T_STRING)
    {
      yyerror(catch_value.u.string);
    }else{
      yyerror("Eval error");
    }
    while(sp>save_sp) pop_stack();
    returnval=-1;
  }else{
    extern char *pc;
    pc=mem_block[A_PROGRAM].block+jump;
    eval_instruction();
    pop_control_stack();
    PC=jump;
    push_pop_error_context(0);
    returnval=sp-save_sp;
  }

  /* free any strings we no longer need */
  bar=(char **)mem_block[A_STRINGS].block;
  e=mem_block[A_STRINGS].current_size/sizeof(char *)-1;
  for(;e>=foo;e--) free_string(bar[e]);
  mem_block[A_STRINGS].current_size=foo*sizeof(char *);

  /* remove constants not needed */
  gazonk=(struct svalue *)mem_block[A_CONSTANTS].block;
  e=mem_block[A_CONSTANTS].current_size/sizeof(struct svalue)-1;
  for(;e>=fo2;e--) free_svalue(gazonk+e);
  mem_block[A_CONSTANTS].current_size=fo2*sizeof(struct svalue);

  command_giver=save_command_giver;
  current_object=save_current_object;
  return returnval;
}

node *eval(node *n)
{
  int args;
  extern struct svalue *sp;
  if(!is_const(n)) return n; /* fix this later */
  args=eval_low(n);

  switch(args)
  {
  case -1:
    n=copy_node(n);
    break;

  case 0:
    n=NULL;
    break;

  case 1:
    n=mksvaluenode(sp);
    pop_stack();
    break;

  default:
    n=0;
    while(args--)
    {
      n=mknode(F_ARG_LIST,mksvaluenode(sp),n);
      pop_stack();
    }
  }
  return n;
}

int last_function_opt_info;

void dooptcode(node *n,int args)
{
  last_function_opt_info=OPT_SIDE_EFFECT;
#ifdef LASDEBUG
  if(lasdebug)
  {
    fprintf(stderr,"Optimizing:\n");
    nice_print_tree(n);
  }
#endif
  if(!num_parse_error)
  {
    n=optimize(n,0);
#ifdef LASDEBUG
    if(lasdebug>1)
    {
      fprintf(stderr,"Result after optimize:\n");
      nice_print_tree(n);
    }
#endif
    do_depend_optim(&n,args);
    n=optimize(n,0);
#ifdef LASDEBUG
    if(lasdebug)
    {
      fprintf(stderr,"Result after depend_optim:\n");
      nice_print_tree(n);
    }
#endif
  }
  if(!num_parse_error)
  {
    do_pop(docode(n));
    last_function_opt_info=n->opt_info;
  }
  free_node(n);
}

int get_opt_info()
{
  if (last_function_opt_info&(OPT_SIDE_EFFECT | OPT_GLOBAL_ASSIGNMENT))
    return TYPE_MOD_SIDE_EFFECT;
  if(last_function_opt_info & OPT_GLOBAL_DEPEND)
    return 0;
  else
    return TYPE_MOD_CONSTANT;
}

