#include "efuns.h"
#include "array.h"

void f_m_aggregate(int num_arg,struct svalue *argp)
{
  int i;
  struct vector *v, *w;

  if(num_arg%2)
    error("Odd number of arguments to m_aggregate.\n",0,0,0,0,0,0,0,0);

  v = allocate_array(num_arg / 2);
  w = allocate_array(num_arg / 2);
  for (i=0; i < num_arg; i += 2)
  {
    assign_svalue_no_free(&v->item[i / 2], argp + i);
    assign_svalue_no_free(&w->item[i / 2], argp + i + 1);
  }
  pop_n_elems(num_arg);
  v = allocate_mapping(v, w);
  order_alist(v);
  push_mapping(v);
  v->ref--;
}

void f_m_cleanup(int num_arg,struct svalue *argp)
{
  struct vector *v;
  extern struct vector *cleanup_map(struct vector *);
  v=cleanup_map(argp->u.vec);
  pop_stack();
  push_mapping(v);
  v->ref--;
}

void f_sum_mappings(int num_arg,struct svalue *argp)
{
  struct vector *v;
  extern struct vector *sum_mappings(struct vector *,struct vector *,struct svalue *,struct svalue *,struct svalue *);

  if(!(argp[2].type & (T_FUNCTION | T_NUMBER))) bad_arg(2);
  if(!(argp[3].type & (T_FUNCTION | T_NUMBER))) bad_arg(3);
  if(!(argp[4].type & (T_FUNCTION | T_NUMBER))) bad_arg(4);

  v=sum_mappings(argp[0].u.vec,argp[1].u.vec,argp+2,argp+3,argp+4);
  pop_n_elems(5);
  push_mapping(v);
  v->ref--;
}

void f_m_set_default(int num_arg,struct svalue *argp)
{
  map_set_default(argp[0].u.vec,argp+1);
  pop_stack();
}

void f_m_get_default(int num_arg,struct svalue *argp)
{      
  assign_svalue(sp,map_get_default(argp[0].u.vec));
}

void f_filter_mapping(int num_arg,struct svalue *argp)
{
  struct vector *v,*res,*m;
  int e,s;

  m=argp[0].u.vec;
  check_map_for_destr(m);
  res = map_array(m->item[1].u.vec,argp+1,argp+2,num_arg-2);

  for(s=e=0;e<res->size;e++)
    if(res->item[e].type!=T_NUMBER || res->item[e].u.number!=0) s++;

  v = allocate_mapping(allocate_array(s),allocate_array(s));

  for(s=e=0;e<res->size;e++)
  {
    if(res->item[e].type!=T_NUMBER || res->item[e].u.number!=0)
    {
      assign_svalue(v->item[0].u.vec->item+s,m->item[0].u.vec->item+e);
      assign_svalue(v->item[1].u.vec->item+s,m->item[1].u.vec->item+e);
      s++;
    }
  }
  free_vector(res);
  pop_n_elems(num_arg);
  push_mapping(v);		/* This will make ref count == 2 */
  v->ref--;
}

void f_mkmapping(int num_arg,struct svalue *argp)
{
  struct vector *v;
  int i;

  i = (sp-1)->u.vec->size;
  if (i > sp->u.vec->size)
  {
    i = sp->u.vec->size;
  }
  v = allocate_mapping(slice_array((sp-1)->u.vec, 0, i - 1),
		       slice_array(sp->u.vec, 0, i - 1));
  pop_n_elems(2);
  order_alist(v);
  push_mapping(v);
  v->ref--;
}

void f_indices(int num_arg,struct svalue *argp)
{
  struct vector *v;
  switch(sp->type)
  {
  default:
    bad_arg(1);
    break;
   
  case T_LIST:
  case T_MAPPING:
    v=sp->u.vec;
    check_alist_for_destruct(v);
    v=v->item[0].u.vec;
    v=slice_array(v,0,v->size-1);
    pop_stack();
    push_vector(v);
    v->ref--;
  }
}

void f_m_indices(int num_arg,struct svalue *argp)
{
  struct vector *v;

  v = sp->u.vec;
  check_map_for_destr(v);
  v = v->item[0].u.vec;
  v = slice_array(v, 0, v->size - 1);
  pop_stack();
  push_vector(v);
  v->ref--;			/* Will make ref count == 1 */
}

void f_m_values(int num_arg,struct svalue *argp)
{
  struct vector *v;

  v = sp->u.vec;
  check_map_for_destr(v);
  v = v->item[1].u.vec;
  v = slice_array(v, 0, v->size - 1);
  pop_stack();
  push_vector(v);
  v->ref--;			/* Will make ref count == 1 */
}

void f_m_delete(int num_arg,struct svalue *argp)
{
  struct vector *v;
  check_map_for_destr((sp-1)->u.vec);
  v=remove_mapping2((sp-1)->u.vec,sp);
  pop_stack();
}

void f_map_mapping(int num_arg,struct svalue *argp)
{
  struct vector *res;

  check_eval_cost();

  check_map_for_destr(argp[0].u.vec);
  res = map_array(argp[0].u.vec->item[1].u.vec,argp+1,argp+2,num_arg-2);
  res = allocate_mapping(argp[0].u.vec->item[0].u.vec,res);
  argp[0].u.vec->item[0].u.vec->ref++;
  pop_n_elems(num_arg);
  push_mapping(res);
  res->ref--;
}


void f_solidify_mapping(int num_arg,struct svalue *argp)
{
  int e,num,num2;
  struct vector *v,*ind,*val,*ret_ind,*ret_val,*data;
  v=argp[0].u.vec;
  ind=v->item[0].u.vec;
  val=v->item[1].u.vec;
  if(ind->size)
  {
    num=1;
    for(e=0;e<ind->size-1;e++)
      if(alist_cmp(ind->item+e,ind->item+e+1))
	num++;
  }else{
    num=0;
  }

  ret_ind=allocate_array(num);
  ret_val=allocate_array(num);
  num=0;
  for(e=0;e<ind->size;)
  {
    assign_svalue(ret_ind->item+num,ind->item+e);
    for(num2=0;num2+e<ind->size-1 &&
	!alist_cmp(ind->item+e+num2,ind->item+e+num2+1);num2++);
    data=allocate_array(num2+1);
    while(num2>=0)
    {
      assign_svalue_no_free(data->item+num2,val->item+e);
      num2--;
      e++;
    }
    ret_val->item[num].u.vec=data;
    ret_val->item[num].type=T_POINTER;
    num++;
  }
  pop_stack();
  v=allocate_mapping(ret_ind,ret_val);
  push_mapping(v);
  v->ref--;
}
