#include <math.h>
#include "efuns.h"
#include "array.h"
#include "simul_efun.h"
#include "stralloc.h"
#include "mapping.h"
#include "list.h"
#include "main.h"

extern int continue_loop;

void f_push_array()
{
  int e;
  struct vector *v;
  v=sp[0].u.vec;
  v->ref++;
  pop_stack();
  for(e=0;e<v->size;e++) push_svalue(v->item+e);
  free_vector(v);
}

char * get_relative_offset()
{
  short foo;
  ((char *)&foo)[0]=pc[0];
  ((char *)&foo)[1]=pc[1];
  pc+=2;
  return pc-2+foo;
}

void f_constant()
{
  unsigned short foo;
  ((char *)&foo)[0]=pc[0];
  ((char *)&foo)[1]=pc[1];
  pc+=2;
  sp++;
  assign_svalue(sp,current_prog->constants+foo);
}

void f_short_constant()
{
  sp++;
  assign_svalue(sp,current_prog->constants+EXTRACT_UCHAR(pc));
  pc++;
}

void f_float()
{
  float foo;
  memcpy((char *)&foo,pc,sizeof(float));
  pc+=sizeof(float);
  push_float(foo);
}

void f_foreach()
{
  char *new_pc;
  if((sp-2)->type != T_POINTER) bad_arg(1);
  if((sp-1)->type != T_LVALUE) bad_arg(2);
  if(!(sp-1)->u.lvalue->sval) bad_arg(2);
  if(sp->type != T_NUMBER) bad_arg(3);

  new_pc=get_relative_offset();
  if(sp->u.number<(sp-2)->u.vec->size)
  {
    assign_svalue((sp-1)->u.lvalue->sval,&(sp-2)->u.vec->item[sp->u.number++]);
    pc=new_pc;
  }
  check_eval_cost();
}

void f_inc_loop()
{
  char *new_pc;
  if(sp[-1].type != T_NUMBER) bad_arg(1);
  if(sp->type != T_LVALUE) bad_arg(2);
  if(!sp->u.lvalue->sval) bad_arg(2);
  if(sp->u.lvalue->sval->type != T_NUMBER) bad_arg(2);
    
  new_pc=get_relative_offset();
  if((++(sp->u.lvalue->sval->u.number))<sp[-1].u.number) pc=new_pc;
  check_eval_cost();
}

void f_dec_loop()
{
  char *new_pc;
  if(sp[-1].type != T_NUMBER) bad_arg(1);
  if(sp->type != T_LVALUE) bad_arg(2);
  if(!sp->u.lvalue->sval) bad_arg(2);
  if(sp->u.lvalue->sval->type != T_NUMBER) bad_arg(2);
    
  new_pc=get_relative_offset();
  if((--(sp->u.lvalue->sval->u.number))>sp[-1].u.number) pc=new_pc;
  check_eval_cost();
}

void f_inc_neq_loop()
{
  char *new_pc;
  if(sp[-1].type != T_NUMBER) bad_arg(1);
  if(sp->type != T_LVALUE) bad_arg(2);
  if(!sp->u.lvalue->sval) bad_arg(2);
  if(sp->u.lvalue->sval->type != T_NUMBER) bad_arg(2);

  new_pc=get_relative_offset();
  if((++(sp->u.lvalue->sval->u.number))!=sp[-1].u.number) pc=new_pc;
  check_eval_cost();
}

void f_dec_neq_loop()
{
  char *new_pc;
  if(sp[-1].type != T_NUMBER) bad_arg(1);
  if(sp->type != T_LVALUE) bad_arg(2);
  if(!sp->u.lvalue->sval) bad_arg(2);
  if(sp->u.lvalue->sval->type != T_NUMBER) bad_arg(2);

  new_pc=get_relative_offset();
  if((--(sp->u.lvalue->sval->u.number))!=sp[-1].u.number) pc=new_pc;
  check_eval_cost();
}

void f_push_cost() { push_number(eval_cost); }
void f_pop_value() { pop_stack(); }
void f_pop_n_elems() { pop_n_elems(EXTRACT_UCHAR(pc++)); }

void f_dup()
{
  sp++;
  assign_svalue_no_free(sp, sp-1);
}

void f_land()
{
  char *new_pc;
  new_pc=get_relative_offset();
  if (sp->type == T_NUMBER && sp->u.number == 0)
    pc = new_pc;
  else
    pop_stack();
}

void f_lor()
{
  char *new_pc;
  new_pc=get_relative_offset();
  if (sp->type != T_NUMBER || sp->u.number != 0)
    pc=new_pc;
  else
    pop_stack();
}

void f_constant_function()
{
  unsigned short offset;

  ((char *)&offset)[0] = pc[0];
  ((char *)&offset)[1] = pc[1];
  pc+=2;
  if(current_object->flags & O_DESTRUCTED)
  {
    push_zero();
  }else{
    sp++;
    sp->type=T_FUNCTION;
    sp->subtype=offset+function_index_offset;
    sp->u.ob=current_object;
    add_ref(current_object,"constant_function");
  }
}

void f_short_constant_function()
{
  unsigned short offset;
  offset=EXTRACT_UCHAR(pc++);
  if(current_object->flags & O_DESTRUCTED)
  {
    push_zero();
  }else{
    sp++;
    sp->type=T_FUNCTION;
    sp->subtype=offset+function_index_offset;
    sp->u.ob=current_object;
    add_ref(current_object,"constant_function");
  }
}

void f_push_simul_efun()
{
  unsigned short offset;

  ((char *)&offset)[0] = pc[0];
  ((char *)&offset)[1] = pc[1];
  pc+=2;
  if(simul_efun_object->flags & O_DESTRUCTED)
  {
    push_zero();
  }else{
    sp++;
    sp->type=T_FUNCTION;
    sp->subtype=offset;
    sp->u.ob=simul_efun_object;
    add_ref(simul_efun_object,"push_simul_efun");
  }
}


void f_branch()
{
  pc=get_relative_offset();
  check_eval_cost();
}

void f_branch_when_zero()
{
  char *new_pc;
  new_pc=get_relative_offset();
  if (sp->type == T_NUMBER && sp->u.number == 0) pc=new_pc;
  pop_stack();
  check_eval_cost();
}

void f_branch_when_non_zero()
{
  char *new_pc;
  new_pc=get_relative_offset();
  if (sp->type != T_NUMBER || sp->u.number != 0) pc=new_pc;
  pop_stack();
  check_eval_cost();
}

void f_short_branch()
{
  pc-=EXTRACT_UCHAR(pc);
  check_eval_cost();
}

void f_short_branch_when_zero()
{
  if (sp->type == T_NUMBER && sp->u.number == 0) pc-=EXTRACT_UCHAR(pc);
  else pc++;
  pop_stack();
  check_eval_cost();
}

void f_short_branch_when_non_zero()
{
  if (sp->type != T_NUMBER || sp->u.number != 0) pc-=EXTRACT_UCHAR(pc);
  else pc++;
  pop_stack();
  check_eval_cost();
}

#define T_VECTOR_MASK (T_POINTER | T_MAPPING | T_ALIST_PART)

struct lvalue lvalues[LVALUES];
struct lvalue *lsp=lvalues-1;

void remove_array_from_stack(struct vector *v)
{
  struct lvalue *l;

  for (l = lvalues; l <= lsp; l++)
    if (l->parent &&
	(l->parent->type & (T_VECTOR_MASK | T_LIST)) &&
	l->parent->u.vec==v)
      l->sval=l->parent=NULL;
}

struct lvalue *push_indexed_lvalue(struct svalue *s,struct svalue *i,int needlval)
{
  int ind,len;

  if(!s) error("Illigal indexing.\n",0,0,0,0,0,0,0,0,0);

#ifdef MALLOC_DEBUG
  check_sfltable();
#endif
  switch(s->type)
  {
  case T_LVALUE:
  {
    /* note that piece of code makes it impossible to use the old
     * lvalue after the new one has been made
     */
    struct svalue *s2;
    s2=s->u.lvalue->sval;
    if(!s2)
      error("Illigal lvalue request.\n",0,0,0,0,0,0,0,0,0);
    *s=const0;
    /* this will not work if the lvalue given wasn't the last one...*/
    lsp--;
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return push_indexed_lvalue(s2,i,needlval);
  }

  case T_POINTER:
    if(i->type!=T_NUMBER)
      error("Index isn't an integer.\n",0,0,0,0,0,0,0,0,0);
    ind=i->u.number;
    len=s->u.vec->size;
    if(ind<0) ind+=len;
    if(ind<0 || ind>=len)
      error("Index out of range.\n",0,0,0,0,0,0,0,0,0);
    lsp++;
    if(lsp>=lvalues+LVALUES)
      error("Lvalue stack overflow.\n",0,0,0,0,0,0,0,0,0);
    lsp->type=LVALUE_VECTOR_INDEX;
    lsp->parent=s;
    lsp->sval=s->u.vec->item+ind;
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return lsp;

  case T_STRING:
    if(i->type!=T_NUMBER)
      error("Index isn't an integer.\n",0,0,0,0,0,0,0,0,0);
    ind=i->u.number;
    len=my_strlen(s);
    if(ind<0) ind+=len;
    if(ind<0 || ind>=len) ind=len;
    lsp++;
    if(lsp>=lvalues+LVALUES)
      error("Lvalue stack overflow.\n",0,0,0,0,0,0,0,0,0);
    lsp->type=LVALUE_STRING_INDEX;
    lsp->sval=NULL;
    lsp->parent=s;
    lsp->start=ind;
    lsp->end=ind;
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return lsp;

  case T_MAPPING:
    lsp++;
    if(lsp>=lvalues+LVALUES)
      error("Lvalue stack overflow.\n",0,0,0,0,0,0,0,0,0);
    lsp->type=LVALUE_MAPPING_INDEX;
    lsp->parent=s->u.vec->item+1;
    lsp->sval = get_map_lvalue(s->u.vec, i, needlval);
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return lsp;

  case T_LIST:
    lsp++;
    if(lsp>=lvalues+LVALUES)
      error("Lvalue stack overflow.\n",0,0,0,0,0,0,0,0,0);
    lsp->type=LVALUE_LIST_INDEX;
    lsp->parent=NULL;
    lsp->sval = get_list_lvalue(s->u.vec, i);
    return lsp;
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    break;

  default:
    error("Indexing a basic type.\n",0,0,0,0,0,0,0,0);
    return 0; /* make gcc happy */
  }
}


void assign(struct lvalue *l,struct svalue *s)
{
  if(l->sval)
  {
    if(l->parent && (l->parent->type & T_VECTOR_MASK))
    {
      if((s->type & T_VECTOR_MASK))
      {
	if(l->parent->u.vec->ref>1)
	  if(check_for_circularity(s->u.vec,l->parent->u.vec))
	    error("Illigal assignment, cyclic arrays not supported.\n",0,0,0,0,0,0,0,0,0);
      }
      l->parent->u.vec->types|=s->type;
    }
    assign_svalue(l->sval,s);
    return;
  }else if(l->parent && l->parent->type==T_STRING){
    char *s1,*s2;
    int len;

    if(l->start==l->end &&
	(s->type==T_INT || (s->type==T_STRING && my_strlen(s)==1)))
    {
      int ch;

      if(s->type==T_INT) ch=s->u.number;
      else ch=s->u.string[0];

      switch(l->parent->subtype)
      {
        case STRING_MALLOC:
          /* this _must_ be my lucky day... */
          l->parent->u.string[l->start]=ch;
          return;

        case STRING_CONSTANT:
        case STRING_TEMPORARY:
          if(s->u.number)
          {
            /* let's not make a binary string.... */
            s1=string_copy(l->parent->u.string);
            l->parent->subtype=STRING_MALLOC;
            l->parent->u.string=s1;
            s1[l->start]=ch;
            return;
          }

        case STRING_SHARED:
          /* we make a new shared string */
          len=my_strlen(l->parent);
	  if(l->start>=len || l->start<0)
	    error("Index out of range.\n",0,0,0,0,0,0,0,0,0);
	  s1=begin_shared_string(len);
          memcpy(s1,l->parent->u.string,len);
          s1[l->start]=ch;
          s1=end_shared_string(s1);
          free_svalue(l->parent);
          l->parent->type=T_STRING;
          l->parent->subtype=STRING_SHARED;
          l->parent->u.string=s1;
          return;
      }
    }

    if(s->type==T_STRING)
    {
      int len2;
      /* we make a new shared string */
      len=my_strlen(l->parent);
      len2=my_strlen(s);

      if(l->start<0) l->start=0;
      if(l->start>len) l->start=len;
      if(l->end<l->start-1) l->end=l->start-1;
      if(l->end>len) l->end=len;


      len=len-l->end+l->start-1+len2;
      s2=s1=begin_shared_string(len);
  
      if(l->start>0)
      {
        memcpy(s2,l->parent->u.string,l->start);
        s2+=l->start;
      }

      if(len2)
      {
        memcpy(s2,s->u.string,len2);
        s2+=len2;
      }

      if((len2=len+s1-s2)>0)
      {
        memcpy(s2,l->parent->u.string+l->end+1,len2);
        s2+=len2;
      }
 
      s2=end_shared_string(s1);
      free_svalue(l->parent);
      l->parent->type=T_STRING;
      l->parent->subtype=STRING_SHARED;
      l->parent->u.string=s2;
      return;
    }
  }
  error("Error in assignment.\n",0,0,0,0,0,0,0,0,0);
}

void lvalue_to_svalue(struct svalue *s,struct lvalue *l)
{
#ifdef MALLOC_DEBUG
  check_sfltable();
#endif
  if(l->sval)
  {
    assign_svalue(s,l->sval);
#ifdef MALLOC_DEBUG
    check_sfltable();
#endif
    return;
  }else if(l->parent){

    if(l->type==LVALUE_STRING_INDEX)
    {
      free_svalue(s);
      s->type=T_INT;
      s->u.number=l->parent->u.string[l->start];
      return;
    }

    if(l->type==LVALUE_RANGE)
    {
      push_svalue(l->parent);
      push_number(l->start);
      push_number(l->end);
      f_range(3,sp-3);
      assign_svalue(s,sp);
      pop_stack();
      return;
    }
  }
  error("Error when referencing lvalue.\n",0,0,0,0,0,0,0,0);
}

void f_indirect()
{
#ifdef DEBUG
  if (sp->type != T_LVALUE)
    fatal("Bad type to F_INDIRECT\n",0,0,0,0,0,0,0,0,0);
#endif
  sp++;
  lvalue_to_svalue(sp,sp[-1].u.lvalue);
  f_swap();
  pop_stack();
}

void f_push_indexed_lvalue()
{
  struct lvalue *l;
  l=push_indexed_lvalue(sp-1,sp,1);
  pop_stack();
  pop_stack();
  sp++;
  sp->type=T_LVALUE;
  sp->u.lvalue=l;
}

void f_index()
{
  struct lvalue *l;
  struct svalue s=const0;
  
  l=push_indexed_lvalue(sp-1,sp,0);
  lvalue_to_svalue(&s,l);
  lsp=l-1;
  pop_stack();
  pop_stack();
  sp++;
  *sp=s;
}

/*
 * Get address to a valid global variable.
 */
struct svalue *find_value(int num)
{
#ifdef DEBUG
    if (num >= current_object->prog->num_variables) {
	fatal("Illegal variable access %d(%d) (%s). See trace above.\n",
	    (void *)num, (void *)(int)current_object->prog->num_variables,
	      current_object->prog->name,0,0,0,0,0,0);
    }
#endif
    return current_object->variables+num;
}

void f_identifier()
{
  sp++;
  assign_svalue_no_free(sp,
	find_value((int)(EXTRACT_UCHAR(pc)+variable_index_offset)));
  pc++;
}

void f_push_identifier_lvalue()
{
  lsp++;
  if(lsp>=lvalues+LVALUES)
    error("Lvalue stack overflow.\n",0,0,0,0,0,0,0,0,0);
  lsp->sval = find_value((int)(EXTRACT_UCHAR(pc) +
				  variable_index_offset));
  lsp->parent=NULL;
  sp++;
  sp->type = T_LVALUE;
  sp->u.lvalue=lsp;
  pc++;
}

void f_local_name()
{
  extern struct svalue *fp;
  sp++;
  
  assign_svalue_no_free(sp, fp + EXTRACT_UCHAR(pc));
  pc++;
}

void f_push_local_variable_lvalue()
{
  extern struct svalue *fp;

  lsp++;
  if(lsp>=lvalues+LVALUES)
    error("Lvalue stack overflow.\n",0,0,0,0,0,0,0,0,0);
  lsp->sval = fp + EXTRACT_UCHAR(pc);
  lsp->parent = NULL;
  sp++;
  sp->type = T_LVALUE;
  sp->u.lvalue=lsp;
  pc++;
}

void f_return()
{
  extern struct svalue *fp;
  struct svalue sv;

  sv = *sp;
  *sp=const0;
  sp--;
  /*
   * Deallocate frame and return.
   */
  while((sp+1)!=fp) pop_stack();
  sp++;
#ifdef DEBUG
  if (sp != fp)
    fatal("Bad stack at F_RETURN\n",0,0,0,0,0,0,0,0,0); /* marion */
#endif
  *sp = sv; /* This way, the same ref counts are maintained */
  continue_loop=0;
}

void f_dumb_return() { continue_loop=0; }

void f_catch()
{
  /*
   * Catch/Throw - catch errors in system or other peoples routines.
   */
  extern jmp_buf error_recovery_context;
  extern int error_recovery_context_exists;
  extern struct svalue catch_value;
  char *new_pc;
  struct svalue *save_sp;

  new_pc=get_relative_offset();

  save_sp=sp;
  push_control_stack(0);
  csp->num_local_variables = 0; /* No extra variables */
  csp->pc = new_pc;
  csp->num_local_variables = (csp-1)->num_local_variables; /* marion */
  /*
   * Save some global variables that must be restored separately
   * after a longjmp. The stack will have to be manually popped all
   * the way.
   */
  push_pop_error_context (1);
	
  /* signal catch OK - print no err msg */
  error_recovery_context_exists = 2;
  if (setjmp(error_recovery_context))
  {
    /*
     * They did a throw() or error. That means that the control
     * stack must be restored manually here.
     * Restore the value of expected_stack also. It is always 0
     * for catch().
     */
    push_pop_error_context (-1);
    pop_control_stack();
    assign_svalue_no_free(++sp, &catch_value);
  }else{
    /* next error will return 1 by default */
    assign_svalue(&catch_value, &const1);
    eval_instruction();
    pop_control_stack();
    push_pop_error_context(0);
    while(sp>save_sp) pop_stack();
    push_zero();
  }
}

void f_swap()
{
  struct svalue a;
  a=*(sp-1);
  *(sp-1)=*sp;
  *sp=a;
}

void f_swap_variables()
{
  sp++;
  lvalue_to_svalue(sp,sp[-2].u.lvalue);
  sp++;
  lvalue_to_svalue(sp,sp[-2].u.lvalue);
  assign(sp[-2].u.lvalue,sp-1);
  assign(sp[-3].u.lvalue,sp);
  pop_n_elems(4);
}

void f_push_ltosval()
{
  if(sp[-1].type!=T_LVALUE)
    error("RHS: not lvalue!\n",0,0,0,0,0,0,0,0,0);
  sp++;
  lvalue_to_svalue(sp,sp[-2].u.lvalue);
  f_swap();
  /* this will make some things faster */
  if(sp[-2].u.lvalue->sval)
    free_svalue(sp[-2].u.lvalue->sval);
}


void f_cast_to_object()
{
  struct object *v;
  switch(sp->type)
  {
  case T_OBJECT:
    break;
  case T_FUNCTION:
    sp->type=T_OBJECT;
    break;
  case T_STRING:
    v=find_hashed(sp->u.string);
    if(v)
    {
      pop_stack();
      push_object(v);
    }else{
      error("Couldn't find object '%s'\n",sp->u.string,0,0,0,0,0,0,0,0);
    }
    break;

  case T_INT:
  {
    /* Here I try to use a somewhat odd 'skiplist' /Profezzorn */
    int i;
    i=sp->u.number;
    pop_stack();
    for(v=obj_list;v && v->clone_number>i;)
    {
#define DOCHECK(X) if(v->X && v->X->clone_number<v->clone_number && v->X->clone_number>=i) { v=v->X; continue; }
      DOCHECK(next_heart_beat);
      DOCHECK(next_obj_hashed);
      DOCHECK(next_inv);
      DOCHECK(contains);
      DOCHECK(super);
      v=v->next_all;
    }
    if(!v || v->clone_number!=i)
    {
      push_zero();
    }else{
      push_object(v);
    }
    break;
  }
    
  default:
    error("Casting a strange value.\n",0,0,0,0,0,0,0,0,0);
    break;
  }
}

void f_cast_to_function()
{
  int function;
  switch(sp->type)
  {
  case T_FUNCTION:
    break;

  default:
    error("Casting a strange value.\n",0,0,0,0,0,0,0,0,0);
    break;

  case T_STRING:
    if(sp[0].subtype==STRING_SHARED)
      function=find_shared_string_function(sp[0].u.string,
					   current_object->prog);
    else
      function=find_function(sp[0].u.string,current_object->prog);
      
    pop_stack();
    if(function>=0)
    {
      sp++;
      sp->type=T_FUNCTION;
      sp->u.ob=current_object;
      add_ref(current_object, "(function)");
      sp->subtype=function;
    }else{
      push_number(0);
    }
  }
}
    
void f_cast_to_float()
{
  float f;
  switch(sp->type)
  {
  case T_NUMBER:
    f=(float)sp->u.number;
    pop_stack();
    push_float(f);
    break;

  case T_FLOAT:
    break;

  case T_STRING:
    f=(float)atof(sp->u.string);
    pop_stack();
    push_float(f);
    break;

  default:
    error("Casting a strange value.\n",0,0,0,0,0,0,0,0,0);
  }
}

void f_cast_to_int()
{
  int i;
  switch(sp->type)
  {
  case T_FLOAT:
    i=sp->u.fnum;
    pop_stack();
    push_number((int)i);
    break;

  case T_NUMBER:
    break;

  case T_STRING:
    i=atoi(sp->u.string);
    pop_stack();
    push_number(i);
    break;

  default:
    error("Casting a strange value.\n",0,0,0,0,0,0,0,0,0);
  }
}

void f_cast_to_string()
{
  char buf[30];
  switch(sp->type)
  {
  case T_FLOAT:
    sprintf(buf,"%f",sp->u.fnum);
    pop_stack();
    push_new_shared_string(buf);
    break;

  case T_NUMBER:
    sprintf(buf,"%d",sp->u.number);
    pop_stack();
    push_new_shared_string(buf);
    break;

  case T_STRING:
    break;

  default:
    error("Casting a strange value.\n",0,0,0,0,0,0,0,0,0);
  }
}

void f_switch()
{ 
  /* Ta-da, and the wonderful wizard of oz waved his wand, */ 
  /* and all amylaars code was gone.        /Profezzorn    */
  unsigned short foo;
  struct vector *m,*val,*ind;
  int def;

  ((char *)&foo)[0]=pc[0];
  ((char *)&foo)[1]=pc[1];
  pc+=2;
  m=current_prog->switch_mappings[foo];
  ind=m->item[0].u.vec;
  val=m->item[1].u.vec;
  def=m->item[2].u.number;
  if(sp->type==T_STRING && sp->subtype!=STRING_SHARED)
  {
    char *foo=make_shared_string(sp->u.string);
    pop_stack();
    push_shared_string(foo);
  }
  foo=search_alist(sp,ind);
  if(foo<ind->size && !alist_cmp(ind->item+foo,sp))
  {
    pc=current_prog->program+val->item[foo].u.number;
  }else{
    if(foo<ind->size && val->item[foo].subtype)
    {
      pc=current_prog->program+val->item[foo].subtype;
    }else{
      pc=current_prog->program+def;
    }
  }

  pop_stack();
}


/*
  flags:
   *
  operators:
  %d
  %s
  %f
  %c
  %n
  %[
  %%
*/

int read_set(char *match,int cnt,char *set,int match_len)
{
  int init;
  int last=0;
  int e;

  if(cnt>=match_len)
    error("Error in sscanf format string.\n",0,0,0,0,0,0,0,0);

  if(match[cnt]=='^')
  {
    for(e=0;e<256;e++) set[e]=1;
    init=0;
    cnt++;
    if(cnt>=match_len)
      error("Error in sscanf format string.\n",0,0,0,0,0,0,0,0);
  }else{
    for(e=0;e<256;e++) set[e]=0;
    init=1;
  }
  if(match[cnt]==']')
  {
    set[last=']']=init;
    cnt++;
    if(cnt>=match_len)
      error("Error in sscanf format string.\n",0,0,0,0,0,0,0,0);
  }

  for(;match[cnt]!=']';cnt++)
  {
    if(match[cnt]=='-')
    {
      cnt++;
      if(cnt>=match_len)
	error("Error in sscanf format string.\n",0,0,0,0,0,0,0,0);
      if(match[cnt]==']')
      {
	set['-']=init;
	break;
      }
      for(e=last;e<match[cnt];e++) set[e]=init;
    }
    set[last=match[cnt]]=init;
  }
  return cnt;
}

int inter_sscanf(int num_arg)
{
  char *input;
  int input_len;
  char *match;
  int match_len;
  struct svalue sval;
  int e,cnt,matches,eye,arg;
  int no_assign;
  char set[256];
  struct svalue *argp;
  
  argp=sp-num_arg+1;

  check_argument(0,T_STRING);
  check_argument(1,T_STRING);

  input=argp[0].u.string;
  input_len=my_strlen(argp+0);

  match=argp[1].u.string;
  match_len=my_strlen(argp+1);

  arg=eye=matches=0;

  for(cnt=0;cnt<match_len;)
  {
    for(;cnt<match_len;cnt++)
    {
      if(match[cnt]=='%')
      {
        if(match[cnt+1]=='%')
        {
          cnt++;
        }else{
          break;
        }
      }
      if(eye>=input_len ||input[eye]!=match[cnt])
	return matches;
      eye++;
    }
    if(cnt>=match_len) return matches;

#ifdef DEBUG
    if(match[cnt]!='%' || match[cnt+1]=='%')
    {
      fatal("Error in sscanf.\n",0,0,0,0,0,0,0,0,0);
    }
#endif

    cnt++;
    if(cnt>=match_len)
      error("Error in sscanf format string.\n",0,0,0,0,0,0,0,0);

    if(match[cnt]=='*')
    {
      no_assign=1;
      cnt++;
      if(cnt>=match_len)
        error("Error in sscanf format string.\n",0,0,0,0,0,0,0,0);
    }else{
      no_assign=0;
    }

    switch(match[cnt])
    {
      case 'c':
        if(eye>=input_len) return matches;
        sval.type=T_NUMBER;
	sval.subtype=NUMBER_NUMBER;
        sval.u.number=input[eye];
        eye++;
        break;

      case 'd':
        if(eye>=input_len) return matches;

        if(isdigit(input[eye]) || input[eye]=='-')
        {
          char *t;          
          sval.type=T_NUMBER;
	  sval.subtype=NUMBER_NUMBER;
          sval.u.number=strtol(input+eye,&t,0);
          eye=t-input;
        }else{
          return matches;
        }
        break;

      case 's':
        if(cnt+1>=match_len)
        {
          sval.type=T_STRING;
          sval.subtype=STRING_SHARED;
          sval.u.string=make_shared_binary_string(input+eye,input_len-eye);
	  break;
        }else{
          char *end_str_start;
          char *end_str_end;
          char *s=0; /* make gcc happy */
	  char *p=0; /* make gcc happy */
          int start;

          start=eye;
          end_str_start=match+cnt+1;
          
	  if(match[cnt+1]=='%')
	  {
	    s=match+cnt+2;
	    if(*s=='*') s++;
	    switch(*s)
	    {
	    case 's':
	      error("Illigal to have to adjecent %%s.\n",0,0,0,0,0,0,0,0,0);
	      return 0; /* make gcc happy */
	      
	    case 'd':
	      for(e=0;e<256;e++) set[e]=1;
	      for(e='0';e<='9';e++) set[e]=0;
	      set['-']=0;
	      goto match_set;

	    case 'f':
	      for(e=0;e<256;e++) set[e]=1;
	      for(e='0';e<='9';e++) set[e]=0;
	      set['.']=set['-']=0;
	      goto match_set;

	    case '[': /* oh dear */
	      read_set(match,s-match+1,set,match_len);
	      for(e=0;e<256;e++) set[e]=!set[e];
	      goto match_set;
	    }
	  }

          for(e=cnt;e<match_len;e++)
            if(match[e]=='%' && match[e+1]!='%')
              break;
	   
          end_str_end=match+e;

          for(;eye<input_len;eye++)
          {
            p=input+eye;
            for(s=end_str_start;s<end_str_end;s++,p++)
            {
              if(*s!=*p) break;
              if(*s=='%') s++;
            }
            if(s==end_str_end)
              break;
          }
	  if(eye==input_len)
	    return matches;
          sval.type=T_STRING;
          sval.subtype=STRING_SHARED;
          sval.u.string=make_shared_binary_string(input+start,eye-start);
          if(s==end_str_end)
          {
            cnt=s-match-1;
            eye=p-input;
          }         
	  break;
        }

      case '[':
        if(eye>=input_len)
        {
          return matches;
        }else{

	  cnt=read_set(match,cnt+1,set,match_len);
/* I want to be able to match the empty set as well.
   if(!set[input[eye]]) return matches;
   */

	match_set:
          for(e=eye;eye<input_len && set[input[eye]];eye++);
          sval.type=T_STRING;
          sval.subtype=STRING_SHARED;
          sval.u.string=make_shared_binary_string(input+e,eye-e);
          break;
        }

      case 'n':
        sval.type=T_NUMBER;
	sval.subtype=NUMBER_NUMBER;
        sval.u.number=eye;
        break;
    
      case 'f':
        if(eye>=input_len) return matches;

        if(isdigit(input[eye]) || input[eye]=='-' || input[eye]=='.')
        {
	  char *t;
          sval.u.fnum=(float)strtod(input+eye,&t);
          sval.type=T_FLOAT;
          eye=t-input;
        }else{
          return matches;
        }
        break;        

      default:
        error("Unknown sscanf token %%%c\n",(void *)(int)match[cnt],0,0,0,0,0,0,0,0);
    }
    matches++;

    if(!no_assign)
    {
      arg++;
      if(num_arg-2<arg)
      {
	free_svalue(&sval);
	error("Too few arguments to sscanf.\n",0,0,0,0,0,0,0,0,0);
      }
      assign(argp[1+arg].u.lvalue,&sval);
    }
    free_svalue(&sval);
    cnt++;
  }
  return matches;
}

void f_sscanf()
{
  int num_arg,i;
  num_arg = EXTRACT_UCHAR(pc++);
  i = inter_sscanf(num_arg);
  pop_n_elems(num_arg);
  push_number(i);
}

void f_const0() { push_zero(); }
void f_const1() { push_one(); }
void f_const_1() { push_number(-1); }
void f_byte() { push_number(EXTRACT_UCHAR(pc++)+2); }
void f_neg_byte() { push_number(-EXTRACT_UCHAR(pc++)-2); }

void f_short()
{
  short foo;
  ((char *)&foo)[0] = pc[0];
  ((char *)&foo)[1] = pc[1];
  pc += 2;
  push_number(foo);
}
	
void f_number()
{
  int i;
  ((char *)&i)[0] = pc[0];
  ((char *)&i)[1] = pc[1];
  ((char *)&i)[2] = pc[2];
  ((char *)&i)[3] = pc[3];
  pc += 4;
  push_number(i);
}

 
void f_string()
{
  unsigned short string_number;
  ((char *)&string_number)[0] = pc[0];
  ((char *)&string_number)[1] = pc[1];
  pc += 2;
  push_shared_string(copy_shared_string(current_prog->strings[string_number]));
}

void f_short_string()
{
  push_shared_string(copy_shared_string(current_prog->strings[EXTRACT_UCHAR(pc++)]));
}

