#include "efuns.h"
#include "list.h"
#include "array.h"
#include "main.h"
#include "stralloc.h"

INLINE int is_eq(struct svalue *a, struct svalue *b)
{
  if (a->type != b->type) return 0;
  switch(a->type)
  {
  case T_NUMBER:
  case T_LIST:
  case T_POINTER:
  case T_OBJECT:
  case T_MAPPING:
  case T_ALIST_PART:
    return a->u.number == b->u.number;

  case T_STRING:
    return my_string_is_equal(a,b);

  case T_FUNCTION:
    return (a->subtype==b->subtype && a->u.ob==b->u.ob);
      
  case T_FLOAT:
    return a->u.fnum == b->u.fnum;

  default:
    fatal("Unknown type %x\n",(void *)(int)a->type,0,0,0,0,0,0,0,0);
    return 0; /* make gcc happy */
  }
}

int is_equal(struct svalue *a,struct svalue *b)
{
  struct svalue *ap,*bp;
  int e;

  if(a->type!=b->type) return 0;
  if(is_eq(a,b)) return 1;
  switch(a->type)
  {
    case T_NUMBER:
    case T_STRING:
    case T_OBJECT:
    case T_FLOAT:
    case T_FUNCTION:
      return 0;

    case T_POINTER:
    case T_ALIST_PART:
    case T_MAPPING:
    case T_LIST:
    if((e=a->u.vec->size)!=b->u.vec->size) return 0;
    ap=a->u.vec->item;
    bp=b->u.vec->item;
    for(;e>0;e--) if(!is_equal(ap++,bp++)) return 0;
      break;

    default:
      fatal("Unknown type in is_equal.\n");
  }
  return 1; /* survived */
}

INLINE int is_gt(struct svalue *a, struct svalue *b)
{
  if (a->type != b->type)
  {
    error("Cannot compare different types.\n",0,0,0,0,0,0,0,0,0);
  }
  switch(a->type)
  {
  case T_LIST:
  case T_POINTER:
  case T_OBJECT:
  case T_MAPPING:
  case T_FUNCTION:
    error("Cannot compare lists, arrays, mappings, functions or objects.\n",0,0,0,0,0,0,0,0,0);

  case T_NUMBER:
    return a->u.number > b->u.number;

  case T_STRING:
    return my_strcmp(a,b)>0;

  case T_FLOAT:
    return a->u.fnum > b->u.fnum;

  default:
    fatal("Unknown type %x\n",(void *)(int)a->type,0,0,0,0,0,0,0,0);
    return 0; /* make gcc happy */
  }
}

INLINE int is_lt(struct svalue *a, struct svalue *b)
{
  if (a->type != b->type)
  {
    error("Cannot compare different types.\n",0,0,0,0,0,0,0,0,0);
  }
  switch(a->type)
  {
  case T_LIST:
  case T_POINTER:
  case T_OBJECT:
  case T_MAPPING:
  case T_FUNCTION:
    error("Cannot compare lists, arrays, mappings, functions or objects.\n",0,0,0,0,0,0,0,0,0);

  case T_NUMBER:
    return a->u.number < b->u.number;

  case T_STRING:
    return my_strcmp(a, b)<0;

  case T_FLOAT:
    return a->u.fnum < b->u.fnum;

  default:
    fatal("Unknown type %x\n",(void *)(int)a->type,0,0,0,0,0,0,0,0);
    return 0; /* make gcc happy */
  }
}

INLINE void f_eq()
{
  int i;
  i=is_eq(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}


INLINE void f_ne()
{
  int i;
  i=!is_eq(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

INLINE void f_lt()
{
  int i;
  i=is_lt(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

INLINE void f_gt()
{
  int i;
  i=is_gt(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

INLINE void f_ge()
{
  int i;
  i=!is_lt(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

INLINE void f_le()
{
  int i;
  i=!is_gt(sp-1,sp);
  pop_stack();
  pop_stack();
  push_number(i);
}

char *add_buf=0;
char *buffered_string=0;
int add_buffered,add_buffer_size;

void add_strcat(char *b,int l)
{
  if(add_buffered+l>add_buffer_size)
  {
    add_buffer_size*=2;
    if(add_buffered+l>add_buffer_size)
      add_buffer_size=add_buffer_size+l+1;
    add_buf=realloc(add_buf,add_buffer_size);
  }
  memcpy(add_buf+add_buffered,b,l);
  add_buffered+=l;
}

void init_add_buf(char *foo,int len)
{
  if(buffered_string==foo)
  {
    if(buffered_string)
      free_string(buffered_string);
    return;
  }else{
    if(buffered_string)
      free_string(buffered_string);
  }
  
  if(!add_buf)
  {
    add_buffer_size=200;
    add_buf=(char *)xalloc(add_buffer_size);
  }
  add_buffered=0;
  add_strcat(foo,len);
}

char *free_add_buf()
{
  if(!add_buf) return 0;
  buffered_string=
    copy_shared_string(make_shared_binary_string(add_buf,add_buffered));
  return buffered_string;
}


void f_sum(int num_arg,struct svalue *argp)
{
  int e,type,size;
  type=argp[0].type;
  if(type==T_OBJECT || type==T_FUNCTION) bad_arg(1);

  for(e=1;e<num_arg;e++)
  {
    if(type!=argp[e].type)
    {
      type=T_STRING;
      break;
    }
  }
  for(e=0;e<num_arg;e++)
  {
    if(argp[e].type!=type &&
       argp[e].type!=T_STRING &&
       argp[e].type!=T_INT &&
       argp[e].type!=T_FLOAT)
      bad_arg(e+1);
  }
  switch(type)
  {
    case T_STRING:
    {
      char *buf,buffer[40];
      for(size=e=0;e<num_arg;e++)
      {
        switch(argp[e].type)
        {
          case T_STRING:
            size+=my_strlen(argp+e);
            break;

          case T_INT:
            size+=14;
            break;

          case T_FLOAT:
            size+=22;
            break;
        }
      }
      buf=malloc(size+1);
      size=0;
      for(size=e=0;e<num_arg;e++)
      {
        switch(argp[e].type)
        {
          case T_STRING:
            memcpy(buf+size,argp[e].u.string,my_strlen(argp+e));
            size+=my_strlen(argp+e);
            break;

          case T_INT:
            sprintf(buffer,"%d",argp[e].u.number);
            memcpy(buf+size,buffer,strlen(buffer));
            size+=strlen(buffer);
            break;

          case T_FLOAT:
            sprintf(buffer,"%f",argp[e].u.fnum);
            memcpy(buf+size,buffer,strlen(buffer));
            size+=strlen(buffer);
            break;
        }
      }
      buf[size]=0;
      pop_n_elems(num_arg);
      push_shared_string(make_shared_binary_string(buf,size));
      free(buf);
    }
    break;

    case T_INT:
    for(size=e=0;e<num_arg;e++) size+=argp[e].u.number;
    pop_n_elems(num_arg);
    push_number(size);
    break;

    case T_FLOAT:
    {
      float sum;
      sum=0.0;
      for(e=0;e<num_arg;e++) sum+=argp[e].u.fnum;
      pop_n_elems(num_arg);
      push_float(sum);
    }
    break;

    case T_POINTER:
    {
      struct vector *v;
      struct svalue *r;
      int d;
      for(size=e=0;e<num_arg;e++)
      {
	check_for_destr(argp[e].u.vec);
        size+=argp[e].u.vec->size;
      }
      if(argp[0].u.vec->ref==1)
      {
	e=argp[0].u.vec->size;
	v=resize_array(argp[0].u.vec,size);
	argp[0]=const0;
	size=e;
	e=1;
      }else{
	v=allocate_array(size);
	e=size=0;
      }
      for(;e<num_arg;e++)
      {
	r=argp[e].u.vec->item;
	for(d=0;d<argp[e].u.vec->size;d++)
	{
	  assign_svalue_no_free(v->item+size,r+d);
	  size++;
	}
      }
      pop_n_elems(num_arg);
      push_vector(v);
      v->ref--;
    }
    break;

    case T_LIST:
    {
      struct vector *v,*w;
      struct svalue *r;
      int d;
      for(size=e=0;e<num_arg;e++)
      {
	check_alist_for_destruct(argp[e].u.vec);
        size+=argp[e].u.vec->size;
      }
      if(num_arg==2)
      {
	w = do_array_surgery(argp[0].u.vec,argp[1].u.vec,OPER_ADD,T_LIST);
      }else{
        v=allocate_array(size);
        for(size=e=0;e<num_arg;e++)
        {
          r=argp[e].u.vec->item[0].u.vec->item;
          for(d=0;d<argp[e].u.vec->size;d++)
          {
            assign_svalue_no_free(v->item+size,r+d);
            size++;
          }
        }
        w=allocate_array(1);
        w->item[0].u.vec=v;
        w->item[0].type=T_ALIST_PART;
        order_alist(w);
      }
      pop_n_elems(num_arg);
      push_list(w);
      w->ref--;
    }
    break;

    case T_MAPPING:
    {
      struct vector *v,*values,*w;
      struct svalue *r,*rr;
      int d;
      for(size=e=0;e<num_arg;e++)
      {
	check_alist_for_destruct(argp[e].u.vec);
        size+=argp[e].u.vec->size;
      }
      if(num_arg==2)
      {
	w = do_array_surgery(argp[0].u.vec,argp[1].u.vec,OPER_ADD,T_MAPPING);
      }else{
        v=allocate_array(size);
        values=allocate_array(size);
        for(size=e=0;e<num_arg;e++)
        {
          r=argp[e].u.vec->item[0].u.vec->item;
          rr=argp[e].u.vec->item[1].u.vec->item;
          for(d=0;d<argp[e].u.vec->size;d++)
          {
            assign_svalue_no_free(v->item+size,r+d);
            assign_svalue_no_free(values->item+size,rr+d);
            size++;
          }
        }
        w=allocate_array(2);
        w->item[0].u.vec=v;
        w->item[0].type=T_ALIST_PART;
        w->item[1].u.vec=values;
        w->item[1].type=T_ALIST_PART;
        order_alist(w);
      }
      pop_n_elems(num_arg);
      push_mapping(w);
      w->ref--;
    }
    break;
  }
}

INLINE void f_add() { f_sum(2,sp-1); }

void f_subtract()
{
  if ((sp-1)->type != sp->type )
    error("Subtract on different types.\n",0,0,0,0,0,0,0,0,0);

  switch(sp->type)
  {
  case T_POINTER:
  {
    extern struct vector *subtract_array
      PROT((struct vector *,struct vector*));
    struct vector *v;

    /* subtract_array already takes care of destructed objects */
    check_for_destr(sp[-1].u.vec);
    check_for_destr(sp[0].u.vec);
    if(sp[0].u.vec->types & sp[-1].u.vec->types)
    {
      v = subtract_array((sp-1)->u.vec, sp->u.vec);
    }else{
      v=slice_array(sp[-1].u.vec,0,sp[-1].u.vec->size-1);
    }
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    return;
  }
  case T_LIST:
  case T_MAPPING:
  {
    struct vector *v;
    int type=sp->type;

    v=do_array_surgery((sp-1)->u.vec,sp->u.vec,OPER_SUB,type);
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    sp->type=type;
    return;
  }
  case T_FLOAT:
  {
    float foo;
    foo=(sp-1)->u.fnum-sp->u.fnum;
    pop_stack();
    sp->u.fnum=foo;
    return;
  }
  case T_NUMBER:
  {
    int i;
    i = (sp-1)->u.number - sp->u.number;
    pop_stack();
    sp->u.number = i;
    return;
  }

  case T_STRING:
  {
    push_string("",STRING_CONSTANT);
    f_replace(3,sp-2);
    return;
  }
  default:
    bad_arg(1);
  }
}

void f_and()
{
  int i;
  if(sp->type == (sp-1)->type && 
     (sp->type & (T_POINTER | T_LIST | T_MAPPING )))
  {
    struct vector *v;
    int type=sp->type;

    check_for_destr(sp[-1].u.vec);
    check_for_destr(sp[0].u.vec);
    if(sp[-1].u.vec->types & sp[0].u.vec->types)
    {
      v=do_array_surgery((sp-1)->u.vec,sp->u.vec,OPER_AND,type);
    }else{
      v=allocate_array(0);
    }
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    sp->type=type;
    return;
  }

  if ((sp-1)->type != T_NUMBER) bad_arg(1);
  if (sp->type != T_NUMBER) bad_arg(2);
  i = (sp-1)->u.number & sp->u.number;
  pop_stack();
  sp->u.number = i;
}

void f_or()
{
  int i;
  if(sp->type == (sp-1)->type &&
     (sp->type & (T_POINTER | T_LIST | T_MAPPING)))
  {
    struct vector *v;
    int type=sp->type;

    check_for_destr(sp[-1].u.vec);
    check_for_destr(sp[0].u.vec);
    if(sp[-1].u.vec->types & sp[0].u.vec->types)
    {
      v=do_array_surgery((sp-1)->u.vec,sp->u.vec,OPER_OR,type);
    }else{
      f_add();
      return;
    }
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    sp->type=type;
    return;
  }

  if ((sp-1)->type != T_NUMBER) bad_arg(1);
  if (sp->type != T_NUMBER) bad_arg(2);
  i = (sp-1)->u.number | sp->u.number;
  pop_stack();
  sp->u.number = i;
}

void f_xor()
{
  int i;
  if(sp->type == (sp-1)->type &&
     (sp->type & (T_POINTER | T_LIST | T_MAPPING )))
  {
    struct vector *v;
    int type=sp->type;

    check_for_destr(sp[-1].u.vec);
    check_for_destr(sp[0].u.vec);
    if(sp[-1].u.vec->types & sp[0].u.vec->types)
    {
      v=do_array_surgery((sp-1)->u.vec,sp->u.vec,OPER_XOR,type);
    }else{
      f_add();
      return;
    }
    pop_stack();
    pop_stack();
    push_vector(v);
    v->ref--;
    sp->type=type;
    return;
  }

  if ((sp-1)->type != T_NUMBER) bad_arg(1);
  if (sp->type != T_NUMBER) bad_arg(2);
  i = (sp-1)->u.number ^ sp->u.number;
  pop_stack();
  sp->u.number = i;
}

void f_lsh()
{
  int i;
  if ((sp-1)->type != T_NUMBER) bad_arg(1);
  if (sp->type != T_NUMBER) bad_arg(2);
  i = (sp-1)->u.number << sp->u.number;
  pop_stack();
  sp->u.number = i;
}

void f_rsh()
{
  int i;
  if ((sp-1)->type != T_NUMBER) bad_arg(1);
  if (sp->type != T_NUMBER) bad_arg(2);
  i = (sp-1)->u.number >> sp->u.number;
  pop_stack();
  sp->u.number = i;
}

void f_multiply()
{
  int i;
  if ((sp-1)->type == T_POINTER && sp->type == T_STRING)
  {
    f_implode(2,sp-1);
    return;
  }
  if ((sp-1)->type == T_FLOAT && sp->type == T_FLOAT)
  {
    float foo;
    foo=(sp-1)->u.fnum*sp->u.fnum;
    pop_stack();
    sp->u.fnum=foo;
    return;
  }

  if ((sp-1)->type != T_NUMBER) bad_arg(1);
  if (sp->type != T_NUMBER) bad_arg(2);
  i = (sp-1)->u.number * sp->u.number;
  pop_stack();
  sp->u.number = i;
}

void f_divide()
{
  int i;
  if ((sp-1)->type == T_STRING && (sp-1)->type == T_STRING)
  {
    f_explode(2,sp-1);
    return;
  }

  if ((sp-1)->type == T_FLOAT && sp->type == T_FLOAT)
  {
    float foo;
    if(sp->u.fnum==0.0)
      error("Division by zero.\n",0,0,0,0,0,0,0,0,0);
    foo=(sp-1)->u.fnum / sp->u.fnum;
    pop_stack();
    sp->u.fnum=foo;
    return;
  }

  if ((sp-1)->type != T_NUMBER) bad_arg(1);
  if (sp->type != T_NUMBER) bad_arg(2);
  if (sp->u.number == 0)
    error("Division by zero\n",0,0,0,0,0,0,0,0,0);
  i = (sp-1)->u.number / sp->u.number;
  pop_stack();
  sp->u.number = i;
}

void f_mod()
{
  int i;
  if ((sp-1)->type == T_FLOAT && sp->type == T_FLOAT)
  {
    float foo;
    if(sp->u.fnum==0.0)
      error("Modulo by zero.\n",0,0,0,0,0,0,0,0,0);
    foo=(sp-1)->u.fnum / sp->u.fnum;
    foo=(sp-1)->u.fnum-sp->u.fnum*(float)((int)floor(foo));
    pop_stack();
    sp->u.fnum=foo;
    return;
  }
  if ((sp-1)->type != T_NUMBER)  bad_arg(1);
  if (sp->type != T_NUMBER) bad_arg(2);
  if (sp->u.number == 0)
    error("Modulus by zero.\n",0,0,0,0,0,0,0,0,0);
  i = (sp-1)->u.number % sp->u.number;
  pop_stack();
  sp->u.number = i;
}

void f_not()
{
  if(sp->type==T_FLOAT)
  {
    pop_push_conditional(sp->u.fnum==0.0);
    return;
  } 
  pop_push_conditional(sp->type == T_NUMBER && sp->u.number == 0);
}

void f_compl()
{
  if (sp->type != T_NUMBER)
    error("Bad argument to ~\n",0,0,0,0,0,0,0,0,0);
  sp->u.number = ~ sp->u.number;
}

void f_negate()
{
  if(sp->type==T_FLOAT)
  {
    sp->u.fnum=-sp->u.fnum;
    return;
  }
  if (sp->type != T_NUMBER)
    error("Bad argument to unary minus\n",0,0,0,0,0,0,0,0,0);
  sp->u.number = - sp->u.number;
}

void f_inc()
{
  if (sp->type != T_LVALUE)
    error("Bad argument to ++\n",0,0,0,0,0,0,0,0,0);
  if(sp->u.lvalue->sval)
  {
    if (sp->u.lvalue->sval->type != T_NUMBER)
      error("++ of non-numeric argument\n",0,0,0,0,0,0,0,0,0);
    sp->u.lvalue->sval->u.number++;
    lvalue_to_svalue(sp,sp->u.lvalue);
  }else{
    sp++;
    lvalue_to_svalue(sp,sp[-1].u.lvalue);
    push_number(1);
    f_add();
    f_assign();
  }
}
void f_inc_and_pop() { f_inc(); f_pop_value(); }

void f_dec()
{
  if (sp->type != T_LVALUE)
    error("Bad argument to --\n",0,0,0,0,0,0,0,0,0);
  if(sp->u.lvalue->sval)
  {
    if (sp->u.lvalue->sval->type != T_NUMBER)
      error("-- of non-numeric argument\n",0,0,0,0,0,0,0,0,0);
    sp->u.lvalue->sval->u.number--;
    lvalue_to_svalue(sp,sp->u.lvalue);
  }else{
    sp++;
    lvalue_to_svalue(sp,sp[-1].u.lvalue);
    push_number(-1);
    f_add();
    f_assign();
  }
}

void f_dec_and_pop() { f_dec(); f_pop_value(); }
 

void f_post_inc()
{
  if (sp->type != T_LVALUE)
    error("Bad argument to ++\n",0,0,0,0,0,0,0,0,0);
  if(sp->u.lvalue->sval)
  {
    if (sp->u.lvalue->sval->type != T_NUMBER)
      error("++ of non-numeric argument\n",0,0,0,0,0,0,0,0,0);
    sp->u.lvalue->sval->u.number++;
    lvalue_to_svalue(sp,sp->u.lvalue);
    sp->u.number--;
  }else{
    sp++;
    lvalue_to_svalue(sp,sp[-1].u.lvalue);
    f_swap();
    sp++;
    assign_svalue(sp,sp-2);
    push_number(1);
    f_add();
    f_assign_and_pop();
  }
}

void f_post_dec()
{
  if (sp->type != T_LVALUE)
    error("Bad argument to --\n",0,0,0,0,0,0,0,0,0);
  if(sp->u.lvalue->sval)
  {
    if (sp->u.lvalue->sval->type != T_NUMBER)
      error("-- of non-numeric argument\n",0,0,0,0,0,0,0,0,0);
    sp->u.lvalue->sval->u.number--;
    lvalue_to_svalue(sp,sp->u.lvalue);
    sp->u.number++;
  }else{
    sp++;
    lvalue_to_svalue(sp,sp[-1].u.lvalue);
    f_swap();
    sp++;
    assign_svalue(sp,sp-2);
    push_number(-1);
    f_add();
    f_assign_and_pop();
  }
}

void f_assign()
{
#ifdef DEBUG
  if (sp[-1].type != T_LVALUE)
    error("Bad argument to F_ASSIGN\n",0,0,0,0,0,0,0,0,0);
#endif
  assign(sp[-1].u.lvalue,sp);
  f_swap();
  pop_stack();
}

void f_assign_and_pop()
{
  f_assign();
  pop_stack();
}

void f_is_equal(int num_arg,struct svalue *argp)
{
  int i;
  i=is_equal(argp,argp+1);
  pop_stack();
  pop_stack();
  push_number(i);
}
