#include <math.h>
#include <string.h>
#include "config.h"
#include "lint.h"
#include "array.h"
#include "interpret.h"
#include "exec.h"
#include "object.h"
#include "main.h"
#include "stralloc.h"
#include "save_objectII.h"

extern int d_flag;

int one_line=0;

void my_error(char *t)
{
  char *s;
  if(s=simple_free_buf()) free(s);
  error(t,0,0,0,0,0,0,0,0,0);
}

void save_svalue_list(struct vector *v)
{
  int e;
  for(e=0;e<v->size;e++)
  {
    if(e) my_putchar(',');
    save_svalue(&(v->item[e]));
  }
}

void save_mapping(struct vector *v)
{
  int e;
  for(e=0;e<v->item[0].u.vec->size;e++)
  {
    if(e) my_putchar(',');
    save_svalue(&(v->item[0].u.vec->item[e]));
    my_putchar(':');
    save_svalue(&(v->item[1].u.vec->item[e]));
  }
}

void save_string(char *e)
{
  if(one_line)
  {
    for(;*e;e++)
    {
      switch(*e)
      {
        case '\n':
          my_putchar('\\');
          my_putchar('n');
          break;

        case '\t':
          my_putchar('\\');
          my_putchar('t');
          break;

        case '\b':
          my_putchar('\\');
          my_putchar('b');
          break;

        case '"':
        case '\\':
          my_putchar('\\');
        default:
          my_putchar(*e);
      }
    }
  }else{
    for(;*e;e++)
    {
      switch(*e)
      {
        case '"':
        case '\n':
        case '\\':
          my_putchar('\\');
        default:
          my_putchar(*e);
      }
    }
  }
}

void save_object_desc(struct object *ob)
{
  char b[20];
  if(ob->obj_index)
  {
    my_strcat("H(\"");
    save_string(ob->obj_index);
    my_strcat("\")");
  }else{
    sprintf(b,"O(%d)",ob->clone_number);
    my_strcat(b);
  }
}

void save_svalue(struct svalue *s)
{
  switch(s->type)
  {
  case T_LVALUE:
    my_strcat("Lvalue");
    break;

  case T_INVALID:
    my_strcat("--");
    break;
    
  case T_STRING:
  {
    my_putchar('"');
    save_string(s->u.string);
    my_putchar('"');
    break;
  }

  case T_NUMBER:
  {
    char b[14];
    sprintf(b,"%d",s->u.number);
    my_strcat(b);
    break;
  }

  case T_FLOAT:
  {
    char b[30];
    sprintf(b,"F%0.8f",s->u.fnum);
    my_strcat(b);
    break;
  }

  case T_POINTER:
    my_strcat("({");
    save_svalue_list(s->u.vec);
    my_strcat("})");
    break;

  case T_ALIST_PART:
    my_strcat("(*");
    save_svalue_list(s->u.vec);
    my_strcat("*)");
    break;

  case T_LIST:
    my_strcat("(<");
    save_svalue_list(s->u.vec->item[0].u.vec);
    my_strcat(">)");
    break;

  case T_MAPPING:
    my_strcat("([");
    save_mapping(s->u.vec);
    my_strcat("])");
    break;

  case T_OBJECT:
    save_object_desc(s->u.ob);
    break;

  case T_FUNCTION:
    save_object_desc(s->u.ob);
    my_strcat("->\"");
    my_strcat(FUNC(s->u.ob->prog,s->subtype)->name);
    my_putchar('"');
    break;
     
  default:
    fatal("Unknown type %d in save_object.\n",(void *)(int)s->type,
	  0,0,0,0,0,0,0,0);
  }
}


char *restore_string(char **t)
{
  init_buf();
  (*t)++;
  while(**t!='"')
  {
    switch(**t)
    {
      case 0:
        my_error("Unexpected end of string in restore.\n");

      case '\\':
      {
        (*t)++;
        switch(**t)
        {
          case 0:
            my_error("Unexpected end of string in restore.\n");

          case 'n': my_putchar('\n'); break;
          case 't': my_putchar('\t'); break;
          case 'b': my_putchar('\b'); break;
          case '0':
          case '1':
          case '2':
          {
	    char c;
            c=(**t-'0')*8*8;
            (*t)++;
            if(**t<'0' || **t>'8') 
              my_error("Error in restore string.\n");
            c+=(**t-'0')*8;
            (*t)++;
            if(**t<'0' || **t>'8') 
              my_error("Error in restore string.\n");
            c+=(**t-'0');
            my_putchar(c);
          }
          default:  my_putchar(**t);  break;
        }
        (*t)++;
        break;
      }
      default:
        my_putchar(*((*t)++));
    }
  }
  (*t)++;
  return free_buf();
}

char *restore_svalue(char *t,struct svalue *s)
{
  free_svalue(s);
  switch(*t)
  {
  case '-': 
  case '0': case '1': case '2': case '3': case '4':
  case '5': case '6': case '7': case '8': case '9':
    s->u.number=(int)strtol(t,&t,10);
    s->type=T_NUMBER;
    s->subtype=NUMBER_NUMBER;
    return t;

  case 'O':
    s->type=T_NUMBER;
    s->subtype=NUMBER_DESTRUCTED_OBJECT;
    s->u.number=0;
    if(strncmp("O(",t,2)) my_error("Error in restore object");
    t+=2;
    while(*t>='0' && *t<='9') t++;
    if(*t!=')') my_error("Error in restore object");
    t++;
    return t;

  case 'F':
    s->type=T_FLOAT;
    t++;
    s->u.number=strtod(t,&t);
    return t;

  case 'H':
  {
    char *p;
    struct object *o;
    if(strncmp("H(\"",t,3)) my_error("Error in restore object");
    t+=2;
    p=restore_string(&t);
    if(*t!=')') free_string(p),my_error("Error in restore object");
    t++;
    o=find_hashed(p);
    if(o)
    {
      add_ref(o,"Restore object");
      s->type=T_OBJECT;
      s->u.ob=o;
    }else{
      s->type=T_NUMBER;
      s->subtype=NUMBER_DESTRUCTED_OBJECT;
      s->u.number=0;
    }
    free_string(p);
    if(*t=='-')
    {
      if(strncmp("->\"",t,3)) my_error("Error in restore object.\n");
      t+=2;
      p=restore_string(&t);
      if(o)
      {
	int tmp;
	tmp=find_function(p,o->prog);
	if(tmp==-1)
	{
	  s->subtype=NUMBER_DESTRUCTED_FUNCTION;
	  s->type=T_NUMBER;
	}else{
	  s->subtype=tmp;
	  s->type=T_FUNCTION;
	}
      }
      free_string(p);
    }
    return t;
  }
    
  case '"':
    s->u.string=restore_string(&t);
    s->type=T_STRING;
    s->subtype=STRING_SHARED;
    return t;

  case '(':
    switch(*++t)
    {
      int e;
      struct vector *v;

    case '{':
      t++; e=0;
      s->type=T_POINTER;
      s->u.vec=allocate_array(MAX_ARRAY_SIZE);
      while(*t!='}' && *t)
      {
	t=restore_svalue(t,&(s->u.vec->item[e]));
	if(*t==',') t++;
	e++;
      }
      if(!*t) my_error("Unexpected end of string in restore object.\n");
      if(*++t==')') t++;
      s->u.vec=shrink_array(s->u.vec,e);
      return t;

    case '<':
      t++; e=0;
      s->type=T_POINTER;
      s->u.vec=allocate_array(MAX_ARRAY_SIZE);
      while(*t!='>' && *t)
      {
	t=restore_svalue(t,&(s->u.vec->item[e]));
	if(*t==',') t++;
	e++;
      }
      if(!*t) my_error("Unexpected end of string in restore object.\n");
      if(*++t==')') t++;
      s->u.vec=shrink_array(s->u.vec,e);
      s->type=T_LIST;		/* Just in case */
      v=allocate_array(1);
      v->item[0].type=T_ALIST_PART;
      v->item[0].u.vec=s->u.vec;
      order_alist(v);
      s->u.vec=v;
      return t;

    case '[':
      t++; e=0;
      s->u.vec=allocate_mapping(
				allocate_array(MAX_ARRAY_SIZE),
				allocate_array(MAX_ARRAY_SIZE));
      s->type=T_POINTER;

      while(*t!=']' && *t)
      {
	t=restore_svalue(t,&(s->u.vec->item[0].u.vec->item[e]));
	if(*t!=':') my_error("Error while restoring mapping.\n");
	t++;
	t=restore_svalue(t,&(s->u.vec->item[1].u.vec->item[e]));
	if(*t==',') t++;
	e++;
      }
      if(!*t) my_error("Unexpected end of string in restore object.\n");
      if(*++t==')') t++;
      v=allocate_mapping(
			 shrink_array(s->u.vec->item[0].u.vec,e),
			 shrink_array(s->u.vec->item[1].u.vec,e));

      s->u.vec->item[0].type=T_NUMBER; /* freed by shrink_array */
      s->u.vec->item[1].type=T_NUMBER;
      free_vector(s->u.vec);
      order_alist(v);
      s->u.vec=v;
      s->type=T_MAPPING;
      return t;
    }
  }
  return t;
}

char *save_object()
{
    int i;

    if (current_object->flags & O_DESTRUCTED)
	error("Save_object on destructed object.\n",0,0,0,0,0,0,0,0,0);

    init_buf();
    for (i=0;i < current_object->prog->num_variables; i++)
    {
	struct svalue *v = &current_object->variables[i];

	if (current_object->prog->variable_names[i].type & TYPE_MOD_STATIC)
	    continue;
	my_strcat(current_object->prog->variable_names[i].name);
	my_putchar(' ');
	save_svalue(v);
	my_putchar('\n');
    }
    return free_buf();
}


int restore_object(char *t)
{
  char *var,q[150];
  struct variable *p;

  if (current_object->flags & O_DESTRUCTED)
    error("Restore object on destructed object.\n",0,0,0,0,0,0,0,0,0);

  while(*t)
  {
    extern struct svalue const0;
    struct svalue *v,tmp=const0;
    if(*t=='\n') { t++; continue; }
    t=strchr(var=t, ' ');
    if (t == 0 || t-var>=sizeof(q)-1)
      error("Illegal format when restore.\n",0,0,0,0,0,0,0,0,0);
    strncpy(q,var,t-var);
    q[t-var]=0;
    p = find_status(q, 0);
    t++;
    if (p == 0 || (p->type & TYPE_MOD_STATIC))
    {
      t=restore_svalue(t,&tmp);
      free_svalue(&tmp);
    }else{
      v = &current_object->variables[p - current_object->prog->variable_names];
      t=restore_svalue(t,v);
    }
    if(*t=='\n') t++;
  }
  if (d_flag > 1)
    debug_message("Object %s restored.\n", current_object->prog->name);
  return 1;
}
