#include <stdio.h>
#include <string.h>
#include "config.h"
#include "lint.h"
#include "array.h"
#include "wiz_list.h"
#include "interpret.h"
#include "main.h"
#include "stralloc.h"

#define DOMAINS
/*
 * Maintain the wizards high score list about most popular castle.
 */


struct wiz_list *all_wiz;

/*
 * Sort the wiz list in ascending order.
 */

static struct wiz_list *insert(struct wiz_list *w,struct wiz_list *wl)
{
    if (wl == 0)
    {
	w->next = 0;
	return w;
    }
    if (w->score > wl->score) {
	wl->next = insert(w, wl->next);
	return wl;
    }
    w->next = wl;
    return w;
}

/*
 * Find the data, if it exists.
 */
static struct wiz_list *find_wiz(char *name)
{
    int length;
    struct wiz_list *wl;

    length = strlen(name);
    for (wl = all_wiz; wl; wl = wl->next)
        if (wl->length == length && strcmp(wl->name, name) == 0)
	    return wl;
    return 0;
}

/*
 * Check that a name exists. Add it, if it doesn't.
 */
struct wiz_list *add_name(char *str)
{
  struct wiz_list *wl;

  wl = find_wiz(str);
  if (wl)
    return wl;
  wl = (struct wiz_list *)xalloc(sizeof (struct wiz_list));
  wl->name = make_shared_string(str);
  wl->length = strlen(str);
  wl->score = 0;
  wl->cost = 0;
  wl->heart_beats = 0;
  wl->total_worth = 0;
  wl->next = all_wiz;
  wl->size_array = 0;
  all_wiz = wl;
  return wl;
}

/*
 * Add score to an existing name.
 */
void add_score(char *name,int score)
{
  struct wiz_list *wl;

  wl = find_wiz(name);
  if (!wl)
    fatal("Add_score: could not find wizard %s\n", name,0,0,0,0,0,0,0,0);
  wl->score += score;
}

/*
 * This one is called at every complete walkaround of reset.
 */
void wiz_decay()
{
    struct wiz_list *wl;
    static int next_time;
    extern int current_time;

    /* Perform this once every hour. */
    if (next_time > current_time)
	return;
    next_time = current_time + 60 * 60;
    for (wl = all_wiz; wl; wl = wl->next) {
        wl->score = wl->score * 99 / 100;
	wl->total_worth = wl->total_worth * 99 / 100;
	wl->cost = wl->cost * 9 / 10;
	wl->heart_beats = wl->heart_beats * 9 / 10;
    }
}

/*
 * Load the wizlist file.
 */
void load_wiz_file()
{
    char buff[1000];		/* I hate not knowing how much I need. */
    FILE *f;

    f = fopen("WIZLIST", "r");
    if (f == NULL)
        return;
    while(fgets(buff, sizeof buff, f) != NULL)
    {
      char *p;
      int score;

      p = strchr(buff, ' ');
      if (p == 0) {
	fprintf(stderr, "Bad WIZLIST file.\n");
	break;
      }
      *p = '\0';
      p++;
      if (*p == '\0') {
	fprintf(stderr, "Bad WIZLIST file.\n");
	break;
      }
      score = atoi(p);
      if (score > 0 )
      {
	(void)add_name(buff);
	add_score(buff, atoi(p));
      }
    }
    fclose(f);
}

/*
 * Save the wizlist file.
 */
void save_wiz_file()
{
    struct wiz_list *wl;
    FILE *f;

    f = fopen("WIZLIST", "w");
    if (f == NULL) {
        fprintf(stderr, "Could not open WIZLIST for write\n");
        return;
    }
    for (wl = all_wiz; wl; wl = wl->next)
        fprintf(f, "%s %d %d\n", wl->name, wl->score, wl->total_worth);
    fclose(f);
}

struct vector *one_wizard(struct wiz_list *w)
{
  struct vector *v;
  v=allocate_array(6);
  v->item[0].u.string=copy_shared_string(w->name);
  v->item[0].type=T_STRING;
  v->item[0].subtype=STRING_SHARED;
  v->item[1].u.number=w->score;
  v->item[2].u.number=w->cost;
  v->item[3].u.number=w->heart_beats;
  v->item[4].u.number=w->total_worth;
  v->item[5].u.number=w->size_array;
}

struct vector *wizlist(char *w)
{
  struct wiz_list *wl;
  int s;
  struct vector *v;
  if(w)
  {
    wl=find_wiz(w);
    if(!wl) return 0;
    return one_wizard(wl);
  }else{
    s=0;
    for(wl=all_wiz;wl;wl=wl->next) s++;
    v=allocate_array(s);
    s=0;
    for(wl=all_wiz;wl;wl=wl->next)
    {
      v->item[s].u.vec=one_wizard(wl);
      v->item[s].type=T_POINTER;
      s++;
    }
    return v;
  }
}

void remove_wiz_list()
{
  struct wiz_list *wl, *w;

  for (w = all_wiz; w; w = wl)
  {
    free_string(w->name);
    wl = w->next;
    free((char *)w);
  }
}

